/*
 * Support for Intel 82596 device (run in 586 compatibility mode)
 * on the HP 700 series machines.
 *
 * Derived from the Mach Intel 82586 driver for the PC.
 *
 * TODO:
 *	Go native
 */
/* 
 * Mach Operating System
 * Copyright (c) 1991,1990,1989 Carnegie Mellon University
 * Copyright (c) 1991,1992,1994 The University of Utah and
 * the Computer Systems Laboratory (CSL).
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 * 
 * CARNEGIE MELLON, THE UNIVERSITY OF UTAH AND CSL ALLOW FREE USE OF
 * THIS SOFTWARE IN ITS "AS IS" CONDITION, AND DISCLAIM ANY LIABILITY
 * OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF
 * THIS SOFTWARE.
 * 
 * Carnegie Mellon requests users of this software to return to
 * 
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 * 
 * any improvements or extensions that they make and grant Carnegie Mellon
 * the rights to redistribute these changes.
 *
 * CSL requests users of this software to return to csl-dist@cs.utah.edu any
 * improvements that they make and grant CSL redistribution rights.
 *
 * 	Utah $Hdr: if_i596.c 3.42 94/12/16$
 */
/* 
 *	Olivetti PC586 Mach Ethernet driver v1.0
 *	Copyright Ing. C. Olivetti & C. S.p.A. 1988, 1989
 *	All rights reserved.
 *
 */ 
/*
  Copyright 1988, 1989 by Olivetti Advanced Technology Center, Inc.,
Cupertino, California.

		All Rights Reserved

  Permission to use, copy, modify, and distribute this software and
its documentation for any purpose and without fee is hereby
granted, provided that the above copyright notice appears in all
copies and that both the copyright notice and this permission notice
appear in supporting documentation, and that the name of Olivetti
not be used in advertising or publicity pertaining to distribution
of the software without specific, written prior permission.

  OLIVETTI DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE
INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS,
IN NO EVENT SHALL OLIVETTI BE LIABLE FOR ANY SPECIAL, INDIRECT, OR
CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
LOSS OF USE, DATA OR PROFITS, WHETHER IN ACTION OF CONTRACT,
NEGLIGENCE, OR OTHER TORTIOUS ACTION, ARISING OUR OF OR IN CONNECTION
WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

/*
  Copyright 1988, 1989 by Intel Corporation, Santa Clara, California.

		All Rights Reserved

Permission to use, copy, modify, and distribute this software and
its documentation for any purpose and without fee is hereby
granted, provided that the above copyright notice appears in all
copies and that both the copyright notice and this permission notice
appear in supporting documentation, and that the name of Intel
not be used in advertising or publicity pertaining to distribution
of the software without specific, written prior permission.

INTEL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE
INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS,
IN NO EVENT SHALL INTEL BE LIABLE FOR ANY SPECIAL, INDIRECT, OR
CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
LOSS OF USE, DATA OR PROFITS, WHETHER IN ACTION OF CONTRACT,
NEGLIGENCE, OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

/*
 * NOTE:
 *		by rvb:
 *  1.	The best book on the 82586 is:
 *		LAN Components User's Manual by Intel
 *	The copy I found was dated 1984.  This really tells you
 *	what the state machines are doing
 *  2.	In the current design, we only do one write at a time,
 *	though the hardware is capable of chaining and possibly
 *	even batching.  The problem is that we only make one
 *	transmit buffer available in sram space.
 *  3.	
 *  n.	Board Memory Map (require 32-byte cache alignment)
	RFA/FD	0x0000 - 0x02df	0x2e0 bytes
		 0x2e0 = 0x17 * 0x20 bytes
	RBD	0x02e0 - 0x369f	0x33c0 bytes
		0x33c0 = 0x17 * 0x240 bytes
				== 0x20 bytes (bd) + 0x220 bytes
	CU	0x36a0 - 0x379f	0x100 bytes
	TBD	0x37a0 - 0x395f	0x1c0 bytes
		 0x1c0 = No 14 * 0x20 bytes
	TBUF	0x3960 - 0x3f9f	0x640 bytes (> 1594(10))
	SCB	0x3fa0 - 0x3fbf	0x20 bytes
	ISCP	0x3fc0 - 0x3fdf	0x20 bytes
	SCP	0x3fe0 - 0x3fff	0x20 bytes
 *		
 */

/*
 * NOTE:
 *
 *	Currently this driver doesn't support trailer protocols for
 *	packets.  Once that is added, please remove this comment.
 */

#if 0
#define DEBUG
#define WATCHDOG
#define IF_CNTRS	MACH
#endif

#include "lan.h"
#if NLAN > 0

#include <sys/types.h>
#include <sys/time.h>
#include <machine/asp.h>
#ifdef USELEDS
#include <machine/led.h>
#endif

#ifdef	MACH_KERNEL
#include <machine/machspl.h>
#include <mach/machine/vm_types.h>
#include <kern/time_out.h>
#include <device/device_types.h>
#include <device/errno.h>
#include <device/io_req.h>
#include <device/if_hdr.h>
#include <device/if_ether.h>
#include <device/net_status.h>
#include <device/net_io.h>
#include <mach_glue.h>
#else
#include <machine/spl.h>
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/mbuf.h>
#include <sys/buf.h>
#include <sys/protosw.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <sys/errno.h>
#include <sys/syslog.h>

#include <net/if.h>
#include <net/netisr.h>
#include <net/route.h>

#ifdef	INET
#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/in_var.h>
#include <netinet/ip.h>
#include <netinet/if_ether.h>
#endif

#ifdef	NS
#include <netns/ns.h>
#include <netns/ns_if.h>
#endif

#include <machine/mtpr.h>
#include <hpsgc/bsd_glue.h>

#include "bpfilter.h"
#if NBPFILTER > 0
#include <net/bpf.h>
#include <net/bpfdesc.h>
#endif
#endif	/* MACH_KERNEL */

#ifdef	MACH_KERNEL
#include <device.h>
#include <if_i596var.h>
#else
#include <hpdev/device.h>
#include <hpsgc/if_i596var.h>
#endif

#define	LAN_BUFF_SIZE	(0x4000 + 16)

int	pc586probe(), pc586attach();
int	pc586intr(), pc586init(), pc586output(), pc586ioctl(), pc586reset();
int	pc586start();
#ifdef WATCHDOG
int	pc586watch();
#endif
int	pc586rcv(), pc586xmt(), pc586bldcu();
int	pc586diag(), pc586config();
char	*pc586bldru();
char	*ram_to_ptr();
u_short	ptr_to_ram();

char	pc586sram[LAN_BUFF_SIZE+32];	/* equiv mapped to make vtop simple */
int	xmt_watch = 0;

pc_softc_t	pc_softc[NLAN];

#ifdef	PC596XMT_BUG
/*
 * The PC596 board on hp700's apparently can lose CX interrupts
 * (i.e. interrupts when transmit is complete).  We used to deal
 * with this by a 3-second watchdog timeout.  This is not adequate
 * for the newer, smaller, all-on-one-board hp710's and hp705's;
 * these newer machines lose many more CX interrupts when busy,
 * seriously degrading network performance.
 * 
 * A simple datapoint illustrates this more clearly.  A 730 which
 * transmitted 4.7 million packets lost 21 CX interrupts.  A 710
 * which transmitted 134 thousand packets lost 139 interrupts.
 *
 * When PC596XMT_BUG is #define'd: On packet transmission, determine
 * how long to wait for CX interrupt.  When a CX interrupt arrives,
 * clear this timer.  After each call to hardclock() (see "trap.c"),
 * if the current time exceeds our timer, set the timer's "tv_sec"
 * field to "-1" and create a LAN interrupt.
 *
 * Note: these data structures are accessed in hardclock(); hence
 * they are not part of `pc_softc_t'.
 *
 * Note: If pc586_waitxmt is too small then you'll end up declaring
 * tx interrupts lost too quickly (hardware may still be retransmitting
 * due to collisions).  When that happens you'll walk on-top of the
 * hardware tx buffer and trash the packet the hardware is still trying
 * to put on the wire.
 *
 * Simple version of the hardware backoff alg:
 *
 * while (retries < MAX)
 *   wait 2**min(retries,10) * slot time
 *
 * By default MAX == 15 and the slot time is ~50us.  Compute it out
 * and it would seem that a packet could live in the hardware for
 * about 332750us.  Round it an even 350000 for a little slack.
 *
 */
struct timeval pc586_toutxmt[NLAN];	/* last time packet xmit'd */
long pc586_waitxmt = 350000;		/* us. to wait b4 declaring CX lost */
int pc586_eir[NLAN];			/* prochpa lan interrupt register */

#define pc586inton(unit) \
    { \
	extern int asp2procibit[], pc586intr(); \
	register int ibit = aspitab(INT_LAN, SPLLAN, pc586intr, unit, 0); \
	if (ibit > 0 && asp2procibit[ibit] > 0) \
		pc586_eir[unit] = asp2procibit[ibit]; \
	else \
		pc586_eir[unit] = 0; \
    }
#define pc586intoff(unit) \
    { \
	aspitab(INT_LAN, SPLLAN, (int (*)())NULL, unit, 0); \
	pc586_eir[unit] = 0; \
	timerclear(&pc586_toutxmt[unit]); \
    }
#else	/* !PC596XMT_BUG */
#define pc586inton(unit)  (void)aspitab(INT_LAN,SPLLAN,pc586intr,unit,0);
#define pc586intoff(unit) (void)aspitab(INT_LAN,SPLLAN,(int (*)())NULL,unit,0);
#endif	/* PC596XMT_BUG */
#define pc586chatt(unit)  pc_softc[unit].hwaddr->attn = 0

#define pc586putport(lanp, data) \
    { \
	extern int isgecko; \
	if (isgecko) { \
		(lanp)->port = (data); \
		DELAY(1000); \
		(lanp)->port = (data) >> 16; \
		DELAY(1000); \
	} else { \
		(lanp)->port = (data) >> 16; \
		DELAY(1000); \
		(lanp)->port = (data); \
		DELAY(1000); \
	 } \
    }

struct pc586_cntrs {
	struct {
		u_int xmt, xmti, xmtsi;
		u_int defer;
		u_int busy;
		u_int sleaze, intrinsic, intrinsic_count;
		u_int chain;
	} xmt; 
	struct {
		u_int rcv;
		u_int ovw;
		u_int crc;
		u_int frame;
		u_int rscerrs, ovrnerrs;
		u_int partial, bad_chain, fill;
	} rcv;
	u_int watch;
	u_int xmtwatch;
} pc586_cntrs[NLAN];

#ifdef	DEBUG
int	pc586debug = 0;
#endif

#ifdef	WATCHDOG
int	pc586timo = 5;
#endif

#ifdef	IF_CNTRS
int pc586_narp = 1, pc586_arp = 0;
int pc586_ein[32], pc586_eout[32]; 
int pc586_lin[128/8], pc586_lout[128/8]; 
static
log_2(no)
unsigned long no;
{
	return ({ unsigned long _temp__;
		asm("bsr %1, %0; jne 0f; xorl %0, %0; 0:" :
		    "=r" (_temp__) : "a" (no));
		_temp__;});
}
#endif

/*
 * pc586probe:
 *
 *	This function "probes" or checks for the pc586 board on the bus to see
 *	if it is there.  As far as I can tell, the best break between this
 *	routine and the attach code is to simply determine whether the board
 *	is configured in properly.  Currently my approach to this is to write
 *	and read a word from the SRAM on the board being probed.  If the word
 *	comes back properly then we assume the board is there.  The config
 *	code expects to see a successful return from the probe routine before
 *	attach will be called.
 *
 * input	: address device is mapped to, and unit # being checked
 * output	: a '1' is returned if the board exists, and a 0 otherwise
 *
 */
pc586probe(dev)
struct hp_device	*dev;
{
	caddr_t		addr = dev->hp_addr;
	int		unit = dev->hp_unit;
	volatile char	*b_prom;
	volatile char	*b_sram;
	pc_softc_t	*sp = &pc_softc[unit];
	extern int	isgecko;

	if (unit < 0 || unit > NLAN) {
		printf("pc%d: board out of range [0..%d]\n", unit, NLAN);
		return(0);
	}
	if (addr < (caddr_t)0xF0000000)
		return(0);
	if (*(int *)dev->hp_ctlr != 0)
		b_prom = (volatile char *) dev->hp_ctlr;
	else
		b_prom = (volatile char *)
			(isgecko ? LASI_LANADDR : LAN_PROM_ADDR);
	b_sram = (volatile char *) (((int)pc586sram + 31) & ~31);
#ifdef DEBUG
	if (pc586debug)
		printf("pc%d: probe: addr=0x%x pc586sram=0x%x b_sram=0x%x\n",
		       unit, dev->hp_addr, pc586sram, b_sram);
#endif
	sp->seated = TRUE;
#ifndef __GNUC__	/* HP compiler complains */
	sp->prom = (volatile char *)b_prom;
	sp->sram = (volatile char *)b_sram;
#else
	(volatile char *)sp->prom = (volatile char *)b_prom;
	(volatile char *)sp->sram = (volatile char *)b_sram;
#endif
	return(1);
}

/*
 * pc586attach:
 *
 *	This function attaches a PC586 board to the "system".  The rest of
 *	runtime structures are initialized here (this routine is called after
 *	a successful probe of the board).  Once the ethernet address is read
 *	and stored, the board's ifnet structure is attached and readied.
 *
 * input	: hp_device structure setup in autoconfig
 * output	: board structs and ifnet is setup
 *
 */
pc586attach(dev)
	struct hp_device *dev;
{
	struct	ifnet	*ifp;
	u_char		*addr_p;
	u_char		*b_addr;
	u_char		unit = (u_char)dev->hp_unit;	
	pc_softc_t	*sp = &pc_softc[unit];
	volatile scb_t	*scb_p;

#ifdef WATCHDOG
	sp->timer = -1;
	sp->xmttimer = -1;
#endif
#ifdef	MACH_KERNEL
	sp->mode = 0;
#endif
	sp->flags = 0;
	sp->open = 0;
#ifdef	MULTICAST
	{
		register int i;

		sp->sc_mcast.mc_count = 0;
		for (i = 0; i < PC_MCASTLEN; i++)
			sp->sc_mcrefcnt[i] = 0;
	}
#endif
	sp->hwaddr = (struct core_lan *) dev->hp_addr;
	sp->hwaddr->reset = 0;
	b_addr = (u_char *)(sp->prom + OFFSET_PROM);
	addr_p = (u_char *)sp->ds_addr;
	addr_p[0] = b_addr[0];
	addr_p[1] = b_addr[1];
	addr_p[2] = b_addr[2];
	addr_p[3] = b_addr[3];
	addr_p[4] = b_addr[4];
	addr_p[5] = b_addr[5];
	printf("ethernet id [%x:%x:%x:%x:%x:%x]\n",
		addr_p[0], addr_p[1], addr_p[2],
		addr_p[3], addr_p[4], addr_p[5]);

	scb_p = (volatile scb_t *)(sp->sram + OFFSET_SCB);
	scb_p->scb_crcerrs = 0;			/* initialize counters */
	scb_p->scb_alnerrs = 0;
	scb_p->scb_rscerrs = 0;
	scb_p->scb_ovrnerrs = 0;
	flushscball(scb_p);

	ifp = &(sp->ds_if);
	ifp->if_unit = unit;
	ifp->if_mtu = ETHERMTU;
	ifp->if_flags = IFF_BROADCAST;
#ifdef	MACH_KERNEL
	ifp->if_header_size = sizeof(struct ether_header);
	ifp->if_header_format = HDR_ETHERNET;
	ifp->if_address_size = 6;
	ifp->if_address = (char *)&sp->ds_addr[0];
	if_init_queues(ifp);
#else
	ifp->if_name = "pc";
	ifp->if_init = pc586init;
	ifp->if_output = ether_output;
	ifp->if_start = pc586start;
	ifp->if_ioctl = pc586ioctl;
	ifp->if_reset = pc586reset;
#ifdef IFF_SIMPLEX
	ifp->if_flags |= IFF_SIMPLEX;
#endif
#if defined(MULTICAST) && defined(IFF_MULTICAST)
	ifp->if_flags |= IFF_MULTICAST;
#endif
	ifp->if_next = NULL;
#if NBPFILTER > 0
	bpfattach(&sp->sc_bpf, ifp, DLT_EN10MB, sizeof(struct ether_header));
#endif
	if_attach(ifp);
#endif
}

/*
 * pc586reset:
 *
 *	This routine is in part an entry point for the "if" code.  Since most 
 *	of the actual initialization has already (we hope already) been done
 *	by calling pc586attach().
 *
 * input	: unit number or board number to reset
 * output	: board is reset
 *
 */
pc586reset(unit)
int	unit;
{
	pc_softc[unit].ds_if.if_flags &= ~IFF_RUNNING;
#ifdef IFF_OACTIVE
	pc_softc[unit].ds_if.if_flags &= ~IFF_OACTIVE;
#endif
	pc_softc[unit].flags &= ~(DSF_LOCK|DSF_RUNNING);
	return(pc586init(unit));

}

/*
 * pc586init:
 *
 *	Another routine that interfaces the "if" layer to this driver.  
 *	Simply resets the structures that are used by "upper layers".  
 *	As well as calling pc586hwrst that does reset the pc586 board.
 *
 * input	: board number
 * output	: structures (if structs) and board are reset
 *
 */	
pc586init(unit)
int	unit;
{
	struct	ifnet	*ifp;
	int		stat;
	unsigned long	oldpri;

	ifp = &(pc_softc[unit].ds_if);
#ifndef	MACH_KERNEL
	if (ifp->if_addrlist == (struct ifaddr *)0)
		return(FALSE);
#endif
	oldpri = splimp();
	if ((stat = pc586hwrst(unit)) == TRUE) {
#ifdef WATCHDOG
		untimeout(pc586watch, &(ifp->if_unit));
		timeout(pc586watch, &(ifp->if_unit), pc586timo*hz);
		pc_softc[unit].timer = pc586timo;
#endif
		pc_softc[unit].ds_if.if_flags |= IFF_RUNNING;
		pc_softc[unit].flags |= DSF_RUNNING;
#ifdef IFF_OACTIVE
		pc_softc[unit].ds_if.if_flags &= ~IFF_OACTIVE;
#else
		pc_softc[unit].tbusy = 0;
#endif
#ifdef MACH_KERNEL
		pc586start(unit);
#else
		pc586start(ifp);
#endif
	} else
		printf("pc%d init(): trouble resetting board.\n", unit);
	splx(oldpri);
	return(stat);
}

#ifdef	MACH_KERNEL
/*ARGSUSED*/
pc586open(dev, flag)
	dev_t	dev;
	int	flag;
{
	register int	unit;
	pc_softc_t	*sp;

	unit = minor(dev);	/* XXX */
	if (unit < 0 || unit >= NLAN || !pc_softc[unit].seated)
	    return (ENXIO);

	pc_softc[unit].ds_if.if_flags |= IFF_UP;
	pc586init(unit);
	return (0);
}
#endif	/* MACH_KERNEL */

/*
 * pc586start:
 *
 *	This is yet another interface routine that simply tries to output a
 *	in an mbuf after a reset.
 *
 * input	: board number
 * output	: stuff sent to board if any there
 *
 */
#ifdef	MACH_KERNEL
pc586start(unit)
int	unit;
{
	io_req_t		m;
	struct	ifnet		*ifp;
	register pc_softc_t	*is = &pc_softc[unit];

#ifdef DEBUG
	if (pc586debug)
		printf("pc%d: start() busy %d\n", unit, is->tbusy);
#endif

	if (is->tbusy) {
		pc586_cntrs[unit].xmt.busy++;
		return;
	}

	ifp = &(pc_softc[unit].ds_if);
	IF_DEQUEUE(&ifp->if_snd, m);
	if (m) {
		is->tbusy++;
		pc586_cntrs[unit].xmt.xmt++;
		pc586xmt(unit, m);
	}
}
#else
pc586start(ifp)
	struct ifnet *ifp;
{
	struct	mbuf		*m;
	int			unit = ifp->if_unit;
	register pc_softc_t	*is = &pc_softc[unit];

#ifdef DEBUG
	if (pc586debug)
		printf("pc%d: start() busy %x\n", unit,
#ifdef IFF_OACTIVE
		       is->ds_if.if_flags & IFF_OACTIVE);
#else
		       is->tbusy);
#endif
#endif

	if ((is->ds_if.if_flags & IFF_RUNNING) == 0)
		return(0);

#ifdef IFF_OACTIVE
	if (is->ds_if.if_flags & IFF_OACTIVE) {
#else
	if (is->tbusy) {
#endif
		pc586_cntrs[unit].xmt.busy++;
		return (0);
	}

	do {
		IF_DEQUEUE(&is->ds_if.if_snd, m);
		if (m == 0)
			return(0);
#if NBPFILTER > 0
		/*
		 * If bpf is listening on this interface, let it
		 * see the packet before we commit it to the wire.
		 */
		if (is->sc_bpf)
			bpf_mtap(is->sc_bpf, m);
#endif
#ifdef IFF_OACTIVE
		is->ds_if.if_flags |= IFF_OACTIVE;
#else
		is->tbusy++;
#endif
		pc586_cntrs[unit].xmt.xmt++;
		pc586xmt(unit, m);
#ifdef IFF_OACTIVE
	} while ((is->ds_if.if_flags & IFF_OACTIVE) == 0);
#else
	} while (is->tbusy == 0);
#endif

	return(0);
}
#endif

/*
 * pc586read:
 *
 *	This routine does the actual copy of data (including ethernet header
 *	structure) from the pc586 to an mbuf chain that will be passed up
 *	to the "if" (network interface) layer.  NOTE:  we currently
 *	don't handle trailer protocols, so if that is needed, it will
 *	(at least in part) be added here.  For simplicities sake, this
 *	routine copies the receive buffers from the board into a local (stack)
 *	buffer until the frame has been copied from the board.  Once in
 *	the local buffer, the contents are copied to an mbuf chain that
 *	is then enqueued onto the appropriate "if" queue.
 *
 * input	: board number, and an frame descriptor pointer
 * output	: the packet is put into an mbuf chain, and passed up
 * assumes	: if any errors occur, packet is "dropped on the floor"
 *
 */
pc586read(unit, fd_p)
int	unit;
fd_t	*fd_p;
{
	register pc_softc_t	*is = &pc_softc[unit];
	register struct ifnet	*ifp = &is->ds_if;
	struct	ether_header	eh;
#ifdef	MACH_KERNEL
	ipc_kmsg_t		new_kmsg;
	struct ether_header	*ehp;
	struct packet_header	*pkt;
	char			*dp;
	u_short			bytes_in_msg, bytes_in_mbuf, bytes;
	u_short			mlen, len, clen;
	u_char			*mb_p;
	volatile rbd_t		*rbd_p;
	u_char			*buffer_p;
#else
    	struct	mbuf		*m;
	extern struct mbuf	*m_pc586devget();
#endif

#if defined(KGDB_LAN) || defined(TGDB)
	if (pc586debug_read(unit, fd_p))
		return 1;
	if ((ifp->if_flags & (IFF_UP|IFF_RUNNING)) != (IFF_UP|IFF_RUNNING))
		return 1;	/* leave interrupts on, silently drop packet */
#else
	if ((ifp->if_flags & (IFF_UP|IFF_RUNNING)) != (IFF_UP|IFF_RUNNING)) {
		printf("pc%d read(): board is not running.\n", ifp->if_unit);
		pc586intoff(ifp->if_unit);
	}
#endif
	pc586_cntrs[unit].rcv.rcv++;
#ifdef	MACH_KERNEL
	new_kmsg = net_kmsg_get();
	if (new_kmsg == IKM_NULL) {
	    /*
	     * Drop the received packet.
	     */
	    is->ds_if.if_rcvdrops++;

	    /*
	     * not only do we want to return, we need to drop the packet on
	     * the floor to clear the interrupt.
	     */
	    return 1;
	}
	ehp = (struct ether_header *) (&net_kmsg(new_kmsg)->header[0]);
	pkt = (struct packet_header *)(&net_kmsg(new_kmsg)->packet[0]);

	/*
	 * Get ether header.
	 */
	ehp->ether_type = fd_p->length;
	len = sizeof(struct ether_header);
	bcopy16(fd_p->source, ehp->ether_shost, ETHER_ADD_SIZE);
	bcopy16(fd_p->destination, ehp->ether_dhost, ETHER_ADD_SIZE);

	/*
	 * Get packet body.
	 */
	dp = (char *)(pkt + 1);

	rbd_p = (volatile rbd_t *)ram_to_ptr(fd_p->rbd_offset, unit);
	if (rbd_p == 0) {
	    printf("pc%d read(): Invalid buffer\n", unit);
	    if (pc586hwrst(unit) != TRUE) {
		printf("pc%d read(): hwrst trouble.\n", unit);
	    }
	    net_kmsg_put(new_kmsg);
	    return 0;
	}

	do {
	    pcacheline(KERNSPC, (caddr_t)rbd_p);
	    buffer_p = (u_char *)
		    (rbd_p->buffer_base << 16) + rbd_p->buffer_addr;
	    bytes_in_msg = rbd_p->status & RBD_SW_COUNT;
	    pdcache(KERNSPC, (caddr_t)buffer_p, bytes_in_msg);
	    bcopy16((u_short *)buffer_p,
		       (u_short *)dp,
		       (bytes_in_msg + 1) & ~1);	/* but we know it's even */
	    len += bytes_in_msg;
	    dp += bytes_in_msg;
	    if (rbd_p->status & RBD_SW_EOF)
		break;
	    rbd_p = (volatile rbd_t *)ram_to_ptr(rbd_p->next_rbd_offset, unit);
	} while ((int) rbd_p);

	pkt->type = ehp->ether_type;
	pkt->length =
		len - sizeof(struct ether_header)
		    + sizeof(struct packet_header);

	/*
	 * Send the packet to the network module.
	 */
	net_packet(ifp, new_kmsg, pkt->length, ethernet_priority(new_kmsg));
	return 1;
#else
	ifp->if_ipackets++;
	eh.ether_type = ntohs(fd_p->length);
	bcopy16(fd_p->source, eh.ether_shost, ETHER_ADD_SIZE);
	bcopy16(fd_p->destination, eh.ether_dhost, ETHER_ADD_SIZE);

	if ((m = m_pc586devget(ifp, fd_p)) == NULL) {
		/*
		 * We've lost this packet, either due to lack of mbufs
		 * or because of hardware problems.  Regardless, just
		 * reset the hardware so that we never see it again.
		 */
		if (pc586hwrst(unit) != TRUE) {
			printf("pc%d read: hardware reset failed\n", unit);
#ifdef WATCHDOG
			pc_softc[unit].timer = 0;
#endif
		}
		return (0);
	}

#if NBPFILTER > 0
	/*
	 * Check if there's a bpf filter listening on this interface.
	 * If so, hand off the raw packet to bpf, which must deal with
	 * trailers in its own way.
	 */
	if (is->sc_bpf) {
		struct mbuf *meth, *mtmp;

		/*
		 * We need to prepend the Ethernet header for BPF (and
		 * then remove it again).  Rather than screw around with
		 * our nice packet, we rip off another mbuf and stuff
		 * the Ethernet header into it.
		 */
		MGET(meth, M_DONTWAIT, MT_DATA);
		if (meth) {
			bcopy16((caddr_t)&eh, mtod(meth, caddr_t), sizeof(eh));
			meth->m_len = sizeof(eh);
			meth->m_next = m;
			bpf_mtap(is->sc_bpf, meth);
			MFREE(meth, mtmp);
		}
	}
#endif

	/*
	 * If we are in promiscuous mode, we have to check if this
	 * packet is really ours.
	 *
	 * XXX This test does not support multicasts.
	 */
	if ((ifp->if_flags & IFF_PROMISC)
	    && bcmp(eh.ether_dhost, is->ds_addr, sizeof(eh.ether_dhost)) != 0
	    && bcmp(eh.ether_dhost, etherbroadcastaddr,
	            sizeof(eh.ether_dhost)) != 0) {
		m_freem(m);
		return 1;	/* successfully received, just dont want it */
	}

#ifdef	IF_CNTRS
	clen = m->m_pkthdr.len + sizeof (struct ether_header);
	pc586_ein[log_2(clen)]++;
	if (clen < 128) pc586_lin[clen>>3]++;

	if (eh.ether_type == ETHERTYPE_ARP) {
		pc586_arp++;
		if (pc586_narp) {
			pc586_ein[log_2(clen)]--;
			if (clen < 128) pc586_lin[clen>>3]--;
		}
	}
#endif

	/*
	 * received packet is now in a chain of mbuf's.  next step is
	 * to pass the packet upwards.
	 */
	ether_input(ifp, &eh, m);
	return 1;
#endif	/* MACH_KERNEL */
}

#ifndef MACH_KERNEL
/*
 * Read in a packet from the ethernet card.
 * You can only read one buffer (`bp') of size `blen' at a given time.
 * Each read must be halfword aligned (regardless of `blen').
 * Keep reading and filling up mbufs until the hardware says "stop".
 * And yes, this is very brain-damaged (not to mention, SLOW!)...
 */
struct mbuf *
m_pc586devget(ifp, fd)
	struct ifnet *ifp;
	fd_t *fd;
{
	struct mbuf *top = 0, **mp = &top;
	register struct mbuf *m;
	register int len, blen, mclrem = 0;
	int unit = ifp->if_unit;
	volatile rbd_t *rbd;
	u_char *bp;
	int btot = 0;

	rbd = (volatile rbd_t *) ram_to_ptr(fd->rbd_offset, unit);
	if (rbd == NULL) {
		printf("pc%d read: invalid buffer\n", unit);
		return(NULL);
	}

	MGETHDR(m, M_DONTWAIT, MT_DATA);
	if (m == 0) {
		printf("pc%d read: out of mbufs (MGETHDR)\n", unit);
		return (0);
	}
	m->m_pkthdr.rcvif = ifp;
	m->m_len = MHLEN;

	do {
		/*
		 * Determine length and location of next buffer.
		 */
		pcacheline(KERNSPC, (caddr_t)rbd);	/* XXX: unnecessary? */
		blen = rbd->status & RBD_SW_COUNT;
		bp = (u_char *)(((u_int)rbd->buffer_base << 16) + rbd->buffer_addr);
#ifdef DEBUG
		if (pc586debug > 1 || pc586debug) {
			printf("pc%d rbuf base=%x addr=%x bufp=%x len=%d\n",
			  unit, rbd->buffer_base, rbd->buffer_addr, bp, blen);
			pdcache(KERNSPC, (caddr_t)bp, blen);
			dumpbytes(bp, blen);
		} else
#endif
		pdcache(KERNSPC, (caddr_t)bp, blen);
getmbuf:
		len = blen;
		if (len > mclrem || (mclrem & 1)) {
			mclrem = 0;
			/*
			 * Grab an mbuf for the data.
			 * `len' will eventually be set to the size of our mbuf.
			 */
			if (top) {
				MGET(m, M_DONTWAIT, MT_DATA);
				if (m == 0) {
					printf("pc%d read: out of mbufs\n",
					       unit);
					m_freem(top);
					return (0);
				}
				m->m_len = MLEN;
			}
			if (len >= MINCLSIZE) {
				MCLGET(m, M_DONTWAIT);
				if (m->m_flags & M_EXT) {
					m->m_len = len = min(len, MCLBYTES);
					mclrem = MCLBYTES;
				} else
					len = m->m_len;
			} else {
				/*
				 * Place initial small packet/header at end.
				 */
				if (len < m->m_len) {
					if (top == 0 &&
					    len + max_linkhdr <= m->m_len)
						m->m_data += max_linkhdr;
					m->m_len = len;
				} else
					len = m->m_len;
			}
			*mp = m;
			mp = &m->m_next;
		}

		/*
		 * Determine how many bytes we can really read.
		 * This simple decision is complicated because reads
		 * must be halfword aligned?
		 *
		 * We are guaranteed that we have an even number
		 * of bytes in our mbuf (i.e. `len' will be even).
		 * Therefore, if `len' is odd after the MIN(), then
		 * `blen' is odd *and* we can safely stuff one more
		 * byte into the mbuf!
		 */
		len = MIN(blen,len);
		if (mclrem && (m->m_flags & M_EXT)) {	/* M_EXT for sanity */
			char *dp = mtod(m, caddr_t) + (MCLBYTES - mclrem);
			bcopy16(bp, dp, (len + 1) & ~1);
			if (mclrem != MCLBYTES)
				m->m_len += len;
			mclrem -= len;
		} else {
			bcopy16(bp, mtod(m, caddr_t), (len + 1) & ~1);
			m->m_len = len;
		}
		bp += len;
		btot += len;

		/*
		 * Determine if we have read everything in this buffer.
		 * If so, maybe were finished... then again, maybe there
		 * is another buffer which is part of this same packet.
		 */
		blen -= len;
		if (blen)
			goto getmbuf;
		if (rbd->status & RBD_SW_EOF)
			break;
		rbd = (volatile rbd_t *)ram_to_ptr(rbd->next_rbd_offset, unit);
	} while (rbd != NULL);
	top->m_pkthdr.len = btot;
	return (top);
}
#endif	/* !MACH_KERNEL */

#ifdef	MACH_KERNEL
pc586output(dev, ior)
	dev_t	dev;
	io_req_t ior;
{
	register int	unit;

	unit = minor(dev);	/* XXX */
	if (unit < 0 || unit >= NLAN || !pc_softc[unit].seated)
	    return (ENXIO);

	return (net_write(&pc_softc[unit].ds_if, pc586start, ior));
}

pc586setinput(dev, receive_port, priority, filter, filter_count)
	dev_t		dev;
	mach_port_t	receive_port;
	int		priority;
	filter_t	filter[];
	unsigned int	filter_count;
{
	register int unit = minor(dev);
	if (unit < 0 || unit >= NLAN || !pc_softc[unit].seated)
	    return (ENXIO);

	return (net_set_filter(&pc_softc[unit].ds_if,
			receive_port, priority,
			filter, filter_count));
}

pc586getstat(dev, flavor, status, count)
	dev_t	dev;
	int	flavor;
	dev_status_t	status;		/* pointer to OUT array */
	unsigned int	*count;		/* out */
{
	register int	unit = minor(dev);
	register pc_softc_t	*sp;

	if (unit < 0 || unit >= NLAN || !pc_softc[unit].seated)
	    return (ENXIO);

	sp = &pc_softc[unit];
	return (net_getstat(&sp->ds_if, flavor, status, count));
}

pc586setstat(dev, flavor, status, count)
	dev_t	dev;
	int	flavor;
	dev_status_t	status;
	unsigned int	count;
{
	register int	unit = minor(dev);
	register pc_softc_t	*sp;

	if (unit < 0 || unit >= NLAN || !pc_softc[unit].seated)
	    return (ENXIO);

	sp = &pc_softc[unit];

	switch (flavor) {
	    case NET_STATUS:
	    {
		/*
		 * All we can change are flags, and not many of those.
		 */
		register struct net_status *ns = (struct net_status *)status;
		int	mode = 0;

		if (count < NET_STATUS_COUNT)
		    return (D_INVALID_OPERATION);

		if (ns->flags & IFF_ALLMULTI)
		    mode |= MOD_ENAL;
		if (ns->flags & IFF_PROMISC)
		    mode |= MOD_PROM;

		/*
		 * Force a complete reset if the receive mode changes
		 * so that these take effect immediately.
		 */
		if (sp->mode != mode) {
		    sp->mode = mode;
		    if (sp->flags & DSF_RUNNING) {
			sp->flags &= ~(DSF_LOCK|DSF_RUNNING);
			pc586init(unit);
		    }
		}
		break;
	    }

	    default:
		return (D_INVALID_OPERATION);
	}
	return (D_SUCCESS);

}
#else
/*
 * pc586ioctl:
 *
 *	This routine processes an ioctl request from the "if" layer
 *	above.
 *
 * input	: pointer the appropriate "if" struct, command, and data
 * output	: based on command appropriate action is taken on the
 *	 	  pc586 board(s) or related structures
 * return	: error is returned containing exit conditions
 *
 */
pc586ioctl(ifp, cmd, data)
struct ifnet	*ifp;
int	cmd;
caddr_t	data;
{
	register struct ifaddr	*ifa = (struct ifaddr *)data;
	int			unit = ifp->if_unit;
	register pc_softc_t	*is = &pc_softc[unit];
	int			opri, error = 0;

	opri = splimp();
	switch (cmd) {
	case SIOCSIFADDR:
		ifp->if_flags |= IFF_UP;
		pc586init(unit);
		switch (ifa->ifa_addr->sa_family) {
#ifdef INET
		case AF_INET:
			((struct arpcom *)ifp)->ac_ipaddr = IA_SIN(ifa)->sin_addr;
			arpwhohas((struct arpcom *)ifp, &IA_SIN(ifa)->sin_addr);
			break;
#endif
#ifdef NS
		case AF_NS:
    		    {
			register struct ns_addr *ina = &(IA_SNS(ifa)->sns_addr);

			if (ns_nullhost(*ina))
				ina->x_host = *(union ns_host *)(is->ds_addr);
			else {
				bcopy((caddr_t)ina->x_host.c_host,
				      (caddr_t)is->ds_addr,sizeof(is->ds_addr));
				if (pc586hwrst(unit) != TRUE)
					pc586start(ifp);
                        }
			break;
    		    }
#endif
		}
		break;
	case SIOCSIFFLAGS:
		if ((ifp->if_flags & IFF_UP) == 0 && is->flags & DSF_RUNNING) {
			printf("pc%d ioctl(): board is not running\n", unit);
			is->flags &= ~(DSF_LOCK | DSF_RUNNING);
#ifdef WATCHDOG
			is->timer = -1;
#endif
			pc586intoff(unit);
		} else if (ifp->if_flags & IFF_UP && (is->flags & DSF_RUNNING) == 0) {
			pc586init(unit);
		}
		/*
		 * If the state of the promiscuous bit changes, the interface
		 * must be reset to effect the change.
		 */
		if (((ifp->if_flags ^ is->sc_iflags) & IFF_PROMISC) &&
		    (is->flags & DSF_RUNNING)) {
			is->flags &= ~(DSF_LOCK|DSF_RUNNING);
			is->sc_iflags = ifp->if_flags;
			pc586init(unit);
		}
		break;

#ifdef MULTICAST
#define ifr ((struct ifreq *)ifa)
	/*
	 * XXX this needs to be integrated with the generic routines in
	 * net/if_ethersubr.c.
	 */
	case SIOCADDMULTI:
		if ( ifr->ifr_addr.sa_data[0]&1 == 0 )
			error = EADDRNOTAVAIL;
		else {
			/*
			 * XXX see ETHER_MAP_IP_MULTICAST in if_ether.h
			 */
                        ifr->ifr_addr.sa_data[0] = 1;
                        ifr->ifr_addr.sa_data[1] = 0;
                        ifr->ifr_addr.sa_data[2] = 0x5e;

			error = pc586multicast(1, unit, ifr->ifr_addr.sa_data);
			is->flags &= ~(DSF_LOCK|DSF_RUNNING);
			pc586init(unit);
		}
		break;
 
	case SIOCDELMULTI:
		if ( ifr->ifr_addr.sa_data[0]&1 == 0 )
			error = EADDRNOTAVAIL;
		else {
			error = pc586multicast(0, unit, ifr->ifr_addr.sa_data);
			is->flags &= ~(DSF_LOCK|DSF_RUNNING);
			pc586init(unit);
		}
		break;
#undef ifr
#endif MULTICAST

#ifdef	IF_CNTRS
	case SIOCCIFCNTRS:
		if (!suser()) {
			error = EPERM;
			break;
		}
		bzero((caddr_t)pc586_ein, sizeof (pc586_ein));
		bzero((caddr_t)pc586_eout, sizeof (pc586_eout));
		bzero((caddr_t)pc586_lin, sizeof (pc586_lin));
		bzero((caddr_t)pc586_lout, sizeof (pc586_lout));
		bzero((caddr_t)&pc586_arp, sizeof (int));
		bzero((caddr_t)&pc586_cntrs, sizeof (pc586_cntrs));
		break;
#endif
	default:
		error = EINVAL;
	}
	splx(opri);
	return (error);
}
#endif	/* MACH_KERNEL */

#ifdef	MULTICAST
int
pc586multicast(addmc, unit, addr)
	int addmc, unit;
	char addr[ETHER_ADD_SIZE];
{
	register pc_softc_t *pcp = &pc_softc[unit];
	register u_char *mcap;
	register u_short *mref;
	register int i, match = 0;
	int s, error = 0;

	s = splimp();

	/*
	 * See if this multicast address is already in our list;
	 * we will need this information no matter what we do.
	 */
	for (i=0, mcap = &pcp->sc_mcast.mc_addr[0], mref = &pcp->sc_mcrefcnt[0];
	     i < PC_MCASTLEN; i++, mcap += ETHER_ADD_SIZE, mref++) {
		if (*mref == 0)
			break;
		if (bcmp((char *)mcap, &addr[0], ETHER_ADD_SIZE) == 0) {
			match = 1;
			break;
		}
	}

	/*
	 * Here is the situation...
	 *
	 *	(i < PC_MCASTLEN) && (!match):
	 *		no matching entry found,
	 *		`mcap' & `mref' point to next available entry.
	 *	(i < PC_MCASTLEN) && (match):
	 *		matching entry found,
	 *		`mcap' points to entry matching `addr',
	 *		`mref' points to refcount for `mcap'.
	 *	(i == PC_MCASTLEN) || (!match):
	 *		no matching entry found,
	 *		`mcap' & `mref' invalid (no room left in `mcap').
	 */
	if (addmc) {
		if (i < PC_MCASTLEN) {
			if (match)
				(*mref)++;
			else {
				bcopy(&addr[0], (char *)mcap, ETHER_ADD_SIZE);
				*mref = 1;
				pcp->sc_mcast.mc_count += ETHER_ADD_SIZE;
			}
		} else
			error = ENOBUFS;
	} else {
		if (!match)
			error = ENXIO;
		else if (--*mref == 0) {
	     		for (; i < (PC_MCASTLEN-1);
			     i++, mcap += ETHER_ADD_SIZE, mref++) {
				bcopy((char *)mcap + ETHER_ADD_SIZE,
					      (char *)mcap, ETHER_ADD_SIZE);
				*mref = *(mref + 1);
			}
			*mref = 0;
			pcp->sc_mcast.mc_count -= ETHER_ADD_SIZE;
		}
	}

	splx(s);

	i = (int)pcp->sc_mcast.mc_count;
	if (i < 0 || i > PC_MCASTLEN*ETHER_ADD_SIZE)
		panic("pc586multicast: mc_count invalid");

	return(error);
}
#endif

/*
 * pc586hwrst:
 *
 *	This routine resets the pc586 board that corresponds to the 
 *	board number passed in.
 *
 * input	: board number to do a hardware reset
 * output	: board is reset
 *
 */
pc586hwrst(unit)
int unit;
{
	u_long data;
	volatile int *sram;
	register struct core_lan *lanp;

	lanp = pc_softc[unit].hwaddr;
	/* turn off interrupt */
	pc586intoff(unit);

	sram = (volatile int *) (&pc_softc[unit].sram[OFFSET_SCP]);
	data = (unsigned) ltor(KERNSPC, sram) | 1;
#ifdef DEBUG
	if (pc586debug)
		printf("pc586hwrst: addr=0x%x, sram=%x(%lx)\n",
		       lanp, sram, data);
#endif

	/* self test */
	sram[1] = -1;
	fcacheline(KERNSPC, (caddr_t)sram);

	pc586putport(lanp, data);
	for (data = 1000000; data > 0; data--) {
		pcacheline(KERNSPC, (caddr_t)sram);
		if (sram[1] != -1)
			break;
	}
#ifdef DEBUG
	if (pc586debug)
		printf("pc%d: self test = 0x%x:0x%x\n",
		       unit, sram[0], sram[1]);
#endif
	if (sram[1] != 0)
		return(FALSE);
	if (pc586bldcu(unit) == FALSE)
		return(FALSE);
	if (pc586diag(unit) == FALSE)
		return(FALSE);

	if (pc586config(unit) == FALSE)
		return(FALSE);
	/* 
	 * insert code for loopback test here
	 */
	pc586rustrt(unit);
	pc586inton(unit);
	return(TRUE);
}

#ifdef WATCHDOG
/*
 * pc586watch():
 *
 *	This routine is the watchdog timer routine for the pc586 chip.  If
 *	chip wedges, this routine will fire and cause a board reset and
 *	begin again.
 *
 * input	: which board is timing out
 * output	: potential board reset if wedged
 *
 */
int watch_dead = 0;
pc586watch(b_ptr)
caddr_t	b_ptr;
{
	int	opri,
		unit = *b_ptr;

	if ((pc_softc[unit].ds_if.if_flags & IFF_UP) == 0)  {
		return;
	}

	opri = splx(SPLLAN);
	if (pc_softc[unit].timer == -1) {
		timeout(pc586watch, b_ptr, pc586timo*hz);
		splx(opri);
		return;
	}
	if (--pc_softc[unit].timer != -1) {
		timeout(pc586watch, b_ptr, 1*hz);
		splx(opri);
		return;
	}

#ifdef	notdef
	printf("pc%d watch(): %dsec timeout no %d\n",
	       unit, pc586timo, ++watch_dead);
#endif
	pc586_cntrs[unit].watch++;
	if (pc586hwrst(unit) != TRUE) {
		printf("pc%d watch(): hwrst trouble.\n", unit);
		pc_softc[unit].timer = 0;
	} else {
		timeout(pc586watch, b_ptr, 1*hz);
		pc_softc[unit].timer = pc586timo;
	}
	splx(opri);
	printf("pc%d: watchdog: %d second timeout #%d\n",
	       unit, pc586timo, ++watch_dead);
}
#endif

/*
 * pc586intr:
 *
 *	This function is the interrupt handler for the pc586 ethernet
 *	board.  This routine will be called whenever either a packet
 *	is received, or a packet has successfully been transfered and
 *	the unit is ready to transmit another packet.
 *
 * input	: board number that interrupted
 * output	: either a packet is received, or a packet is transfered
 *
 */
pc586intr(unit)
int unit;
{
	register pc_softc_t *is = &pc_softc[unit];
	volatile scb_t	*scb_p;
	volatile ac_t	*cb_p;
	u_short		int_type;
#ifdef KGDB_LAN
	extern boolean_t kgdb_stop;
#endif

	if (is->seated == FALSE) { 
		printf("pc%d intr(): board not seated\n", unit);
		return(-1);
	}

#ifdef	PC596XMT_BUG
	if (((long)pc586_toutxmt[unit].tv_sec) < 0) {
		timerclear(&pc586_toutxmt[unit]);
		if ((is->ds_if.if_flags & IFF_UP) == 0)
			return(-1);

		pc586_cntrs[unit].xmtwatch++;
#ifdef IFF_OACTIVE
                is->ds_if.if_flags &= ~IFF_OACTIVE;
#else
                is->tbusy = 0;
#endif
#ifdef MACH_KERNEL
                pc586start(unit);
#else
                pc586start(&pc_softc[unit].ds_if);
#endif
	}
#endif	/* PC596XMT_BUG */

	scb_p = (volatile scb_t *)(is->sram + OFFSET_SCB);
	purgescbstat(scb_p);
	while ((int_type = (scb_p->scb_status & SCB_SW_INT)) != 0) {
		pc586ack(unit);
		if (int_type & SCB_SW_FR) {
			pc586rcv(unit);
#ifdef USELEDS
			if (inledcontrol == 0)
				ledcontrol(0, 0, LED_LANRCV);
#endif
		}
		if (int_type & SCB_SW_RNR) {
			pc586_cntrs[unit].rcv.ovw++;
#ifdef	notdef
			printf("pc%d intr(): receiver overrun! begin_fd = %x\n",
				unit, is->begin_fd);
#endif
			pc586rustrt(unit);
		}
		if (int_type & SCB_SW_CNA) {
			/*
			 * At present, we don't care about CNA's.  We
			 * believe they are a side effect of XMT.
			 */
		}
		if (int_type & SCB_SW_CX) {
#ifdef	PC596XMT_BUG
			timerclear(&pc586_toutxmt[unit]);
#endif	/* PC596XMT_BUG */
			/*
			 * At present, we only request Interrupt for
			 * XMT.
			 */
			cb_p = (volatile ac_t *)(is->sram + OFFSET_CU);
			pdcache(KERNSPC, cb_p, sizeof(*cb_p));
#ifdef	DEBUG
			if (pc586debug)
				printf("pc%d: intr type: %x status %x\n",
				       unit, int_type, cb_p->ac_status);
#endif
			if ((!(cb_p->ac_status & AC_SW_OK)) ||
			    (cb_p->ac_status & (0xfff^TC_SQE))) {
				if (cb_p->ac_status & TC_DEFER) {
					if (xmt_watch) printf("DF");
					pc586_cntrs[unit].xmt.defer++;
				} else if (cb_p->ac_status & (TC_COLLISION|0xf)) {
					if (xmt_watch) printf("%x",cb_p->ac_status & 0xf);
#ifndef	MACH_KERNEL
					if (cb_p->ac_status & TC_COLLISION)
						is->ds_if.if_collisions += 16;
					else
						is->ds_if.if_collisions +=
							cb_p->ac_status & 0xf;
#endif
				} else {
					if (xmt_watch) 
						printf("pc%d XMT: %x %x\n",
						       unit, cb_p->ac_status,
						       cb_p->ac_command);
					/*
					 * If transmitter claims to be busy,
					 * believe it.  This is probably a
					 * left-over interrupt as a result of
					 * the race condition in pc586start.
					 * DON'T restart the transmitter!
					 */
					if (cb_p->ac_status == AC_SW_B) {
						pc586_cntrs[unit].xmt.xmtsi++;
						goto xdone;
					}
					log(LOG_WARNING,
					    "pc%d: XMT: bad status %x\n",
					    unit, cb_p->ac_status);
#ifndef	MACH_KERNEL
					is->ds_if.if_oerrors++;
#endif
				}
			} else {
#ifdef USELEDS
				if (inledcontrol == 0)
					ledcontrol(0, 0, LED_LANXMT);
#endif
#ifndef	MACH_KERNEL
				is->ds_if.if_opackets++;
#endif
			}
			pc586_cntrs[unit].xmt.xmti++;
#ifdef IFF_OACTIVE
			is->ds_if.if_flags &= ~IFF_OACTIVE;
#else
			is->tbusy = 0;
#endif
#ifdef MACH_KERNEL
			pc586start(unit);
#else
			pc586start(&pc_softc[unit].ds_if);
#endif
		}
xdone:
#ifdef WATCHDOG
		is->timer = pc586timo;
		watch_dead = 0;
#endif
		purgescbstat(scb_p);
	}
#ifdef KGDB_LAN
	if (kgdb_stop)
		kgdb_break();
#endif
	return(0);
}

/*
 * pc586rcv:
 *
 *	This routine is called by the interrupt handler to initiate a
 *	packet transfer from the board to the "if" layer above this
 *	driver.  This routine checks if a buffer has been successfully
 *	received by the pc586.  If so, the routine pc586read is called
 *	to do the actual transfer of the board data (including the
 *	ethernet header) into a packet (consisting of an mbuf chain).
 *
 * input	: number of the board to check
 * output	: if a packet is available, it is "sent up"
 *
 */
pc586rcv(unit)
int	unit;
{
	volatile fd_t		*fd_p;
	register pc_softc_t	*is = &pc_softc[unit];

	for (fd_p = is->begin_fd; fd_p != (fd_t *)NULL; fd_p = is->begin_fd) {
		pcacheline(KERNSPC, (caddr_t)fd_p);
		if (fd_p->status == 0xffff) {
			if (pc586hwrst(unit) != TRUE)
				printf("pc%d rcv(): hwrst ffff trouble.\n",
					unit);
			return;
		} else if (fd_p->status & AC_SW_C) {
			is->begin_fd =
				(fd_t *)ram_to_ptr(fd_p->link_offset, unit);

			if (fd_p->status == (RFD_DONE|RFD_RSC)) {
				printf("pc%d RCV: lost packet %x\n",
				       unit, fd_p->status);
#ifndef	MACH_KERNEL
				is->ds_if.if_ierrors++;
#endif
				pc586_cntrs[unit].rcv.partial++;
			} else if (!(fd_p->status & RFD_OK)) {
				printf("pc%d RCV: %x\n", unit, fd_p->status);
#ifndef	MACH_KERNEL
				is->ds_if.if_ierrors++;
#endif
			} else if (fd_p->status & 0xfff) {
				printf("pc%d RCV: %x\n", unit, fd_p->status);
#ifndef	MACH_KERNEL
				is->ds_if.if_ierrors++;
#endif
			} else if (!pc586read(unit, fd_p))
				return;
			if (!pc586requeue(unit, fd_p)) {	/* abort on chain error */
				if (pc586hwrst(unit) != TRUE)
					printf("pc%d rcv(): hwrst trouble.\n", unit);
				return;
			}
		} else
			break;
	}
}

/*
 * pc586requeue:
 *
 *	This routine puts rbd's used in the last receive back onto the
 *	free list for the next receive.
 *
 */
pc586requeue(unit, fd_p)
int	unit;
fd_t	*fd_p;
{
	volatile rbd_t	*l_rbdp;
	rbd_t	*f_rbdp;

#ifndef	REQUEUE_DBG
	if (bad_rbd_chain(fd_p->rbd_offset, unit))
		return 0;
#endif
	f_rbdp = (rbd_t *)ram_to_ptr(fd_p->rbd_offset, unit);
	if (f_rbdp != NULL) {
		l_rbdp = f_rbdp;
		pcacheline(KERNSPC, (caddr_t)l_rbdp);
		while (!(l_rbdp->status & RBD_SW_EOF) &&
		       l_rbdp->next_rbd_offset != 0xffff)
		{
			l_rbdp->status = 0;
			fcacheline(KERNSPC, (caddr_t)l_rbdp);
		   	l_rbdp = (rbd_t *)ram_to_ptr(l_rbdp->next_rbd_offset,
						     unit);
			pcacheline(KERNSPC, (caddr_t)l_rbdp);
		}
		l_rbdp->next_rbd_offset = PC586NULL;
		l_rbdp->status = 0;
		l_rbdp->size |= AC_CW_EL;
		fcacheline(KERNSPC, (caddr_t)l_rbdp);
		pc_softc[unit].end_rbd->next_rbd_offset = 
			ptr_to_ram((char *)f_rbdp, unit);
		pc_softc[unit].end_rbd->size &= ~AC_CW_EL;
		fcacheline(KERNSPC, (caddr_t)pc_softc[unit].end_rbd);
		pc_softc[unit].end_rbd = (rbd_t *)l_rbdp;
	}

	fd_p->status = 0;
	fd_p->command = AC_CW_EL;
	fd_p->link_offset = PC586NULL;
	fd_p->rbd_offset = PC586NULL;
	fcacheline(KERNSPC, (caddr_t)fd_p);

	pc_softc[unit].end_fd->link_offset = ptr_to_ram((char *)fd_p, unit);
	pc_softc[unit].end_fd->command = 0;
	fcacheline(KERNSPC, (caddr_t)pc_softc[unit].end_fd);
	pc_softc[unit].end_fd = fd_p;

	return 1;
}

/*
 * pc586xmt:
 *
 *	This routine fills in the appropriate registers and memory
 *	locations on the PC586 board and starts the board off on
 *	the transmit.
 *
 * input	: board number of interest, and a pointer to the mbuf
 * output	: board memory and registers are set for xfer and attention
 *
 */
#ifdef	DEBUG
int xmt_debug = 0;
#endif
pc586xmt(unit, m)
int	unit;
#ifdef	MACH_KERNEL
io_req_t	m;
#else
struct	mbuf	*m;
#endif
{
	pc_softc_t			*is = &pc_softc[unit];
	register u_char			*xmtdata_p = (u_char *)(is->sram + OFFSET_TBUF);
	register u_short		*xmtshort_p;
#ifdef	MACH_KERNEL
	register struct ether_header	*eh_p = (struct ether_header *)m->io_data;
#else
	struct	mbuf			*tm_p = m;
	register struct ether_header	*eh_p = mtod(m, struct ether_header *);
	u_char				*mb_p = mtod(m, u_char *) + sizeof(struct ether_header);
	u_short				count = m->m_len - sizeof(struct ether_header);
#endif
	volatile scb_t			*scb_p = (volatile scb_t *)(is->sram + OFFSET_SCB);
	volatile ac_t			*cb_p = (volatile ac_t *)(is->sram + OFFSET_CU);
	tbd_t				*tbd_p = (tbd_t *)(is->sram + OFFSET_TBD);
	u_short				len, clen = 0;

	cb_p->ac_status = 0;
	cb_p->ac_command = (AC_CW_EL|AC_TRANSMIT|AC_CW_I);
	cb_p->ac_link_offset = PC586NULL;
	cb_p->cmd.transmit.tbd_offset = OFFSET_TBD;

	bcopy16(eh_p->ether_dhost, cb_p->cmd.transmit.dest_addr, ETHER_ADD_SIZE);
	cb_p->cmd.transmit.length = (u_short)(eh_p->ether_type);
	fdcache(KERNSPC, (caddr_t)cb_p, sizeof *cb_p);

#ifndef	MACH_KERNEL
#ifdef	DEBUG
	if (xmt_debug)
		printf("pc%d: XMT    mbuf: L%u @%x ", unit, count, mb_p);
#endif
#endif
	tbd_p->act_count = 0;
	tbd_p->buffer_addr = (u_short)((u_int)xmtdata_p);
	tbd_p->buffer_base = (int)xmtdata_p >> 16;
#ifdef DEBUG
	if (pc586debug)
		printf("pc%d: create tbuf #1 base=%x addr=%x\n",
		       unit, tbd_p->buffer_base, tbd_p->buffer_addr);
#endif
#ifdef	MACH_KERNEL
	{ int Rlen, Llen;
	    clen = m->io_count - sizeof(struct ether_header);
	    Llen = clen & 1;
	    Rlen = ((int)(m->io_data + sizeof(struct ether_header))) & 1;

	    bcopy16(m->io_data + sizeof(struct ether_header) - Rlen,
		    xmtdata_p,
		    clen + (Rlen + Llen) );
	    xmtdata_p += clen + Llen;
	    tbd_p->act_count = clen;
	    tbd_p->buffer_addr += Rlen;
	}
#else
	do {
		if (count) {
			if (clen + count > ETHERMTU)
				break;
			len = count;
			bcopy16(mb_p, xmtdata_p, count);
			clen += count;
			tbd_p->act_count += count;
			xmtdata_p += len;
			if ((tm_p = tm_p->m_next) == (struct mbuf *)0)
				break;
		} else if ((tm_p = tm_p->m_next) == (struct mbuf *)0)
			break;
		count = tm_p->m_len;
		mb_p = mtod(tm_p, u_char *);
#ifdef	DEBUG
		if (xmt_debug)
			printf("mbuf+ L%u @%x ", count, mb_p);
#endif
	} while (1);
	if (clen & 1)
		xmtdata_p += 1;
#endif	/* MACH_KERNEL */
#ifdef	DEBUG
	if (xmt_debug)
		printf("CLEN = %u\n", clen);
#endif
	if (clen < ETHERMIN) {
		tbd_p->act_count += ETHERMIN - clen;
		for (xmtshort_p = (u_short *)xmtdata_p;
		     clen < ETHERMIN;
		     clen += 2) *xmtshort_p++ = 0;
		xmtdata_p = (u_char *)xmtshort_p;
	}
	tbd_p->act_count |= TBD_SW_EOF;
	tbd_p->next_tbd_offset = PC586NULL;
	fcacheline(KERNSPC, (caddr_t)tbd_p);
	/* XXX take into account padding of odd length buffers */
	len = xmtdata_p - (u_char *)(is->sram + OFFSET_TBUF);
	if (len < clen)
		panic("pc586xmt: screwed the pooch");
	fdcache(KERNSPC, (caddr_t)(is->sram + OFFSET_TBUF), len);
#ifdef	IF_CNTRS
	clen += sizeof (struct ether_header) + 4 /* crc */;
	pc586_eout[log_2(clen)]++;
	if (clen < 128)  pc586_lout[clen>>3]++;
#endif
#ifdef	DEBUG
	if (xmt_debug) {
		pc586tbd(unit);
		printf("\n");
	}
#endif

	/*
	 * This just wedged on me (4/17/92).  We cant have this STUPID
	 * board screwing up our fine machine, can we?!!
	 */
	{
		int i = 1000000;

		do
			purgescbcmd(scb_p);
		while (scb_p->scb_command && i-- > 0);

		if (i == 0) {
			printf("pc%d: I have wedged myself in pc586xmt().\n",
			       unit);
			if (pc586hwrst(unit) != TRUE) {
				printf("pc%d: hardware reset failed.\n", unit);
#ifdef WATCHDOG
				pc_softc[unit].timer = 0;
#endif
			}
#ifdef	MACH_KERNEL
			iodone(m);
#else
			m_freem(m);
#endif
			return;
		}
	}
	scb_p->scb_command = SCB_CU_STRT;
	flushscbcmd(scb_p);
	pc586chatt(unit);
#ifdef	PC596XMT_BUG
	{
		register long tout;
		register struct timeval *tp = &pc586_toutxmt[unit];

		*tp = time;
		for (tout = tp->tv_usec + pc586_waitxmt;
		     tout >= 1000000; tout -= 1000000)
			tp->tv_sec += 1;
		tp->tv_usec = tout;
	}
#endif	/* PC596XMT_BUG */

#ifdef	MACH_KERNEL
	iodone(m);
#else
	m_freem(m);
#endif
}

/*
 * pc586bldcu:
 *
 *	This function builds up the command unit structures.  It inits
 *	the scp, iscp, scb, cb, tbd, and tbuf.
 *
 */
pc586bldcu(unit)
{
	char		*sram = pc_softc[unit].sram;
	scp_t		*scp_p = (scp_t *)(sram + OFFSET_SCP);
	volatile iscp_t	*iscp_p = (volatile iscp_t *)(sram + OFFSET_ISCP);
	volatile scb_t	*scb_p = (volatile scb_t *)(sram + OFFSET_SCB);
	volatile ac_t	*cb_p = (volatile ac_t *)(sram + OFFSET_CU);
	tbd_t		*tbd_p = (tbd_t *)(sram + OFFSET_TBD);
	int		i;

	scp_p->scp_sysbus = 0x68;
	i = (int)ltor(KERNSPC, iscp_p);
	scp_p->scp_iscp_lo = i;
	scp_p->scp_iscp_hi = i >> 16;
	scp_p->scp_unused1 = 0;
	scp_p->scp_unused2 = 0;

	iscp_p->iscp_busy = 1;
	iscp_p->iscp_scb_offset = OFFSET_SCB;
	i = (int)ltor(KERNSPC, sram);
	iscp_p->iscp_scb_lo = i;
	iscp_p->iscp_scb_hi = i >> 16;

	purgescball(scb_p);
	pc586_cntrs[unit].rcv.crc += scb_p->scb_crcerrs;
	pc586_cntrs[unit].rcv.frame += scb_p->scb_alnerrs;
	pc586_cntrs[unit].rcv.rscerrs += scb_p->scb_rscerrs;
	pc586_cntrs[unit].rcv.ovrnerrs += scb_p->scb_ovrnerrs;
	scb_p->scb_status = 0;
	scb_p->scb_command = 0;
	scb_p->scb_cbl_offset = OFFSET_CU;
	scb_p->scb_rfa_offset = OFFSET_RU;
	scb_p->scb_crcerrs = 0;
	scb_p->scb_alnerrs = 0;
	scb_p->scb_rscerrs = 0;
	scb_p->scb_ovrnerrs = 0;

	cb_p->ac_status = 0;
	cb_p->ac_command = (AC_CW_EL|AC_NOP);
	cb_p->ac_link_offset = OFFSET_CU;

	tbd_p->act_count = 0;
	tbd_p->next_tbd_offset = PC586NULL;
	tbd_p->buffer_addr = 0;
	tbd_p->buffer_base = 0;

	i = ltor(KERNSPC, scp_p) | 0x2;
	fdcache(KERNSPC, pc_softc[unit].sram, LAN_BUFF_SIZE);
	pc586putport(pc_softc[unit].hwaddr, i);
	pc586chatt(unit);
	for (i = 1000000; iscp_p->iscp_busy && i > 0; i--)
		pcacheline(KERNSPC, (caddr_t)iscp_p);
#ifdef DEBUG
	if (pc586debug)
		printf("pc%d: iscp status = %x\n", unit, iscp_p->iscp_busy);
#endif
	if (!i) {
		printf("pc%d bldcu(): iscp_busy timeout.\n", unit);
		return(FALSE);
	}
	for (i = STATUS_TRIES; i > 0; i--) {
		if (scb_p->scb_status == (SCB_SW_CX|SCB_SW_CNA)) 
			break;
		purgescbstat(scb_p);
	}
#ifdef DEBUG
	if (pc586debug)
		printf("pc%d: scb status = %x:%x %d\n",
		       unit, scb_p->scb_status, scb_p->scb_command, i);
#endif
	if (!i) {
		printf("pc%d bldcu(): not ready after reset.\n", unit);
		return(FALSE);
	}
	pc586ack(unit);
	return(TRUE);
}

/*
 * pc586bldru:
 *
 *	This function builds the linear linked lists of fd's and
 *	rbd's.  Based on page 4-32 of 1986 Intel microcom handbook.
 *
 */
char *
pc586bldru(unit)
int unit;
{
	fd_t	*fd_p = (fd_t *)(pc_softc[unit].sram + OFFSET_RU);
	ru_t	*rbd_p = (ru_t *)(pc_softc[unit].sram + OFFSET_RBD);
	int 	i;

	pc_softc[unit].begin_fd = fd_p;
	for(i = 0; i < N_FD; i++, fd_p++) {
		fd_p->status = 0;
		fd_p->command	= 0;
		fd_p->link_offset = ptr_to_ram((char *)(fd_p + 1), unit);
		fd_p->rbd_offset = PC586NULL;
	}
	pc_softc[unit].end_fd = --fd_p;
	fd_p->link_offset = PC586NULL;
	fd_p->command = AC_CW_EL;
	fd_p = (fd_t *)(pc_softc[unit].sram + OFFSET_RU);

	fd_p->rbd_offset = ptr_to_ram((char *)rbd_p, unit);
	for(i = 0; i < N_RBD; i++, rbd_p++) {
		rbd_p->r.status = 0;
		rbd_p->r.buffer_addr = (u_short)((u_int)rbd_p->rbuffer);
		rbd_p->r.buffer_base = (int)rbd_p->rbuffer >> 16;
#ifdef DEBUG
		if (pc586debug)
			printf("pc%d: create rbuf #4 base=%x addr=%x\n",
			       unit, rbd_p->r.buffer_base,
			       rbd_p->r.buffer_addr);
#endif
		rbd_p->r.size = RCVBUFSIZE;
		if (i != N_RBD-1) {
			rbd_p->r.next_rbd_offset =
				ptr_to_ram((char *)(rbd_p + 1), unit);
		} else {
			rbd_p->r.next_rbd_offset = PC586NULL;
			rbd_p->r.size |= AC_CW_EL;
			pc_softc[unit].end_rbd = (rbd_t *)rbd_p;
		}
	}
	return (char *)pc_softc[unit].begin_fd;
}

/*
 * pc586rustrt:
 *
 *	This routine starts the receive unit running.  First checks if the
 *	board is actually ready, then the board is instructed to receive
 *	packets again.
 *
 */
pc586rustrt(unit)
int unit;
{
	volatile scb_t	*scb_p;
	char		*strt;

	scb_p = (volatile scb_t *)(pc_softc[unit].sram + OFFSET_SCB);
	purgescball(scb_p);
	if ((scb_p->scb_status & SCB_RUS_READY) == SCB_RUS_READY)
		return;

	strt = pc586bldru(unit);
	scb_p->scb_command = SCB_RU_STRT;
	scb_p->scb_rfa_offset = ptr_to_ram(strt, unit);
	fdcache(KERNSPC, pc_softc[unit].sram, LAN_BUFF_SIZE);
	pc586chatt(unit);
}

/*
 * pc586diag:
 *
 *	This routine does a 586 op-code number 7, and obtains the
 *	diagnose status for the pc586.
 *
 */
pc586diag(unit)
int unit;
{
	volatile scb_t	*scb_p;
	volatile ac_t	*cb_p;
	int		i;

	scb_p = (volatile scb_t *)(pc_softc[unit].sram + OFFSET_SCB);
	purgescball(scb_p);
	if (scb_p->scb_status & SCB_SW_INT) {
		printf("pc%d diag(): bad initial state %u\n",
			unit, scb_p->scb_status);
		pc586ack(unit);
	}
	cb_p = (volatile ac_t *)(pc_softc[unit].sram + OFFSET_CU);
	cb_p->ac_status	= 0;
	cb_p->ac_command = (AC_DIAGNOSE|AC_CW_EL);
	fdcache(KERNSPC, cb_p, sizeof *cb_p);
	scb_p->scb_command = SCB_CU_STRT;
	flushscbcmd(scb_p);
	pc586chatt(unit);
	for (i = 0; i < 0xffff; i++) {
		if ((cb_p->ac_status & AC_SW_C))
			break;
		pdcache(KERNSPC, cb_p, sizeof *cb_p);
	}
	if (i == 0xffff || !(cb_p->ac_status & AC_SW_OK)) {
		printf("pc%d: diag failed; status = %x, command = %x\n",
		       unit, cb_p->ac_status, cb_p->ac_command);
		return(FALSE);
	}
	purgescbstat(scb_p);
	if ( (scb_p->scb_status & SCB_SW_INT) && (scb_p->scb_status != SCB_SW_CNA) )  {
		printf("pc%d diag(): bad final state %x\n",
			unit, scb_p->scb_status);
		pc586ack(unit);
	}
	return(TRUE);
}

/*
 * pc586config:
 *
 *	This routine does a standard config of the pc586 board.
 *
 */
pc586config(unit)
int unit;
{
	volatile scb_t	*scb_p;
	volatile ac_t	*cb_p;
	int 		i;

#ifdef IFF_OACTIVE
	pc_softc[unit].ds_if.if_flags &= ~IFF_OACTIVE;
#endif

	scb_p = (volatile scb_t *)(pc_softc[unit].sram + OFFSET_SCB);
	purgescball(scb_p);
	if ((scb_p->scb_status != SCB_SW_CNA) &&
	    (scb_p->scb_status & SCB_SW_INT) ) {
		printf("pc%d config(): unexpected initial state %x\n",
			unit, scb_p->scb_status);
	}
	pc586ack(unit);

	cb_p = (volatile ac_t *)(pc_softc[unit].sram + OFFSET_CU);
	cb_p->ac_status	= 0;
	cb_p->ac_command = (AC_CONFIGURE|AC_CW_EL);
	cb_p->ac_link_offset = PC586NULL;

	/*
	 * below is the default board configuration from p2-28 from 586 book
	 */
#if 0
	cb_p->cmd.configure.fifolim_bytecnt 	= 0x080c;
	cb_p->cmd.configure.addrlen_mode  	= 0x2600;
	cb_p->cmd.configure.linprio_interframe	= 0x6000;
	cb_p->cmd.configure.slot_time      	= 0xf200;
	if (pc_softc[unit].ds_if.if_flags & IFF_PROMISC)
		cb_p->cmd.configure.hardware	= 0x0001;
	else
		cb_p->cmd.configure.hardware	= 0x0000;
	cb_p->cmd.configure.min_frame_len   	= 0x0040;
#else
	/*
	 * The above code was obviously written for a different
	 * byte order.  I added the `IFF_PROMISC' code just for
	 * completeness.  (jef)
	 */
	cb_p->cmd.configure.fifolim_bytecnt 	= 0x0c08;
	cb_p->cmd.configure.addrlen_mode  	= 0x0026;
	cb_p->cmd.configure.linprio_interframe	= 0x0060;
	cb_p->cmd.configure.slot_time      	= 0x00f2;
	if (pc_softc[unit].ds_if.if_flags & IFF_PROMISC)
		cb_p->cmd.configure.hardware	= 0x0100;
	else
		cb_p->cmd.configure.hardware	= 0x0000;
	cb_p->cmd.configure.min_frame_len   	= 0x4000;
#endif
	fdcache(KERNSPC, cb_p, sizeof(*cb_p));

	scb_p->scb_command = SCB_CU_STRT;
	flushscbcmd(scb_p);
	pc586chatt(unit);

	for (i = 0; i < 0xffff; i++) {
		if ((cb_p->ac_status & AC_SW_C))
			break;
		pdcache(KERNSPC, cb_p, sizeof *cb_p);
	}
	if (i == 0xffff || !(cb_p->ac_status & AC_SW_OK)) {
		printf("pc%d: config-configure failed; status = %x\n",
			unit, cb_p->ac_status);
		return(FALSE);
	}
	purgescbstat(scb_p);
#if 0
	if (scb_p->scb_status & SCB_SW_INT) {
		printf("pc%d configure(): bad configure state %x\n",
			unit, scb_p->scb_status);
		pc586ack(unit);
	}
#else
	/*
	 * I have no idea what the above test was supposed to be doing.
	 * As far as I can tell, the chip is *supposed* to set SCB_SW_CNA
	 * upon completion of a command.  (jef)
	 */
	if ((scb_p->scb_status & SCB_SW_INT) != SCB_SW_CNA) {
		printf("pc%d configure(): unexpected configure state %x\n",
			unit, scb_p->scb_status);
	}
#endif

#ifdef MULTICAST
	{
		register multicast_t *mc_p = (multicast_t *)cb_p;
		register multicast_t *mp = &pc_softc[unit].sc_mcast;

		mc_p->ac_status	= 0;
		mc_p->ac_command = (AC_MCSETUP|AC_CW_EL);
		mc_p->ac_link_offset = PC586NULL;

		mc_p->mc_count = mp->mc_count;
		for (i = 0; i < mp->mc_count; i++)
			mc_p->mc_addr[i] = mp->mc_addr[i];

		fdcache(KERNSPC, mc_p, sizeof(*mc_p));

		scb_p->scb_command = SCB_CU_STRT;
		flushscbcmd(scb_p);
		pc586chatt(unit);

		for (i = 0; i < 0xffff; i++) {
			if ((cb_p->ac_status & AC_SW_C))
				break;
			pdcache(KERNSPC, cb_p, sizeof *cb_p);
		}
		if (i == 0xffff || !(cb_p->ac_status & AC_SW_OK)) {
			printf("pc%d: mcast-configure failed; status = %x\n",
				unit, cb_p->ac_status);
			return(FALSE);
		}
		purgescbstat(scb_p);

		if ((scb_p->scb_status & SCB_SW_INT) != SCB_SW_CNA) {
			printf("pc%d configure(): unexpected mcast state %x\n",
				unit, scb_p->scb_status);
		}
	}
#endif

	cb_p->ac_status = 0;
	cb_p->ac_command = (AC_IASETUP|AC_CW_EL);
	cb_p->ac_link_offset = PC586NULL;

	bcopy16(pc_softc[unit].ds_addr, cb_p->cmd.iasetup, ETHER_ADD_SIZE);
	fdcache(KERNSPC, cb_p, sizeof(*cb_p));

	scb_p->scb_command = SCB_CU_STRT;
	flushscbcmd(scb_p);
	pc586chatt(unit);

	for (i = 0; i < 0xffff; i++) {
		if ((cb_p->ac_status & AC_SW_C))
			break;
		pdcache(KERNSPC, cb_p, sizeof *cb_p);
	}
	if (i == 0xffff || !(cb_p->ac_status & AC_SW_OK)) {
		printf("pc%d: config-address failed; status = %x\n",
			unit, cb_p->ac_status);
		return(FALSE);
	}

	purgescbstat(scb_p);
	if ((scb_p->scb_status & SCB_SW_INT) != SCB_SW_CNA) {
		printf("pc%d configure(): unexpected final state %x\n",
			unit, scb_p->scb_status);
	}
	pc586ack(unit);

	return(TRUE);
}

/*
 * pc586ack:
 */
pc586ack(unit)
{
	volatile scb_t	*scb_p;
	int i;

	scb_p = (volatile scb_t *)(pc_softc[unit].sram + OFFSET_SCB);
	purgescball(scb_p);
	scb_p->scb_command = (scb_p->scb_status & SCB_SW_INT);
	flushscbcmd(scb_p);
	if (!scb_p->scb_command)
		return;
	pc586chatt(unit);
	i = 1000000;
	do
		purgescbcmd(scb_p);
	while (scb_p->scb_command && i-- > 0);
	if (!i)
		printf("pc%d pc586ack(): board not accepting command.\n", unit);
}

char *
ram_to_ptr(offset, unit)
int unit;
u_short	offset;
{
	if (offset == PC586NULL)
		return(NULL);
	if (offset > 0x3fff) {
		printf("pc%d: ram_to_ptr(%x, %d)\n", unit, offset, unit);
		panic("range");
		return(NULL);
	}
	return(pc_softc[unit].sram + offset);
}

#ifndef	REQUEUE_DBG
bad_rbd_chain(offset, unit)
	u_short offset;
	int unit;
{
	volatile rbd_t	*rbdp;
	char	*sram = pc_softc[unit].sram;

	for (;;) {
		if (offset == PC586NULL)
			return 0;
		if (offset > 0x3fff) {
			printf("pc%d: bad_rbd_chain offset = %x\n",
				unit, offset);
			pc586_cntrs[unit].rcv.bad_chain++;
			return 1;
		}

		rbdp = (volatile rbd_t *)(sram + offset);
		pcacheline(KERNSPC, (caddr_t)rbdp);
		offset = rbdp->next_rbd_offset;
	}
}
#endif

u_short
ptr_to_ram(k_va, unit)
char	*k_va;
int unit;
{
	return((u_short)(k_va - pc_softc[unit].sram));
}

#ifdef DEBUG
pc586scb(unit)
{
	volatile scb_t	*scb;
	u_short		 i;

	scb = (volatile scb_t *)(pc_softc[unit].sram + OFFSET_SCB);
	purgescball(scb);
	i = scb->scb_status;
	printf("stat: stat %x, cus %x, rus %x //",
		(i&0xf000)>>12, (i&0x0700)>>8, (i&0x0070)>>4);
	i = scb->scb_command;
	printf(" cmd: ack %x, cuc %x, ruc %x\n",
		(i&0xf000)>>12, (i&0x0700)>>8, (i&0x0070)>>4);

	printf("crc %u[%u], align %u[%u], rsc %u[%u], ovr %u[%u]\n",
		scb->scb_crcerrs, pc586_cntrs[unit].rcv.crc,
		scb->scb_alnerrs, pc586_cntrs[unit].rcv.frame,
		scb->scb_rscerrs, pc586_cntrs[unit].rcv.rscerrs,
		scb->scb_ovrnerrs, pc586_cntrs[unit].rcv.ovrnerrs);

	printf("cbl %u, rfa %u //", scb->scb_cbl_offset, scb->scb_rfa_offset);
}

pc586tbd(unit)
{
	pc_softc_t	*is = &pc_softc[unit];
	volatile tbd_t	*tbd_p = (tbd_t *)(is->sram + OFFSET_TBD);
	int 		i = 0;
	int		sum = 0;

	do {
		pcacheline(KERNSPC, (caddr_t)tbd_p);
		sum += (tbd_p->act_count & ~TBD_SW_EOF);
		printf("pc%d: %d: addr=%x count=%u(%d) next=%x base=%x\n",
		       unit, i++, tbd_p->buffer_addr,
		       (tbd_p->act_count & ~TBD_SW_EOF), sum,
		       tbd_p->next_tbd_offset,
		       tbd_p->buffer_base);
		if (tbd_p->act_count & TBD_SW_EOF)
			break;
		tbd_p = (tbd_t *)(is->sram + tbd_p->next_tbd_offset);
	} while (1);
	dumpbytes((u_char *)(is->sram + OFFSET_TBUF), sum);
}

dumpehdr(eh)
	register struct ether_header *eh;
{
	printf("type=0x%x", eh->ether_type);
	printf(" daddr=%s", ether_sprintf(eh->ether_dhost));
	printf(" saddr=%s", ether_sprintf(eh->ether_shost));
}

#ifndef	MACH_KERNEL
dumpmbufchain(str, eh, m)
	char *str;
	struct ether_header *eh;
	struct mbuf *m;
{
	register struct mbuf *m1;

	printf("%s: ", str);
	dumpehdr(eh);
	printf("\nmbuf chain @ 0x%x\n", m);
	for (m1 = m; m1 != NULL; m1 = m1->m_next) {
		printf("next:0x%x npkt:0x%x len:%d data:%d type:0x%x",
			m1->m_next, m1->m_nextpkt, m1->m_len, m1->m_data,
			m1->m_type);
		if (m1->m_flags & M_PKTHDR)
			printf(" PKTHDR (ln:%d if:0x%x)", m1->m_pkthdr.len,
				m1->m_pkthdr.rcvif);
		if (m1->m_flags & M_EXT)
			printf(" EXT (bf:0x%x f:0x%x sz:%u)",
				m1->m_ext.ext_buf, m1->m_ext.ext_free,
				m1->m_ext.ext_size);
		printf("\n");
		dumpbytes(mtod(m1, u_char *), m1->m_len);
	}
}
#endif

dumpbytes(dp, len)
	register u_char *dp;
	register int len;
{
	static char digits[] = "0123456789ABCDEF";
	char cbuf[3];
	register int i;

	if (pc586debug < 3)
		return;

	cbuf[2] = '\0';
	for (i = 0; i < len; i++) {
		cbuf[0] = digits[*dp >> 4];
		cbuf[1] = digits[*dp++ & 0xF];
		printf("%s%c", cbuf, (i%32)==31 ? '\n' : ' ');
	}
	if (i%32)
		printf("\n");
}
#endif	/* DEBUG */
#endif	/* NLAN > 0 */
