/* 
 * Copyright (c) 1991,1992,1994 The University of Utah and
 * the Computer Systems Laboratory at the University of Utah (CSL).
 * All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software is hereby
 * granted provided that (1) source code retains these copyright, permission,
 * and disclaimer notices, and (2) redistributions including binaries
 * reproduce the notices in supporting documentation, and (3) all advertising
 * materials mentioning features or use of this software display the following
 * acknowledgement: ``This product includes software developed by the
 * Computer Systems Laboratory at the University of Utah.''
 *
 * THE UNIVERSITY OF UTAH AND CSL ALLOW FREE USE OF THIS SOFTWARE IN ITS "AS
 * IS" CONDITION.  THE UNIVERSITY OF UTAH AND CSL DISCLAIM ANY LIABILITY OF
 * ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * CSL requests users of this software to return to csl-dist@cs.utah.edu any
 * improvements that they make and grant CSL redistribution rights.
 *
 *	Utah $Hdr: medreg.h 1.2 94/12/14$
 */

#ifndef _MEDUSA_
#define _MEDUSA_

#define MED_MAC_LEN 	6    		/* size of a MAC address */
#define CRC_SIZE	6		/* Trailing CRC */
#define CRC_SIZE_ALGN   8		/* But lets be aligned. */
#define MED_STI_ID	0x280B31AF	/* Medusa STI ROM graphics ID */
#define MED_MAC_OFFSET	0x3ffe3  	/* offset to MAC address in STI ROM */

#define RESET_MISC        med_card_base + 0x01000000
#define RESET_FIFO        med_card_base + 0x01000004
#define RESET_FDDI        med_card_base + 0x01000008
#define TEST_FCG          med_card_base + 0x0100000C
#define TEST_ELM          med_card_base + 0x01000010
#define TEST_MAC          med_card_base + 0x01000014
#define INTERRUPT_MASK    med_card_base + 0x01000018
#define INTERRUPT_EVENT   med_card_base + 0x0100001C
#define MISC_OUTPUT       med_card_base + 0x01000024
#define MISC_INPUT        med_card_base + 0x01000030
#define STATUS_SYNCH_FIFO med_card_base + 0x01000028
#define STATUS_BUFF_FIFO  med_card_base + 0x0100002C
#define CHECKSUM_UNIT     med_card_base + 0x01400000
#define TX_READY_FIFO     med_card_base + 0x02000000
#define TX_FREE_FIFO      med_card_base + 0x02400000
#define RX_READY_FIFO     med_card_base + 0x02800000
#define RX_FREE_FIFO      med_card_base + 0x02C00000
#define CARD_MEMORY_BASE  med_card_base + 0x03000000
#define CARD_CKSUM_BASE   med_card_base + 0x03800000

/*
 * Neither of the two constants below may exceed 64, which is the depth of
 * the on-card buffer FIFOs.
 */
#define MED_NUM_BUFS		128     /* Total buffer count */
#define MED_NUM_TX_BUFS 	64	/* No. buffers allocated to transmit */
#define MED_NUM_RX_BUFS 	64	/* No. buffers allocated to receive */
#define MED_BUFSIZE 		0x2000	/* Size in bytes of each buffer */

#define MED_PKT_HDR 		0x307800d0 /* Magic at start of FDDI packet */

/*
 * Initial Board Register Values
 */
#define INIT_RESET_MISC		0xC0
#define INIT_RESET_FIFO		0xFC
#define INIT_RESET_FDDI         0xC0
#define INIT_TEST_FCG           0xC0
#define INIT_TEST_ELM           0x00
#define INIT_TEST_MAC           0x00
#define INIT_INTERRUPT_MASK     0x00
#define INIT_INTERRUPT_EVENT    0x00
#define INIT_MISC_OUTPUT        0xC0
#define INIT_MISC_INPUT         0xe0
#define INIT_STATUS_SYNCH_FIFO  0xA0
#define INIT_STATUS_BUFF_FIFO   0xAA
#define INIT_CHECKSUM_UNIT      0x00

/*
 * MAC Control and status registers
 */
#define MAC_CNTRL_A             med_card_base + 0x00C00000
#define MAC_CNTRL_B             med_card_base + 0x00C00004
#define MAC_INTR_MASK_A         med_card_base + 0x00C00008
#define MAC_INTR_MASK_B         med_card_base + 0x00C0000C
#define MAC_MSA                 med_card_base + 0x00C00040
#define MAC_MLA_A               med_card_base + 0x00C00044
#define MAC_MLA_B               med_card_base + 0x00C00048
#define MAC_MLA_C               med_card_base + 0x00C0004C
#define MAC_T_REQ               med_card_base + 0x00C00050
#define MAC_T_MAX               med_card_base + 0x00C00054
#define MAC_RX_STATUS           med_card_base + 0x00C00090
#define MAC_TX_STATUS           med_card_base + 0x00C00094
#define MAC_INFO_REG_A          med_card_base + 0x00C000A0
#define MAC_INFO_REG_B          med_card_base + 0x00C000A4

/*
 * Configuration parameters
 */
#define MAC_T_REQ_CONFIG        0xFF3C
#define MAC_T_MAX_CONFIG        0x80E0
#define MAC_CNTRL_A_CONFIG      0x8000
#define MAC_CNTRL_B_CONFIG      0x0000
#define MAC_ON                  0x8000

/*
 * TX Status masks
 */
#define MAC_RING_OPERATIONAL    0x0800

/*
 * ELM Control and status registers
 */

#define ELM_CNTRL_A             med_card_base + 0x00800000
#define ELM_CNTRL_B             med_card_base + 0x00800004
#define ELM_INTR_MASK           med_card_base + 0x00800008
#define ELM_XMIT_VECTOR         med_card_base + 0x0080000C
#define ELM_VECTOR_LENGTH       med_card_base + 0x00800010
#define ELM_LE_THRESHOLD        med_card_base + 0x00800014
#define ELM_A_MAX               med_card_base + 0x00800018
#define ELM_LS_MAX              med_card_base + 0x0080001C
#define ELM_TB_MIN              med_card_base + 0x00800020
#define ELM_T_OUT               med_card_base + 0x00800024
#define ELM_LC_SHORT            med_card_base + 0x0080002C
#define ELM_T_SCRUB             med_card_base + 0x00800030
#define ELM_NS_MAX              med_card_base + 0x00800034
#define ELM_TPC_LOAD_VALUE      med_card_base + 0x00800038
#define ELM_TNE_LOAD_VALUE      med_card_base + 0x0080003C
#define ELM_STATUS_A            med_card_base + 0x00800040
#define ELM_STATUS_B            med_card_base + 0x00800044
#define ELM_TPC                 med_card_base + 0x00800048
#define ELM_TNE                 med_card_base + 0x0080004C
#define ELM_CLK_DIV             med_card_base + 0x00800050
#define ELM_BIST_SIGNATURE      med_card_base + 0x00800054
#define ELM_RCV_VECTOR          med_card_base + 0x00800058
#define ELM_INTR_EVENT          med_card_base + 0x0080005C
#define ELM_VIOL_SYN_CTR        med_card_base + 0x00800060
#define ELM_MIN_IDLE_CTR        med_card_base + 0x00800064
#define ELM_LINK_ERR_CTR        med_card_base + 0x00800068

/*
 * Defines for ELM_STATUS_B
 */
#define ELM_PCM_STATE           0x0780
#define ELM_PC_MAINT            0x0004

/*
 * Configuration parameters for the ELM registers,
 * taken from the ELM data sheet, page 19.
 */
#define ELM_INTR_MASK_CONFIG      0x0000
#define ELM_LE_THRESHOLD_CONFIG   0x0001
#define ELM_A_MAX_CONFIG          0xFFF6
#define ELM_LS_MAX_CONFIG         0xFFFF
#define ELM_TB_MIN_CONFIG         0xFC2F
#define ELM_T_OUT_CONFIG          0xECED
#define ELM_LC_SHORT_CONFIG       0xF676
#define ELM_T_SCRUB_CONFIG        0xFF6D
#define ELM_NS_MAX_CONFIG         0xE796
#define ELM_CNTRL_A_CONFIG        0x0000

/*
 * PCM States
 */
#define ELM_PCM_OFF             0x0000

/*
 * Defines for PCM Control
 */
#define ELM_PCM_CNTRL_STOP      0x0003

/*
 * Venom Stuff.
 */
#define VENOM_CONTROL		med_card_base & 0x0C000000
#define VENOM_SUP_CONTROL	0xfffbd080

#define out_word(addr, datum) \
	*((volatile unsigned long *)(addr)) = (unsigned long)(datum)

#define in_word(addr) \
	(*(volatile unsigned long *)(addr))

#endif /* _MEDUSA_ */
