/* 
 * Copyright (c) 1990,1991,1992,1993,1994 The University of Utah and
 * the Computer Systems Laboratory (CSL).  All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software is hereby
 * granted provided that (1) source code retains these copyright, permission,
 * and disclaimer notices, and (2) redistributions including binaries
 * reproduce the notices in supporting documentation, and (3) all advertising
 * materials mentioning features or use of this software display the following
 * acknowledgement: ``This product includes software developed by the
 * Computer Systems Laboratory at the University of Utah.''
 *
 * THE UNIVERSITY OF UTAH AND CSL ALLOW FREE USE OF THIS SOFTWARE IN ITS "AS
 * IS" CONDITION.  THE UNIVERSITY OF UTAH AND CSL DISCLAIM ANY LIABILITY OF
 * ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * CSL requests users of this software to return to csl-dist@cs.utah.edu any
 * improvements that they make and grant CSL redistribution rights.
 *
 * 	Utah $Hdr: act.h 1.5 95/01/20$
 */

/*
 *      This file defines machine specific, thread related structures,
 *      variables and macros.
 */

#ifndef	_PARISC_ACT_H_
#define _PARISC_ACT_H_

#ifndef ASSEMBLER

#ifndef __dead /* XXX */
#define __dead
#endif

/*
 * saved state structure
 *
 * This structure is used in the PCB, below,
 * to hold saved state (either user or kernel) for activations.
 *
 * It is also used privately by interrupt code
 * as the base of the interrupt stack frame.
 * This use is completely contained in interrupt.s and is invisible elsewhere.
 *
 * Finally, it is used in the 'tmp_saved_state' variable,
 * declared in parisc_init.c and used only in locore.s.
 * It is used as a general-purpose storage bag by TLB and PDC routines.
 *
 * NB: If you change this structure you must also change the KGDB stub as well.
 *
 * In order to avoid the mess of converting between HP-UX/BSD and Mach for 
 * debuggers we'll just adopt the same HP-UX layout.
 */
struct parisc_saved_state {
	unsigned flags;
	unsigned r1;
	unsigned r2;
	unsigned r3;
	unsigned r4;
	unsigned r5;
	unsigned r6;
	unsigned r7;
	unsigned r8;
	unsigned r9;
	unsigned r10;
	unsigned r11;
	unsigned r12;
	unsigned r13;
	unsigned r14;
	unsigned r15;
	unsigned r16;
	unsigned r17;
	unsigned r18;
	unsigned r19;
	unsigned r20;
	unsigned r21;
	unsigned r22;
	unsigned r23;
	unsigned r24;
	unsigned r25;
	unsigned r26;
	unsigned r27;
	unsigned r28;
	unsigned r29;
	unsigned r30;
	unsigned r31;
	unsigned cr11;			/* sar */
	unsigned iioq_head;
	unsigned iisq_head;
	unsigned iioq_tail;
	unsigned iisq_tail;
	unsigned cr15;			/* eiem */
	unsigned cr19;			/* iir */
	unsigned cr20;			/* isr */
	unsigned cr21;			/* ior */
	unsigned cr22;			/* ipsw */
	unsigned bsd_goto;
	unsigned sr4;
	unsigned sr0;
	unsigned sr1;
	unsigned sr2;
	unsigned sr3;
	unsigned sr5;
	unsigned sr6;
	unsigned sr7;			/* not used */
	unsigned cr0;			/* rctr */
	unsigned cr8;			/* pidr1: same as act->mact.pid */
	unsigned cr9;			/* pidr2: not used */
	unsigned cr10;			/* ccr:   not used */
	unsigned cr12;			/* pidr3: not used */
	unsigned cr13;			/* pidr4: not used */
	unsigned cr24;			/* not used in Mach */
	unsigned cr25;			/* not used in Mach */
	unsigned cr26;			/* not used in Mach */
	unsigned mpsfu_high;		/* not used in Mach */
	unsigned mpsfu_low;		/* not used in Mach */
	unsigned mpsfu_ovflo;		/* not used in Mach */
	struct Act *act;		/* Virtual back pointer to activation (in PCBs only) */
};	/* 64 words (256 bytes) total */

/*
 * The following structure holds floating-point state.
 * Initially, the status register (the upper 32 bits of fr0)
 * contains -1, an illegal value, indicating no floating-point state.
 * The floating-point state is then cleared the first time the user process
 * attempts to use the FPU.
 * The variable fpu_act points to the activation corresponding
 * to the FPU state currently active in the FPU; it is used for lazy switching
 * of FPU state.
 */
struct parisc_fp_state {
	double	fr0;	/* sr, exp1 */
	double	fr1;	/* exp2, exp3 */
	double	fr2;	/* exp4, exp5 */
	double	fr3;	/* exp6, exp7 */
	double	fr4;
	double	fr5;
	double	fr6;
	double	fr7;
	double	fr8;
	double	fr9;
	double	fr10;
	double	fr11;
	double	fr12;
	double	fr13;
	double	fr14;
	double	fr15;
	double	fr16;
	double	fr17;
	double	fr18;
	double	fr19;
	double	fr20;
	double	fr21;
	double	fr22;
	double	fr23;
	double	fr24;
	double	fr25;
	double	fr26;
	double	fr27;
	double	fr28;
	double	fr29;
	double	fr30;
	double	fr31;
};	/* 32 doublewords (256 bytes) total */


/* User state for this activation.

   The user state has to be kept separate from the main activation
   because of two aspects of PA idiocy:
   First, the chip kicks itself out of virtual mode on all traps and interrupts.
   Second, accesses to the same memory through both physical and virtual addresses
   are not allowed unless the physical and virtual addresses are equal
   (or unless a bunch of silly cache magic is done).
   When a trap occurs, getting back into virtual mode
   involves saving state into the saved_state structure,
   meaning it must be accessible by physical address.

   In the short term, it is acceptable to just equivalently map
   all activation structures.
   However, equivalently-mapped memory can never be paged normally.
   Therefore, my current long-term plan is to define activations as being
   potentially pageable, with the ss pointer pointing to equivalently-mapped memory,
   probably on the same page, but on non-intersecting cache lines.
   Paging activations must be done specially, keeping the ss pointer correct.
   It must remain constant as long as the activation is paged in,
   and since pageout is not done at arbitrary points in glue code,
   it can be relied on within glue code (but nowhere else).
   Note that paging out an activation must first flush the FPU
   if that activation's pcb is fpu_pcb.

   All these shenanigans won't be necessary on processors
   that nicely stay in virtual mode for traps (i.e. most processors).  */
struct PCB
{
	/* Primary user state.
	   Some assembly language code depends on this being first (PCB_SS == 0).  */
	struct parisc_saved_state ss;

	/* User floating-point state.
	   If the first word is -1, then the FPU state is zeroed
	   the first time the activation tries to use the FPU.  */
	struct parisc_fp_state fs;
};

/*
 * Machine-dependent per-activation information
 */
struct MachineAct
{
	/* Pointer to process control block,
	   accessible in both physical and virtual mode.  */
	struct PCB *pcb;

	/* XXX for now, activations are equivalently mapped, and pcb points here.  */
	/* XXX ensure that this is doubleword-aligned, for FP registers */
	struct PCB xxx_pcb;

	/* This points to the return-transition routine in glue code.
	   It is set in an activation when it is called from another,
	   and invoked to return to the client activation.
	   It is responsible for copying any return data,
	   popping the called activation back off the thread,
	   and returning it to its actpool.

	   For slow, message-based calls,
	   it invokes the client's reentry_routine once the return message is built.
	   For fast, direct-copy calls,
	   it also takes care of all activation manipulation on the client side,
	   and returns directly into the client activation.

	   Nudging/encapsulating any activation
	   involves examining its return_routine and, based on that,
	   invoking a return-linkage-encapsulation routine.
	   This will convert any direct-copy return linkage into a message-based linkage,
	   allowing the encapsulated activation to be controlled or migrated away.
	   Encapsulating a non-topmost activation similarly
	   involves examining the return_routine in the activation it called
	   and invoking its return-encapsulation routine.

	   Register conventions:
		t2 = server (callee) activation
		ret0 = return code from server (XXX)
	   All other registers are just residual trash from the server.  */
	__dead void	(*return_routine)();

	/* During message-based calls,
	   this points to the glue routine to invoke
	   when returning from a called activation.
	   This pointer is undefined
	   during direct-copy calls or when no call is in progress.

	   Register conventions:
		r1 = thread
		t3 = client (caller) activation
		t1 = client (caller) PCB
		arg0 = return message
		ret0 = return code (XXX)
	   All other registers are just residual trash from the server.  */
	__dead void	(*reentry_routine)();


	/* User stack pointer for initialization on migrating RPCs */
	vm_offset_t	user_stack;

	/* User receive buffer/size for migrating RPCs with memory */
	vm_offset_t	user_rbuf;
	vm_size_t	user_rbuf_size;

	/* XXX when in a trap or syscall, contains a reference to the user activation.  */
	struct Act	*user_act;


	/* These values are simply cached from the task's map's pmap's values.  */
	unsigned space;
	int pid;

	/* Software-managed user single-step */
	int	 sstep;

	/* tmp storage for RPC code */
	int	spare[16];

#ifdef	KGDB
	/* hacks for kgdb single stepping */
	unsigned kgdb_trace;
	unsigned saved_insn;
	unsigned saved_oqt;
	unsigned saved_sqt;
#endif
};
typedef struct MachineAct MachineAct;

/*
 * useful synonyms for registers
 */
#define rp	r2
#define arg3	r23
#define arg2	r24
#define arg1	r25
#define arg0	r26
#define	dp	r27
#define	ret0	r28
#define	ret1	r29
#define	sp	r30

#define fret	fr4
#define farg0	fr4
#define farg1	fr5
#define farg2	fr6
#define farg3	fr7

#define rctr	cr0
#define pidr1	cr8
#define pidr2	cr9
#define ccr	cr10
#define sar	cr11
#define pidr3	cr12
#define pidr4	cr13
#define iva	cr14
#define eiem	cr15
#define itmr    cr16
#define pcsq	cr17
#define pcoq	cr18
#define iir	cr19
#define isr	cr20
#define ior	cr21
#define ipsw    cr22
#define eirr	cr23
#define ppda    cr24
#define hta     cr25

#endif ASSEMBLER

/*
 * Save State Flags
 */
#define	SS_INTRAP	0x01	/* State saved from trap */
#define	SS_INSYSCALL	0x02	/* State saved from system call */
#define SS_ININT	0x04	/* On the interrupt stack */
#define	SS_PSPKERNEL	0x08	/* Previous context stack pointer kernel */

#define	SS_INTRAP_POS	 31	/* State saved from trap */
#define	SS_INSYSCALL_POS 30	/* State saved from system call */
#define SS_ININT_POS	 29	/* On the interrupt stack */
#define	SS_PSPKERNEL_POS 28	/* Previous context stack pointer kernel */



#define ACT_STATIC_KLUDGE	300 /* XXX saved-state must be direct-mapped */


#endif	/* _PARISC_ACT_H_ */
