/*
 *  (c) Copyright 1988 HEWLETT-PACKARD COMPANY
 *
 *  To anyone who acknowledges that this file is provided "AS IS"
 *  without any express or implied warranty:
 *      permission to use, copy, modify, and distribute this file
 *  for any purpose is hereby granted without fee, provided that
 *  the above copyright notice and this notice appears in all
 *  copies, and that the name of Hewlett-Packard Company not be
 *  used in advertising or publicity pertaining to distribution
 *  of the software without specific, written prior permission.
 *  Hewlett-Packard Company makes no representations about the
 *  suitability of this software for any purpose.
 */

/*
 * Copyright (c) 1990,1991,1992,1994 The University of Utah and
 * the Computer Systems Laboratory (CSL).  All rights reserved.
 *
 * THE UNIVERSITY OF UTAH AND CSL PROVIDE THIS SOFTWARE IN ITS "AS IS"
 * CONDITION, AND DISCLAIM ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
 * WHATSOEVER RESULTING FROM ITS USE.
 *
 * CSL requests users of this software to return to csl-dist@cs.utah.edu any
 * improvements that they make and grant CSL redistribution rights.
 *
 * 	Utah $Hdr: clock.c 1.17 94/11/10$
 */

/*
 *	PA-RISC low-level clock handling
 */

#include <mach/machine/vm_types.h>
#include <machine/act.h>
#include <machine/pdc.h>
#include <machine/iomod.h>
#include <machine/psl.h>
#include <machine/machspl.h>
#include <machine/regs.h>
#include <machine/clock.h>
#ifdef hp700
#include <iotypes.h>
#include <machine/asp.h>
#ifdef USELEDS
#include <machine/led.h>
#endif
#endif
#include <mach/time_value.h>

struct time_value time;
int iticks_per_tick;
extern int tick;

unsigned time_inval;
int tick_too_soon = 0;

extern void softclock();
void hardclock();

startrtclock()
{

	/*
	 * The PDC puts the value of itics / 10 msec into low memory at the
	 * location MEM_10MSEC.
	 */
	iticks_per_tick = PAGE0->mem_10msec;
	/*
	 * Install interrupt handlers for hard and soft clock.
	 */
	(void)procitab(SPLCLOCK, hardclock, 0, 0);
	(void)procitab(SPLSCLK, softclock, 0, 0);
	/*
	 * Start the interval timer.
	 */
	mtctl(CR_ITMR, time_inval = (mfctl(CR_ITMR) + iticks_per_tick));
}


/*
 * initialize the system time from the time of day clock
 */

inittodr()
{
	static struct pdc_tod tod;	/* XXX must be equiv. mapped */

	/*
	 * call the PDC to get the time of day. Fortunately the PDC returns
	 * the time in standard UNIX format.
	 */
	pdc_iodc(PAGE0->mem_pdc, 1, PDC_TOD, PDC_TOD_READ, &tod);

	time = *(struct time_value *)&tod;

	/*
	 * is the time completely absurd?
	 *
	 * XXX - should really test this against the time in the superblock
	 */
	if ((long)time.seconds < 0) {
		time.seconds = SECYR * (1990 - 1970);
		printf("Warning: time of day clock not initialized -- check and reset\n");
	}
}

/*
 * reset the time of day clock to the value in time
 */

resettodr()
{
	/*
	 * since the PDC wants the time in UNIX format we don't have to
	 * do anything fancy...
	 */

	pdc_iodc(PAGE0->mem_pdc, 1, PDC_TOD, PDC_TOD_WRITE, 
		 time.seconds, time.microseconds);
}

#ifdef	PC596XMT_BUG
#include <sys/time.h>
#endif

/*
 * This is the clock interrupt handling routine.
 * It is enterred from the eirr jump table (eirr_switch.int_action).
 */

void
hardclock(int_bit, ssp)
	int int_bit;
	struct parisc_saved_state *ssp;
{
	register int time_remaining, s;
	register long ticks;
	unsigned int now;

	ticks = 0;
	s = splclock();

	now = mfctl(CR_ITMR);
#ifdef DEBUG
	/*
	 * Ticks should only happen at or after the time they're
	 * scheduled for.  If one arrives too early, we complain, but 
	 * just ignore it.
	 */
	if ((now - time_inval) < 0) {
		tick_too_soon++;
		/*
		 * Fudge this and keep going otherwise we may hang
		 * around for a complete wrap around of the interval
		 * timer before the machine comes back.  Kernel debuggers
		 * can cause this.
		 */
		time_inval = now;
	}
#endif

	/*
	 * See if we missed any clock ticks.
	 */
	do {
		ticks++;
		mtctl(CR_ITMR, (int)(time_inval += iticks_per_tick));
		mtctl(CR_EIRR, SPL_IMASK_CLOCK);
		time_remaining = time_inval - mfctl(CR_ITMR);
	} while (time_remaining < 0);

	splx(s);

	clock_interrupt(ticks * tick, USERMODE(ssp), ssp->eiem == SPL0);

#ifdef	PC596XMT_BUG	/* consult "../hpsgc/if_i596.c" for details */
#include "lan.h"
#if NLAN > 0
	{
		extern int pc586_eir[];
		extern struct iomod *prochpa;
		extern struct timeval pc586_toutxmt[];
		register struct timeval *timep = (struct timeval *) &time;
		register struct timeval *toutp;

		/*
		 * Check if our pc596 chip failed to interrupt on transmit.
		 * If it did, force a LAN interrupt for our CPU.
		 *
		 * Note: When there is only one LAN board, help the compiler
		 * optimize this hack (we're in hardclock, ya know).
		 *
		 * Note: Since we are in the clock interrupt code, it's safe
		 * to treat the system time as non-volatile.
		 */

#if NLAN == 1
#define	unit	0
#else
		register int unit;

		for (unit = 0; unit < NLAN; unit++)
#endif
		{
			toutp  = &pc586_toutxmt[unit];
			if (((long)toutp->tv_sec) > 0 && pc586_eir[unit] > 0 &&
			    timercmp(timep, toutp, >)) {
				toutp->tv_sec = -1;
				prochpa->io_eir = pc586_eir[unit];
			}
		}
#if NLAN == 1
#undef	unit
#endif
	}
#endif	/* NLAN > 0 */
#endif	/* PC596XMT_BUG */

#ifdef USELEDS
	/*
	 * Do really important stuff: pulse the heartbeat LED.  Turn off
	 * the disk/LAN LEDs in case we are idle and they got left on.
	 */
	{
		static int heartbeat;
		extern int hz;

		if (++heartbeat == hz >> 1) {
			if (inledcontrol == 0)
				ledcontrol(0, LED_DISK|LED_LANRCV|LED_LANXMT,
					   LED_PULSE);
			heartbeat = 0;
		}
	}
#endif
}

/* 
 * Routine to turn on the softclock interrupt bit. 
 */

setsoftclock()

{
	PAGE0->mem_hpa->io_eir = SPL_IBIT_SOFTCLK;
}
