/*
 * Mach Operating System
 * Copyright (c) 1989 Carnegie Mellon University.
 * Copyright (c) 1990,1991,1992,1994 The University of Utah and
 * the Computer Systems Laboratory (CSL).
 * All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON, THE UNIVERSITY OF UTAH AND CSL ALLOW FREE USE OF
 * THIS SOFTWARE IN ITS "AS IS" CONDITION, AND DISCLAIM ANY LIABILITY
 * OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF
 * THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 * 
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 * 
 * any improvements or extensions that they make and grant Carnegie Mellon
 * the rights to redistribute these changes.
 *
 * 	Utah $Hdr: conf.c 1.21 94/12/15$
 */

/*
 *	Device switch table
 */
#include <mach/machine/vm_types.h>
#include <device/conf.h>

extern int block_io_mmap();

/*
 * we don't define a console here. After we have configured the devices
 * we will assign a console though the indirect table.
 */
#define	cnname		"cn"
#define cnopen 		nulldev
#define cnclose 	nulldev
#define cnread 		nulldev
#define cnwrite 	nulldev
#define cngetstat 	nulldev
#define cnsetstat 	nulldev
#define cnreset 	nulldev
#define cnportdeath 	nulldev

extern int timeopen(),timeclose(),timemmap();
#define timename	"time"

#include <mux.h>
#define	muxname		"mux"
#if NMUX > 0
extern int muxopen(),muxclose(),muxread(),muxwrite(),muxgetstat(),
	muxsetstat(),muxreset(),muxportdeath();
#else
#define muxopen 	nodev
#define muxclose 	nodev
#define muxread 	nodev
#define muxwrite 	nodev
#define muxgetstat 	nodev
#define muxsetstat 	nodev
#define muxreset 	nodev
#define muxportdeath 	nodev
#endif

#include <dca.h>
#define	dcaname		"dca"
#if NDCA > 0
extern int dcaopen(),dcaclose(),dcaread(),dcawrite(),dcagetstat(),
	dcasetstat(),dcareset(),dcaportdeath();
#else
#define dcaopen 	nodev
#define dcaclose 	nodev
#define dcaread 	nodev
#define dcawrite 	nodev
#define dcagetstat 	nodev
#define dcasetstat 	nodev
#define dcareset 	nodev
#define dcaportdeath 	nodev
#endif

#include <grf.h>
#define	grfname		"grf"
#if NGRF > 0
extern int grfopen(),grfclose(),grfgetstat(),grfsetstat(),grfmap();
#else
#define grfopen 	nodev
#define grfclose 	nodev
#define grfgetstat 	nodev
#define grfsetstat 	nodev
#define grfmap		nodev
#endif

#include <ite.h>
#define	itename		"ite"
#if NITE > 0
extern int iteopen(),iteclose(),iteread(),itewrite(),itegetstat(),
	itesetstat(),itereset(),iteportdeath();
#else
#define iteopen 	nodev
#define iteclose 	nodev
#define iteread 	nodev
#define itewrite 	nodev
#define itegetstat 	nodev
#define itesetstat 	nodev
#define itereset 	nodev
#define iteportdeath 	nodev
#endif

#include <hil.h>
#define	hilname		"hil"
#if NHIL > 0
extern int hilopen(),hilclose(),hilread(),hilgetstat(),hilsetstat(),hilmap();
#else
#define hilopen 	nodev
#define hilclose 	nodev
#define hilread 	nodev
#define hilgetstat 	nodev
#define hilsetstat 	nodev
#define hilmap	 	nodev
#endif

#include <rd.h>
#define	rdname		"rd"
#if NRD > 0
int	rdopen(),rdread(),rdwrite(),rdioctl(),rdsize();
#else
#define rdopen 		nodev
#define rdread 		nodev
#define rdwrite 	nodev
#define rdioctl 	nodev
#define rdsize		nodev
#endif

#include <sd.h>
#define	sdname		"sd"
#if NSD > 0
int	sdopen(),sdclose(),sdread(),sdwrite(),sdgetstat(),sdsetstat();
#else
#define sdopen 		nodev
#define sdclose 	nodev
#define sdread 		nodev
#define sdwrite 	nodev
#define sdgetstat 	nodev
#define sdsetstat	nodev
#endif

#include <ciolan.h>
#define	lanname		"ln"
#if NCIOLAN > 0
extern int lanopen(),lanwrite(),langetstat(),lansetstat(),lanreset(),
	lansetinput();
#else
#define lanopen		nodev
#define lanwrite	nodev
#define langetstat	nodev
#define lansetstat	nodev
#define lanreset	nodev
#define lansetinput	nodev
#endif

#include <lan.h>
#define	pc586name	"pc"
#if NLAN > 0
extern int pc586open(),pc586output(),pc586getstat(),pc586setstat(),
	pc586reset(), pc586setinput();
#else
#define pc586open	nodev
#define pc586output	nodev
#define pc586getstat	nodev
#define pc586setstat	nodev
#define pc586reset	nodev
#define pc586setinput	nodev
#endif

#include <medusa.h>
#define medname		"fddi"
#if NMEDUSA > 0
extern int medopen(),medclose(),medoutput(),medgetstat(),medsetstat(),
	medreset(), medsetinput();
#else
#define medopen		nodev
#define medclose	nodev
#define medoutput	nodev
#define medgetstat	nodev
#define medsetstat	nodev
#define medreset	nodev
#define medsetinput	nodev
#endif

#include <gkd.h>
#define	gkdname		"gkd"
#if NHIL > 0
extern int gkdopen(),gkdclose(),gkdread(),gkdgetstat(),gkdsetstat(),gkdmap();
#else
#define gkdopen 	nodev
#define gkdclose 	nodev
#define gkdread 	nodev
#define gkdgetstat 	nodev
#define gkdsetstat 	nodev
#define gkdmap	 	nodev
#endif

#define mmname		"mkmem"
extern int mmopen(), mmread(), mmwrite();

/*
 * List of devices - console must be at slot 0
 */
struct dev_ops	dev_name_list[] =
{
	/*name,		open,		close,		read,
	  write,	getstat,	setstat,	mmap,
	  async_in,	reset,		port_death,	subdev,
	  dev_info */

	{ cnname,	cnopen,		cnclose,	cnread,
	  cnwrite,	cngetstat,	cnsetstat,	nodev,
	  nodev,	nulldev,	cnportdeath,	0,
	  nodev },

	{ timename,	timeopen,	timeclose,	nulldev,
	  nulldev,	nulldev,	nulldev,	timemmap,
	  nodev,	nulldev,	nulldev,	0,
	  nodev },

	{ muxname,	muxopen,	muxclose,	muxread,
	  muxwrite,	muxgetstat,	muxsetstat,	nodev,
	  nodev,	muxreset,	muxportdeath,	0,
	  nodev },

	{ rdname,       rdopen,         nulldev,        rdread,
          rdwrite,      rdioctl,        rdsize,		block_io_mmap,
          nodev,        nulldev,        nulldev,        8,
	  nodev },

	{ sdname,       sdopen,         sdclose,        sdread,
          sdwrite,      sdgetstat,	sdsetstat,	block_io_mmap,
          nodev,        nulldev,        nulldev,        8,
	  nodev },

	{ lanname,	lanopen,	nulldev,	nodev,
	  lanwrite,	langetstat,	lansetstat,	nodev,
	  lansetinput,	lanreset,	nulldev,	0,
	  nodev },
	
	{ pc586name,	pc586open,	nulldev,	nodev,
	  pc586output,	pc586getstat,	pc586setstat,	nodev,
	  pc586setinput,pc586reset,	nulldev,	0,
	  nodev },
	
	{ mmname,	mmopen,		nulldev,	mmread,
	  mmwrite,	nulldev,	nulldev,	nodev,
	  nulldev,	nulldev,	nulldev,	0,
	  nodev },

	{ grfname,	grfopen,	grfclose,	nodev,
	  nodev,	grfgetstat,	grfsetstat,	grfmap,
	  nodev,	nulldev,	nulldev,	0,
	  nodev },

	{ hilname,	hilopen,	hilclose,	hilread,
	  nodev,	hilgetstat,	hilsetstat,	hilmap,
	  nodev,	nulldev,	nulldev,	0,
	  nodev },

	{ dcaname,	dcaopen,	dcaclose,	dcaread,
	  dcawrite,	dcagetstat,	dcasetstat,	nodev,
	  nodev,	dcareset,	dcaportdeath,	0,
	  nodev },

	{ itename,	iteopen,	iteclose,	iteread,
	  itewrite,	itegetstat,	itesetstat,	nodev,
	  nodev,	itereset,	iteportdeath,	0,
	  nodev },

	{ medname,	medopen,	medclose,	nodev,
	  medoutput,	medgetstat,	medsetstat,	nodev,
	  medsetinput,	medreset,	nulldev,	0,
	  nodev },

	{ gkdname,	gkdopen,	gkdclose,	gkdread,
	  nodev,	gkdgetstat,	gkdsetstat,	gkdmap,
	  nodev,	nulldev,	nulldev,	0,
	  nodev },
};

int	dev_name_count = sizeof(dev_name_list)/sizeof(dev_name_list[0]);

/*
 * Indirect list. We'll stomp on the console device when we finally figure
 * out where the console should be.
 */
struct dev_indirect dev_indirect_list[] = {

	/* console */
	{ "console",	&dev_name_list[0],	0 }
};
int	dev_indirect_count = sizeof(dev_indirect_list)
				/sizeof(dev_indirect_list[0]);
