/*
 *  (c) Copyright 1988 HEWLETT-PACKARD COMPANY
 *
 *  To anyone who acknowledges that this file is provided "AS IS"
 *  without any express or implied warranty:
 *      permission to use, copy, modify, and distribute this file
 *  for any purpose is hereby granted without fee, provided that
 *  the above copyright notice and this notice appears in all
 *  copies, and that the name of Hewlett-Packard Company not be
 *  used in advertising or publicity pertaining to distribution
 *  of the software without specific, written prior permission.
 *  Hewlett-Packard Company makes no representations about the
 *  suitability of this software for any purpose.
 */

/*
 * Copyright (c) 1990,1991,1994 The University of Utah and
 * the Computer Systems Laboratory (CSL).  All rights reserved.
 *
 * THE UNIVERSITY OF UTAH AND CSL PROVIDE THIS SOFTWARE IN ITS "AS IS"
 * CONDITION, AND DISCLAIM ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
 * WHATSOEVER RESULTING FROM ITS USE.
 *
 * CSL requests users of this software to return to csl-dist@cs.utah.edu any
 * improvements that they make and grant CSL redistribution rights.
 *
 * 	Utah $Hdr: interrupt.s 1.39 94/12/14$
 */

#include <machine/asm.h>
#include <assym.s>
#include <machine/act.h>
#include <machine/trap.h>
#include <machine/psl.h>
#include <machine/machspl.h>
#include <machine/pmap.h>
#include <mach/machine/vm_param.h>
#ifdef	GPROF
#include <machine/gprof.h>
#endif

#include <machine/debug.h>

/*
 * Stack frame used for interrupts:
 *	parisc_saved_state (SS_SIZE)
 *	padding
 *	C stack frame (FM_SIZE)
 */
/* #define FR_SIZE		   eval(roundup((SS_SIZE+FM_SIZE),8)) */
#define FR_SIZE  ((SS_SIZE+FM_SIZE+FR_ROUND-1) & ~(FR_ROUND - 1))

/* 
 * Data space variables imported by this file. 
 */

	.space	$PRIVATE$
	.subspa	$DATA$

	.import $global$,data
	.import istackptr,data
	.import	cur_pcb,data
	.import	kernel_port,data
	.import intstack_top,data
	.import	need_ast,data

/* 
 * zero indicates interrupt stack has overflowed 
 */

istk_overflow
	.word	1

	.space $TEXT$
	.subspa $CODE$

/*
 * ihandler(type)
 *
 * Build a saved state structure on the interrupt stack, switch the processor
 * into virtual addressing mode and call the routine interrupt().
 */

	.export	ihandler,entry
	.proc
	.callinfo calls, frame=FR_SIZE, save_sp, hpux_int

ihandler
	.entry

	/*
	 * get the value of istackptr, if it's zero then we're already on the
	 * interrupt stack.
	 */

	ldil	L%istackptr,t1
	ldw	R%istackptr(t1),sp
	comb,<>,n	sp,r0,$istackatbase

	/* 
	 * We're already on the interrupt stack, get the old stack pointer back
	 * that we shoved off into tr3. Then put a value into t1 to
	 * indicate that we are in the situation that the previous stack
	 * pointer was the kernel stack and we are in an interrupt.
	 */

	mfctl	tr3,sp
	movib,tr SS_PSPKERNEL+SS_ININT,t1,$istacknotatbase
	/* XXX note branch is not nullified.  */

$istackatbase

	/* Allocate a stack frame.  */ 	
	ldo	FR_SIZE(sp),sp 
	/*
	 * We're not on the interrupt stack. The value of istackptr is the 
	 * interrupt stack.
	 */

	stw	r0,R%istackptr(t1)
	ldi	SS_ININT,t1

$istacknotatbase

	/*
	 * We are creating a standard stack frame here for us to put our save 
	 * state onto.
	 */

	stw	r0,FM_PSP(sp)
	stw	t1,-FR_SIZE+SS_FLAGS(sp)

$intcontinue

	/*
	 * Save away r26, r30 and r22. This gets us completely out of the
	 * temporary registers.
	 */

	mfctl	tr2,t1
	stw	t1,-FR_SIZE+SS_R26(sp)
	mfctl	tr3,t1
	stw	t1,-FR_SIZE+SS_R30(sp)
	mfctl	tr4,t1
	stw	t1,-FR_SIZE+SS_R22(sp)

	/*
	 * save registers r21 (t2) and r20 (t3) for our use
	 */

	stw	t3,-FR_SIZE+SS_R20(sp)
	stw	t2,-FR_SIZE+SS_R21(sp)

	/* 
	 * Now we're about to turn on the PC queue.  We'll also go to virtual
	 * mode in the same step. We need to get the space registers pointing
	 * at the kernel.
	 */

	mfsp	sr4,t1
	stw	t1,-FR_SIZE+SS_SR4(sp)

	mfsp	sr5,t1
	stw	t1,-FR_SIZE+SS_SR5(sp)

	mfsp	sr6,t1
	stw	t1,-FR_SIZE+SS_SR6(sp)

	mfsp	sr7,t1
	stw	t1,-FR_SIZE+SS_SR7(sp)

	mfsp	sr1,t1
	stw	t1,-FR_SIZE+SS_SR1(sp)

	mtsp	r0,sr4
	mtsp	r0,sr5
	mtsp	r0,sr6
	mtsp	r0,sr7
	mtsp	r0,sr1


	/*	
	 * save the protection ID registers and then set them to be kernel 
	 * only. We shouldn't be touching user space since this is an 
	 * interrupt...
	 */
	mfctl	pidr1,t1
	stw	t1,-FR_SIZE+SS_CR8(sp)

	mfctl	pidr2,t1
	stw	t1,-FR_SIZE+SS_CR9(sp)

	mfctl	pidr3,t1
	stw	t1,-FR_SIZE+SS_CR12(sp)

	mfctl	pidr4,t1
	stw	t1,-FR_SIZE+SS_CR13(sp)

	ldi	PARISC_PID_KERNEL,t1
	mtctl	t1,pidr1
	mtctl	t1,pidr2
	mtctl	t1,pidr3
	mtctl	t1,pidr4

	/*
	 * Now, save away other volatile state that prevents us from turning
	 * the PC queue back on, namely, the pc queue and ipsw.  Note that
	 * I don't have to save the iir, isr or ior because I'm an interrupt,
	 * and not a fault, trap or check.
	 */
	mfctl	pcoq,t1
	stw	t1,-FR_SIZE+SS_IIOQH(sp)

	mtctl	r0,pcoq			/* advance the queue */

	mfctl	pcoq,t1
	stw	t1,-FR_SIZE+SS_IIOQT(sp)


	mfctl	pcsq,t1
	stw	t1,-FR_SIZE+SS_IISQH(sp)

	mtctl	r0,pcsq			/* advance the queue */

	mfctl	pcsq,t1
	stw	t1,-FR_SIZE+SS_IISQT(sp)

	mfctl	eiem,t1
	stw	t1,-FR_SIZE+SS_CR15(sp)

	mfctl	ipsw,t2
	stw	t2,-FR_SIZE+SS_CR22(sp)

	/*
	 * stuff the space queue
	 */	

	mtctl	r0,pcsq
	mtctl	r0,pcsq

	/*
	 * set the new psw to be data and code translation, interrupts 
	 * disabled, protection enabled, Q bit on
	 */

	ldil	L%KERNEL_PSW,t1
	ldo	R%KERNEL_PSW(t1),t1
		
	mtctl	t1,ipsw

	/*
	 * Load up a real value into eiem to reflect an spl level of splhigh.
	 * That's the SPL level we will be calling interrupt() with.
	 * Right now interrupts are still off.
	 */
	ldi	SPLHIGH,t1
	mtctl	t1,eiem

	/*
	 * Load in the address to "return" to with the rfi instruction.
	 */

	ldil	L%$intnowvirt,t1
	ldo	R%$intnowvirt(t1),t1

	/*
	 * now stuff the offset queue
	 */

	mtctl	t1,pcoq
	ldo	4(t1),t1
	mtctl	t1,pcoq

	/*
	 * Must do rfir not rfi since we may be called from tlbmiss routine
	 * (to handle page fault) and it uses the shadowed registers.
	 */
	rfir
	nop

$intnowvirt

	/* 
	 * Now we're in virtual mode.  
	 */
#ifdef	GPROF
	/*
	 * Kernel profiling. Turn the recovery counter on when entering
	 * the kernel. Use the lower bits of the cycle counter as a
	 * pseudo random interval. 
	 */
	mfctl	cr16,t1
	extru	t1,31,PROFILING_LEN,t1
	mtctl	t1,cr0
	ssm	PSW_R,r0
#endif

#if 0 /* XXX we no longer use the interrupt stack for interrupts!!! */
#ifdef DEBUG
	/* Test if we've overflowed the interrupt stack.
	 *
	 * If so, there is an extra page beyond intstack_top which has no 
	 * access rights. We change this so that we can now access it and 
	 * adjust the interrupt type parameter to indicate an overflow error.
	 * However, if this is a recursive overflow then do nothing: the 
	 * access rights of the extra page have already been changed and 
	 * we're probably just trying to sync the discs.  This is dangerous, 
	 * however.
	 */

	/*
	 * get the value of the top of the interrupt stack
	 */

	ldil	L%intstack_top,t1
	ldo	R%intstack_top(t1),t1

	/*
	 * are we into the red zone page? 
	 */

	comb,<,n  sp,t1,$noistkovfl

	/*
	 * We've overflowed, load the overflow flag
	 */

	ldil	L%istk_overflow,t1
	ldw	R%istk_overflow(t1),t1

	/* 
	 * have we already overflowed before?
	 */

	comb,=,n  r0,t1,$noistkovfl

	/*
	 * This is the first overflow, clear the overflow flag
	 */

	ldil	L%istk_overflow,t1
	stw	r0,R%istk_overflow(t1)

	/*
	 * Change the access rights for the interrupt stack buffer page so 
	 * that we can use it as part of the interrupt stack while we panic.
	 */

	ldil	L%intstack_top,t1
	ldo	R%intstack_top(t1),t1

	/*
	 * flush the TLB entry for the top of the interrupt stack
	 */

	pdtlb	r0(sr1,t1)

	/* 
	 * We need to change the access rights on this page.  Basically
	 * we will chase down the mapping entry (which will be easy since
	 * it's equivalently mapped) and then change the protection.
	 * We will then change the interrupt to be I_IS_OVFL which will
	 * cause a panic.
	 */

	/*
	 * XXX - this needs to be changed if we change the page size
	 */

	/*
	 * Convert the physical page number into the form we need to index
	 * the ptov table and locate the correct entry.
	 */

	depi    0,31,12,t1
	extrs	t1,23,24,t1
	mfctl	ptov,t2	
	add	t1,t2,t1

	/* 
	 * t1 now points to the ptov table entry for this physical page.
	 * For this page the protection is in the ptov table since it's
	 * an equivalently mapped part of the kernel.
	 */

	ldw	PTOV_TLBPROT(t1),t2

	/*
	 * load up the protection that we want on this page
 	 */

	ldil	L%TLB_AR_KRW,t3

	/*
	 * put the new protection into the protection field
	 */

	dep	t3,11,7,t2
	stw	t2,PTOV_TLBPROT(t1)

	/*
	 * change the interrupt type to interrupt control stack overflow
	 */

	ldil	L%I_IS_OVFL,arg0

$noistkovfl
#endif
#endif /* XXX */

	/* 
	 * Save all the general registers
	 */

	stw	r1,-FR_SIZE+SS_R1(sp)
	stw	r2,-FR_SIZE+SS_R2(sp)
	stw	r3,-FR_SIZE+SS_R3(sp)
	stw	r4,-FR_SIZE+SS_R4(sp)
	stw	r5,-FR_SIZE+SS_R5(sp)
	stw	r6,-FR_SIZE+SS_R6(sp)
	stw	r7,-FR_SIZE+SS_R7(sp)
	stw	r8,-FR_SIZE+SS_R8(sp)
	stw	r9,-FR_SIZE+SS_R9(sp)
	stw	r10,-FR_SIZE+SS_R10(sp)
	stw	r11,-FR_SIZE+SS_R11(sp)
	stw	r12,-FR_SIZE+SS_R12(sp)
	stw	r13,-FR_SIZE+SS_R13(sp)
	stw	r14,-FR_SIZE+SS_R14(sp)
	stw	r15,-FR_SIZE+SS_R15(sp)
	stw	r16,-FR_SIZE+SS_R16(sp)
	stw	r17,-FR_SIZE+SS_R17(sp)
	stw	r18,-FR_SIZE+SS_R18(sp)
	stw	r19,-FR_SIZE+SS_R19(sp)

	/*
	 * r20 (t3) was saved above
	 * r21 (t2) was saved above
	 * r22 (t1) was saved above
	 */

	stw	r23,-FR_SIZE+SS_R23(sp)
	stw	r24,-FR_SIZE+SS_R24(sp)
	stw	r25,-FR_SIZE+SS_R25(sp)

	/*
	 * r26 (arg0) was saved above
	 */ 

	stw	r27,-FR_SIZE+SS_R27(sp)
	stw	r28,-FR_SIZE+SS_R28(sp)
	stw	r29,-FR_SIZE+SS_R29(sp)

	/* 
	 * r30 (sp) was saved above
	 */

	stw	r31,-FR_SIZE+SS_R31(sp)


	/* 
	 * Save all the space registers 
	 */

	mfsp	sr0,t1
	stw	t1,-FR_SIZE+SS_SR0(sp)

	/*
	 * sr1 was saved earlier
	 */

	mfsp	sr2,t1
	stw	t1,-FR_SIZE+SS_SR2(sp)

	mfsp	sr3,t1
	stw	t1,-FR_SIZE+SS_SR3(sp)

	/*
	 * sr4 - sr7 already saved above.
	 */

	/*
	 * Save the necessary control registers that were not already saved.
	 */

	mfctl	rctr,t1
	stw	t1,-FR_SIZE+SS_CR0(sp)

	mfctl	sar,t1
	stw	t1,-FR_SIZE+SS_CR11(sp)

	/*
	 * set up dp for the kernel	
	 */

	ldil	L%$global$,dp
	ldo	R%$global$(dp),dp

	/*
	 * interrupt type was setup back before the call to ihandler.
	 * The save state is at the start of locals 
	 */

	ldo	-FR_SIZE(sp),arg1

	.import	interrupt,code
	ldil	L%interrupt,t1
	ldo	R%interrupt(t1),t1
	.call
	blr	r0,rp
	bv	0(t1)
	nop

	/*
	 * Ok, back from C. Disable interrupts while we restore things
	 */
	rsm	PSW_I,r0

	/* 
	 * Restore most of the state, up to the point where we need to turn 
	 * off the PC queue.  Going backwards, starting with control regs.
	 */
	
	ldw	-FR_SIZE+SS_CR15(sp),t1
	mtctl	t1,eiem

	ldw	-FR_SIZE+SS_CR11(sp),t1
	mtctl	t1,sar

	ldw	-FR_SIZE+SS_CR0(sp),t1
	mtctl	t1,rctr

	/*
	 * Restore the lower space registers, we'll restore sr5 - sr7 after
	 * we have turned off translations
	 */

	ldw	-FR_SIZE+SS_SR0(sp),t1
	mtsp	t1,sr0

	ldw	-FR_SIZE+SS_SR1(sp),t1
	mtsp	t1,sr1

	ldw	-FR_SIZE+SS_SR2(sp),t1
	mtsp	t1,sr2

	ldw	-FR_SIZE+SS_SR3(sp),t1
	mtsp	t1,sr3

	/* 
	 * restore most of the general registers
	 */

	ldw	-FR_SIZE+SS_R1(sp),r1
	ldw	-FR_SIZE+SS_R2(sp),r2
	ldw	-FR_SIZE+SS_R3(sp),r3
	ldw	-FR_SIZE+SS_R4(sp),r4
	ldw	-FR_SIZE+SS_R5(sp),r5
	ldw	-FR_SIZE+SS_R6(sp),r6
	ldw	-FR_SIZE+SS_R7(sp),r7
	ldw	-FR_SIZE+SS_R8(sp),r8
	ldw	-FR_SIZE+SS_R9(sp),r9
	ldw	-FR_SIZE+SS_R10(sp),r10
	ldw	-FR_SIZE+SS_R11(sp),r11
	ldw	-FR_SIZE+SS_R12(sp),r12
	ldw	-FR_SIZE+SS_R13(sp),r13
	ldw	-FR_SIZE+SS_R14(sp),r14
	ldw	-FR_SIZE+SS_R15(sp),r15
	ldw	-FR_SIZE+SS_R16(sp),r16
	ldw	-FR_SIZE+SS_R17(sp),r17
	ldw	-FR_SIZE+SS_R18(sp),r18
	ldw	-FR_SIZE+SS_R19(sp),r19

	/* 
	 * r20 (t3) is used as a temporary and will be restored later
	 * r21 (t2) is used as a temporary and will be restored later
	 * r22 (t1) is used as a temporary and will be restored later
	 */

	ldw	-FR_SIZE+SS_R23(sp),r23
	ldw	-FR_SIZE+SS_R24(sp),r24
	ldw	-FR_SIZE+SS_R25(sp),r25
	ldw	-FR_SIZE+SS_R26(sp),r26
	ldw	-FR_SIZE+SS_R27(sp),r27
	ldw	-FR_SIZE+SS_R28(sp),r28
	ldw	-FR_SIZE+SS_R29(sp),r29

	/* 
	 * r30 (sp) will be restored later
	 */

	b	$int_new_page
	ldw	-FR_SIZE+SS_R31(sp),r31


	/*
	 * clear the system mask, this puts us back into physical mode.
	 * 
	 * N.B: We will start running with code translation on, data 
	 * translation off, interrupts off and the Q bit is off. What this 
	 * means is that we can not tolerate a code translation miss from here
	 * until we do the rfi. We will align the next address to be on a page
	 * boundary so that we will be guarenteed that the page is in the TLB.
	 * All of the code from here to the rfi must be within one page.
	 */

	.align  PARISC_PGBYTES
$int_new_page
	rsm	RESET_PSW,r0	

	/*
	 * restore the protection ID registers
	 */

	ldw	-FR_SIZE+SS_CR13(sp),t1
	mtctl	t1,pidr4

	ldw	-FR_SIZE+SS_CR12(sp),t1
	mtctl	t1,pidr3

	ldw	-FR_SIZE+SS_CR9(sp),t1
	mtctl	t1,pidr2

	ldw	-FR_SIZE+SS_CR8(sp),t1
	mtctl	t1,pidr1

	/*
	 * restore space registers sr4 - sr7
	 */

	ldw	-FR_SIZE+SS_SR4(sp),t1
	mtsp	t1,sr4

	ldw	-FR_SIZE+SS_SR5(sp),t1
	mtsp	t1,sr5

	ldw	-FR_SIZE+SS_SR6(sp),t1
	mtsp	t1,sr6

	ldw	-FR_SIZE+SS_SR7(sp),t1
	mtsp	t1,sr7

	/*
	 * Cut back the interrupt stack. First get the flags from the saved
	 * state area to see if this is the last frame on the stack
	 */

	ldw	-FR_SIZE+SS_FLAGS(sp),t1

	/*
	 * Check the SS_PSPKERNEL bit in the saved state flags
	 */
	
	bb,<,n  t1,SS_PSPKERNEL_POS,$notlastistk

	/*
	 * SS_PSPKERNEL bit not set, this means that we are at the base of the
	 * interrupt stack.
	 *
	 * This maens that -FR_SIZE(sp) is the original value of istackptr 
	 * before this stack frame. Save the value into the istackptr variable,
	 * this indicates that we're not on the interrupt stack.
	 */

	ldo     -FR_SIZE(sp),t1
	ldil    L%istackptr,t2
	stw     t1,R%istackptr(t2)


$notlastistk

	/*
	 * Now restore the IIA space queue and the IIA offset queue from the 
	 * saved state. Save the PC we are going to to check if it is in
	 * "user" mode.
	 */
	ldw	-FR_SIZE+SS_IIOQH(sp),t2
	mtctl	t2,pcoq

	ldw	-FR_SIZE+SS_IIOQT(sp),t1
	mtctl	t1,pcoq

	ldw	-FR_SIZE+SS_IISQH(sp),t3
	mtctl	t3,pcsq

	ldw	-FR_SIZE+SS_IISQT(sp),t1
	mtctl	t1,pcsq

	/*
	 * get the ipsw we are returning to.
	 */
	ldw	-FR_SIZE+SS_CR22(sp),t1

	/*
	 * We only want to check for ASTs if we would be in user mode when
	 * it is delivered.  "User mode" means the privilege level is non-
	 * zero and space is non-zero (taking an AST when priv is non-zero
	 * but space is zero, i.e. early in the gateway page, can cause
	 * grief for the signal code.)
	 */
	extru,<> t2,31,2,r0
	b,n	no_int_ast
	comib,=,n 0,t3,no_int_ast

	/*
	 * See if we need to have an AST delivered.
	 */
	ldil	L%need_ast,t2
	ldw	R%need_ast(t2),t2
	comib,=,n 0,t2,no_int_ast

	/*
	 * If we do need an AST we massage things to make it look like
	 * we took a trap and jump into the trap handler.  To do this
	 * we essentially back out of the interrupt and jump to the
	 * trap handler.
	 */

#ifndef DEBUG
	depi	1,PSW_L_P,1,t1
#endif
	mtctl	t1,ipsw
	ldi	I_LPRIV_XFR,arg0

	/*
	 * This is my original attempt at handing ast's from interrupts,
	 * before I realized the reason for the complication of Mike's code.
	 * (The thandler wants to do an rfir on entry,
	 * but our shadow registers are long gone by this point.)
	 * So, I see three options:
	 *	1. Send a flag into thandler, say by setting arg0 to -1,
	 *	   and make it use rfi instead of rfir in that case.
	 *	   Would unnecessarily slow down the common case slightly.
	 *	2. Use Mike's method, which means lots of duplicated code.
	 *	3. Don't do it at all for now - just let the ast get
	 *	   caught at the next trap or syscall.
	 *	   Should work - just has terrible performance characteristics.
	 *	   When I implement microscheduling (hopefully RSN),
	 *	   interrupt and trap handling will become symmetrical
	 *	   and this problem should go away entirely.
	 * For now, since I'm lazy, I'm chosing #3.
	 *
	 * Mike sez: you cannot just ignore them.  Some user code (cthreads?)
	 * spins rather than blocking.  Back to #2.
	 */
#if 0
	/*
	 * Put the original t1, sp, and arg0 back into the temporary registers
	 * where thandler is expecting to find them.
	 */
	ldw	-FR_SIZE+SS_R22(sp),t1
	mtctl	t1,tr4
	ldw	-FR_SIZE+SS_R26(sp),t1
	mtctl	t1,tr2
	ldw	-FR_SIZE+SS_R30(sp),t1
	mtctl	t1,tr3

	/*
	 * Restore the original t2 and t3.
	 */
	ldw	-FR_SIZE+SS_R21(sp),t2
	ldw	-FR_SIZE+SS_R20(sp),t3

	/*
	 * Jump into the beginning of the trap handler.
	 */
	b,n	$handle_as_trap
#else
	/*
	 * Determine the location of the PCB saved state for this thread
	 * and copy over the original values of registers still in use
	 * (t1, t2, arg0, sp, t3).
	 */
        ldil    L%cur_pcb,t1
        ldw     R%cur_pcb(t1),t1

	ldw	-FR_SIZE+SS_R22(sp),t3
	stw	t3,SS_R22(t1)
	ldw	-FR_SIZE+SS_R21(sp),t3
	stw	t3,SS_R21(t1)
	ldw	-FR_SIZE+SS_R26(sp),t3
	stw	t3,SS_R26(t1)
	ldw	-FR_SIZE+SS_R30(sp),t3
	stw	t3,SS_R30(t1)
	ldw	-FR_SIZE+SS_R20(sp),t3
	stw	t3,SS_R20(t1)

	/*
	 * Indicate that we are a trap.
	 */
	ldi	SS_INTRAP,t2
	stw	t2,SS_FLAGS(t1)

	/*
	 * Save PC queues, IPSW and EIEM
	 */
	mfctl	pcoq,t2
	stw	t2,SS_IIOQH(t1)
	mtctl	r0,pcoq
	mfctl	pcoq,t2
	stw	t2,SS_IIOQT(t1)
	mfctl	pcsq,t2
	stw	t2,SS_IISQH(t1)
	mtctl	r0,pcsq
	mfctl	pcsq,t2
	stw	t2,SS_IISQT(t1)

	mfctl	ipsw,t2
	stw	t2,SS_CR22(t1)
	mfctl	eiem,t2
	stw	t2,SS_CR15(t1)

	/*
	 * Save SIDs and PIDs and set them to reflect kernel mode
	 * before we re-enable mapping.
	 *
	 * Note we leave pidr1 pointing to user space.
	 */
	mfsp	sr0,t2
	stw	t2,SS_SR0(t1)
	mfsp	sr1,t2
	stw	t2,SS_SR1(t1)
	mfsp	sr2,t2
	stw	t2,SS_SR2(t1)
	mfsp	sr3,t2
	stw	t2,SS_SR3(t1)
	mfsp	sr4,t2
	stw	t2,SS_SR4(t1)
	mfsp	sr5,t2
	stw	t2,SS_SR5(t1)

	mtsp	r0,sr4
	mtsp	r0,sr5
	mtsp	r0,sr6

	mfctl	pidr1,t2
	stw	t2,SS_CR8(t1)
	mfctl	pidr3,t2
	stw	t2,SS_CR12(t1)
	mfctl	pidr4,t2
	stw	t2,SS_CR13(t1)

	ldi	PARISC_PID_KERNEL,t2
	mtctl	t2,pidr4

	/*
	 * Load space and offset queues, IPSW and EIEM for jump to
	 * virtual mode.  We will land down in the thandler code.
	 *
	 * Note we use rfi and not rfir since we haven't touched the
	 * shadowed registers and they may have been clobbered
	 * anyway when we went virtual for the call to interrupt().
	 */
	ldil	L%KERNEL_PSW,t2
	ldo	R%KERNEL_PSW(t2),t2
	mtctl	t2,ipsw

	ldi	SPLHIGH,t2
	mtctl	t2,eiem

	mtctl	r0,pcsq
	mtctl	r0,pcsq
	ldil	L%$trapnowvirt,t2
	ldo	R%$trapnowvirt(t2),t2
	mtctl	t2,pcoq
	ldo	4(t2),t2
	mtctl	t2,pcoq

	rfi
	nop
#endif

no_int_ast
	mtctl	t1,ipsw

	/*
	 * restore all the registers that we used above
	 */
	ldw	-FR_SIZE+SS_R22(sp),t1
	ldw	-FR_SIZE+SS_R21(sp),t2
	ldw	-FR_SIZE+SS_R20(sp),t3
	ldw	-FR_SIZE+SS_R30(sp),sp

	rfi
	nop
	.exit
	.procend



/*
 * thandler(type)
 *
 * Build a saved state structure on the kernel stack, switch the processor
 * into virtual addressing mode and call the routine trap().
 */

	.export	thandler,entry
	.proc
	.callinfo calls, frame=FM_SIZE, save_sp, hpux_int

thandler

	.entry
	/* If we're already on the interrupt stack (istackptr == 0), then we
	 * simply use the current sp.  Otherwise, we must go to the kernel 
	 * stack.  If we're already on the kernel stack (active_ksps == 0), then
	 * just use the current sp.  In either case, if we're not the first
	 * frame on EITHER stack, mark ss_flags with SS_PSPKERNEL. Either way,
	 * mark the flag with SS_INTRAP so that the save state is indelibly 
	 * marked as an trap.
	 */

	ldil	L%istackptr,t1
	ldw	R%istackptr(t1),t1

	comb,<>,n t1,r0,$handle_as_trap

	/* 
	 * We are currently running on the interrupt stack so use sp from the
	 * interrupt stack
	 *
	 * Create a stack frame and clear the previous stack pointer word. 
	 * Then save some of the state that we will need.
	 */

	ldo	FR_SIZE(sp),sp
	stw	0,FM_PSP(sp)

	mfctl	ior,t1
	stw	t1,-FR_SIZE+SS_CR21(sp)

	mfctl	isr,t1
	stw	t1,-FR_SIZE+SS_CR20(sp)

	mfctl	iir,t1
	stw	t1,-FR_SIZE+SS_CR19(sp)

	ldi	SS_PSPKERNEL+SS_INTRAP,t1

	/*
	 * process the trap in ihandler since we are on the interrupt stack
	 */
	b 	$intcontinue
	stw	t1,-FR_SIZE+SS_FLAGS(sp)

$handle_as_trap

	/* XXX check for trap in glue code */

	/*
	 * Find the current PCB (t1).
	 */
	ldil	L%cur_pcb,t1
	ldw	R%cur_pcb(t1),t1

	/*
	 * Get another temporary (t2).
	 */
	stw	t2,SS_R21(t1)
	
	/* 
	 * Mark the saved state as in a trap.
	 */
	ldi	SS_INTRAP,t2
	stw	t2,SS_FLAGS(t1)

	/* 
	 * Now, save away other volatile state that prevents us from turning
	 * the PC queue back on, namely, the pc queue and ipsw, and the
	 * interrupt information.
	 */

	mfctl	pcoq,t2
	stw	t2,SS_IIOQH(t1)

	mtctl	r0,pcoq

	mfctl	pcoq,t2
	stw	t2,SS_IIOQT(t1)

	mfctl	pcsq,t2
	stw	t2,SS_IISQH(t1)

	mtctl	r0,pcsq

	mfctl	pcsq,t2
	stw	t2,SS_IISQT(t1)

	mfctl	ipsw,t2
	stw	t2,SS_CR22(t1)

	mfctl	eiem,t2
	stw	t2,SS_CR15(t1)

	mfctl	ior,t2
	stw	t2,SS_CR21(t1)

	mfctl	isr,t2
	stw	t2,SS_CR20(t1)

	mfctl	iir,t2
	stw	t2,SS_CR19(t1)

	/*
	 * Important note from Mike: shadowed registers are garbage at
	 * this point if we came here via tlbmiss.  We cannot save these
	 * registers until the RFIR has been done (restoring their interrupt
	 * time values).
	 */

	mfctl	tr4,t2
	stw	t2,SS_R22(t1)

	mfctl	tr2,t2
	stw	t2,SS_R26(t1)

	mfctl	tr3,t2
	stw	t2,SS_R30(t1)

	/* 
	 * Save the space registers. 
	 */
	mfsp	sr0,t2
	stw	t2,SS_SR0(t1)

	mfsp	sr1,t2
	stw	t2,SS_SR1(t1)

	mfsp	sr2,t2
	stw	t2,SS_SR2(t1)

	mfsp	sr3,t2
	stw	t2,SS_SR3(t1)

	mfsp	sr4,t2
	stw	t2,SS_SR4(t1)

	mfsp	sr5,t2		/* XXX not needed once kernel code uses pidr1 */
	stw	t2,SS_SR5(t1)

	/*
	 * Protection ID registers.
	 */
	mfctl	pidr1,t2
	stw	t2,SS_CR8(t1)

	mfctl	pidr3,t2	/* XXX not needed once only glue code uses it */
	stw	t2,SS_CR12(t1)

	mfctl	pidr4,t2	/* XXX not needed once kernel code uses pidr1 */
	stw	t2,SS_CR13(t1)

	/* 
	 * Now we're about to turn on the PC queue.  We'll also go to virtual
	 * mode in the same step.
	 * Point the relevant space registers to the kernel space
	 */
	mtsp	r0,sr4
	mtsp	r0,sr5
	mtsp	r0,sr6

	/*
	 * Set up pidr4 for accessing kernel memory.
	 * pidr1 remains set to the user code's PID.
	 */
	ldi	PARISC_PID_KERNEL,t2
	mtctl	t2,pidr4

	/*
	 * load the space queue
	 */
	mtctl	r0,pcsq
	mtctl	r0,pcsq

	/*
	 * set the new psw to be data and code translation, interrupts 
	 * disabled, protection enabled, Q bit on
	 */
	ldil	L%KERNEL_PSW,t2
	ldo	R%KERNEL_PSW(t2),t2
	
	mtctl	t2,ipsw

	/*
	 * load up a real value into eiem to reflect an spl level of splhigh.
	 * Interrupts are off since we never enabled them with the rfi above.
	 */
	ldi	SPLHIGH,t2
	mtctl	t2,eiem

	/*
	 * load in the address to "return" to with the rfi instruction
	 */
	ldil	L%$trapnowvirt,t2
	ldo	R%$trapnowvirt(t2),t2

	/*
	 * load the offset queue
	 */
	mtctl	t2,pcoq
	ldo	4(t2),t2
	mtctl	t2,pcoq

	/*
	 * Must do rfir not rfi since we may be called from tlbmiss routine
	 * (to handle page fault) and it uses the shadowed registers.
	 */
	rfir
	nop

$trapnowvirt
#ifdef	GPROF
	/*
	 * Kernel profiling. Turn the recovery counter on when entering
	 * the kernel. Use the lower bits of the cycle counter as a
	 * pseudo random interval. 
	 */
	mfctl	cr16,t2
	extru	t2,31,PROFILING_LEN,t2
	mtctl	t2,cr0
	ssm	PSW_R,r0
#endif

	/*
	 * Save the callee-save registers for easy examination.
	 *
	 * XXX this is also needed for the hack backward compatible "immediate"
	 * get/set state routines which might be called before we return
	 * to this activation (this includes the special setstate break used
	 * by the emulator during sigreturn).  The new get/set state routines
	 * would not be invoked until we have returned from trap at which
	 * time the callee save registers would have been restored from the
	 * stack.
	 */
	stw	r3,SS_R3(t1)
	stw	r4,SS_R4(t1)
	stw	r5,SS_R5(t1)
	stw	r6,SS_R6(t1)
	stw	r7,SS_R7(t1)
	stw	r8,SS_R8(t1)
	stw	r9,SS_R9(t1)
	stw	r10,SS_R10(t1)
	stw	r11,SS_R11(t1)
	stw	r12,SS_R12(t1)
	stw	r13,SS_R13(t1)
	stw	r14,SS_R14(t1)
	stw	r15,SS_R15(t1)
	stw	r16,SS_R16(t1)
	stw	r17,SS_R17(t1)
	stw	r18,SS_R18(t1)

	/* 
	 * Save all caller-save registers that we haven't saved already.
	 */
	stw	r1,SS_R1(t1)
	stw	r2,SS_R2(t1)
	stw	r19,SS_R19(t1)
	stw	r20,SS_R20(t1)
	/*
	 * r21 already saved
	 * r22 already saved
	 */
	stw	r23,SS_R23(t1)
	stw	r24,SS_R24(t1)
	stw	r25,SS_R25(t1)
	/*
	 * r26 already saved
	 */
	stw	r27,SS_R27(t1)
	stw	r28,SS_R28(t1)
	stw	r29,SS_R29(t1)
	/*
	 * r30 already saved
	 */
	stw	r31,SS_R31(t1)

	/*
	 * Save the necessary control registers that were not already saved.
	 */
	mfctl	rctr,t2
	stw	t2,SS_CR0(t1)

	mfctl	sar,t2
	stw	t2,SS_CR11(t1)

	/*
	 * This is the base fallback point
	 * for both interruption and encapsulation rollback.
	 * Interruption rollback must still save the callee-saved registers,
	 * which weren't saved above.
	 * Encapsulation rollback only needs to mark the activation as in a trap.
	 *
	 * t1 = PCB
	 */

	/*
	 * Find the user activation (t3).
	 * The user PCB will still hang around in t1 until the call to trap().
	 */
	ldw	PCB_ACT(t1),t3

	/*
	 * Find a kernel activation to run on (t2).
	 */
	ldil	L%kernel_port,t4
	ldw	R%kernel_port(t4),t4
	ldw	IPT_ACTS(t4),t2
	comb,<>,n t2,r0,$kact_available

	/* XXX block */
	ldil	L%istackptr,t3
	ldw	R%istackptr(t3),sp
	stw	r0,R%istackptr(t3)
	.import panic_no_kstacks,code
	ldil	L%panic_no_kstacks,t2
	ldo	R%panic_no_kstacks(t2),t2
	bv	0(t2)
	ldo	FM_SIZE(sp),sp

$kact_available

	/*
	 * Remove it from the port.
	 */
	ldw	ACT_IPT_NEXT(t2),ret1
	stw	ret1,IPT_ACTS(t4)

	/*
	 * Attach the kernel activation (t2) to the current thread (now r1).
	 * The user activation is still in t3.
	 */
	ldw	ACT_THREAD(t3),r1
	stw	t3,ACT_LOWER(t2)
	stw	t2,ACT_HIGHER(t3)
	stw	t2,THREAD_TOP_ACT(r1)
	stw	r1,ACT_THREAD(t2)

	/*
	 * Take a reference on the user activation. XXX
	 */
	ldw	ACT_REF_COUNT(t3),t4
	ldo	1(t4),t4
	stw	t4,ACT_REF_COUNT(t3)
	stw	t3,ACT_USER_ACT(t2)

#ifdef ACTCHECK
	ldw		ACT_REENTRY_ROUTINE(t3),t4
	comb,=,n	r0,t4,$reis0
	ldw		ACT_PCB(t2),ret0
	ldil		L%cur_pcb,ret1
	stw		ret0,R%cur_pcb(ret1)
	ldw		ACT_USER_STACK(t2),sp
	break		0x1f,0x10
$reis0
#endif

	/*
	 * Set the reentry routine in the user activation
	 * to point to the ordinary reentry point.
	 */
	ldil	L%$trap_reentry,t4
	ldo	R%$trap_reentry(t4),t4
	stw	t4,ACT_REENTRY_ROUTINE(t3)

	/*
	 * Make the new activation officially the current one.
	 */
	ldw	ACT_PCB(t2),t3
	ldil	L%cur_pcb,t4
	stw	t3,R%cur_pcb(t4)

	/*
	 * At this point,
	 * the new activation is attached to and owned by the thread.
	 * We are done mucking in the old activation.
	 * Therefore, this is the new base for interruption rollback.
	 */

	/*
	 * Initialize the kernel stack pointer from the new activation.
	 */
	ldw	ACT_USER_STACK(t2),sp

	/* 
	 * Clear out the previous stack pointer word.
	 */
	stw	0,FM_PSP(sp)

	/*
	 * Load the global pointer for the kernel
	 */
	ldil	L%$global$,dp
	ldo	R%$global$(dp),dp

	/*
	 * Now leaving glue code.
	 */

	/*
	 * Call the C routine trap. Interrupt type (arg0) was setup back before
	 * the call to thandler. 
	 *
	 * Don't bother setting the RETURN_ROUTINE,
	 * since we know that the kernel trap handler always returns "nicely".
	 */
	copy	t1,arg1
	.import	trap,code
	ldil	L%trap,t1
	ldo	R%trap(t1),t1
	.call
	blr     r0,rp
	bv	0(t1)
	nop

	/*
	 * Returning to glue code.
	 */
$trap_return

	/*
	 * Find the thread (r1), kernel activation (t2), and user activation (t3).
	 */
	ldil	L%cur_pcb,t4
	ldw	R%cur_pcb(t4),t2
	ldw	PCB_ACT(t2),t2
	ldw	ACT_LOWER(t2),t3
	ldw	ACT_THREAD(t2),r1

	/*
	 * Call any waiting returnhandlers
	 */
	ldw	ACT_HANDLERS(t3),t1
	comb,<>,n t1,r0,$call_handlers

	/*
	 * Jump to the user activation's return transition routine.
	 * This could in fact be the null_act's thread-destruction routine.
	 */

	/*
	 * Remove the kernel activation from the thread.
	 */
	stw	r0,ACT_THREAD(t2)
	stw	t3,THREAD_TOP_ACT(r1)

#ifdef ACTCHECK
	stw	r0,ACT_LOWER(t2)
	stw	r0,ACT_HIGHER(t3)
	stw	r0,ACT_USER_ACT(t2)
#endif

	/* 
	 * Drop the reference count on the user activation. XXX
	 */
	ldw	ACT_REF_COUNT(t3),t1
	ldo	-1(t1),t1
	stw	t1,ACT_REF_COUNT(t3)

	/*
	 * Return the kernel activation to the kernel port.
	 */
	ldil	L%kernel_port,t1
	ldw	R%kernel_port(t1),t1
	ldw	IPT_ACTS(t1),arg3
	stw	t2,IPT_ACTS(t1)
	stw	arg3,ACT_IPT_NEXT(t2)

	/*
	 * Return to the user activation.
	 * This marks the end of the (very simple) return transition.
	 */
	ldw	ACT_PCB(t3),t1
	stw	t1,R%cur_pcb(t4)

	/*
	 * Invoke the user activation's return reentry routine.
	 * This could be the null_act's thread-destruction routine
	 * or an encapsulated activation's APC routine.
	 */
	ldw	ACT_REENTRY_ROUTINE(t3),t4
#ifdef ACTCHECK
	stw		r0,ACT_REENTRY_ROUTINE(t3)
	/*
	 * XXX cannot just check for trap_reentry, may get here with
	 * syscall_reentry if call_handlers were invoked.
	 */
	comb,<>,n	r0,t4,$rerok
	break		0x1f,0x11
$rerok
#endif
	bv,n	0(t4)

$trap_reentry
	/*
	 * This is a very clean point for interruption and encapsulation rollback.
	 * It is also the point where execution starts
	 * in activations in which a new thread has just been created.
	 */
#ifdef ACTCHECK
	ldw		SS_IIOQH(t1),t2
	comb,<>,n	r0,t2,$pcok
	break		0x1f,0x12
$pcok
#endif

	/*
	 * Disable the FPU so the user code will trap if it tries to use it.
	 */
	mtctl	r0,ccr

	/*
	 * XXX restore the callee-save registers in case the "immediate"
	 * get/set state routines were called.  See comment in save code
	 * above for details.
	 */
	ldw	SS_R3(t1),r3
	ldw	SS_R4(t1),r4
	ldw	SS_R5(t1),r5
	ldw	SS_R6(t1),r6
	ldw	SS_R7(t1),r7
	ldw	SS_R8(t1),r8
	ldw	SS_R9(t1),r9
	ldw	SS_R10(t1),r10
	ldw	SS_R11(t1),r11
	ldw	SS_R12(t1),r12
	ldw	SS_R13(t1),r13
	ldw	SS_R14(t1),r14
	ldw	SS_R15(t1),r15
	ldw	SS_R16(t1),r16
	ldw	SS_R17(t1),r17
	ldw	SS_R18(t1),r18

	/* 
	 * Restore most of the general caller-save registers
	 */
	ldw	SS_R1(t1),r1
	ldw	SS_R2(t1),r2
	ldw	SS_R19(t1),r19
	ldw	SS_R20(t1),r20
	/* 
	 * r21 (t2) is used as a temporary and will be restored later
	 * r22 (t1) is used as a temporary and will be restored later
	 */
	ldw	SS_R23(t1),r23
	ldw	SS_R24(t1),r24
	ldw	SS_R25(t1),r25
	ldw	SS_R26(t1),r26
	ldw	SS_R27(t1),r27
	ldw	SS_R28(t1),r28
	ldw	SS_R29(t1),r29
	ldw	SS_R30(t1),r30
	ldw	SS_R31(t1),r31

	rsm	PSW_I,r0

	/* 
	 * Restore most of the state, up to the point where we need to turn 
	 * off the PC queue. Going backwards, starting with control regs.
	 */

	ldw	SS_CR15(t1),t2
	mtctl	t2,eiem

	ldw	SS_CR11(t1),t2
	mtctl	t2,sar

	ldw	SS_CR8(t1),t2
	mtctl	t2,pidr1

	ldw	SS_CR0(t1),t2
	mtctl	t2,rctr

	/*
	 * Restore the lower space registers; we'll restore sr5 - sr7 after
	 * we have turned off translations
	 */

	ldw	SS_SR0(t1),t2
	mtsp	t2,sr0

	ldw	SS_SR1(t1),t2
	mtsp	t2,sr1

	ldw	SS_SR2(t1),t2
	mtsp	t2,sr2

	ldw	SS_SR3(t1),t2
	mtsp	t2,sr3

	/*
	 * clear the system mask; this puts us back into physical mode.
	 * 
	 * N.B: Better not be any code translation traps from this point
	 * on. Of course, we know this routine could never be *that* big.
	 */
	rsm	RESET_PSW,r0	

	/*
	 * restore the protection ID registers
	 */

	ldw	SS_CR9(t1),t2
	mtctl	t2,pidr2
	
	ldw	SS_CR12(t1),t2
	mtctl	t2,pidr3
	
	ldw	SS_CR13(t1),t2
	mtctl	t2,pidr4

	/*
	 * restore the space registers
	 */

	ldw	SS_SR4(t1),t2
	mtsp	t2,sr4

#ifdef MULTIPLE_SPACES_PER_TASK
	ldw	SS_SR5(t1),t2
#endif
	mtsp	t2,sr5
#ifdef MULTIPLE_SPACES_PER_TASK
	ldw	SS_SR6(t1),t2
#endif
	mtsp	t2,sr6

	/*
	 * finally we can restore the space and offset queues and the ipsw
	 */

	ldw	SS_IIOQH(t1),t2
	mtctl	t2,pcoq

	ldw	SS_IIOQT(t1),t2
	mtctl	t2,pcoq

	ldw	SS_IISQH(t1),t2
	mtctl	t2,pcsq

	ldw	SS_IISQT(t1),t2
	mtctl	t2,pcsq

	ldw	SS_CR22(t1),t2
	mtctl	t2,ipsw

	/*	
	 * restore the last registers: r21 (t2) and r22 (t1).
	 */
	ldw	SS_R21(t1),r21
	ldw	SS_R22(t1),r22

	rfi
	nop
	.exit
	.procend


/* XXX this is just a disgusting temporary hack. */
	.export	thread_machine_start_self_user,entry
	.proc
	.callinfo
thread_machine_start_self_user
	.entry
	ldi	I_LPRIV_XFR,arg0
	b,n	$trapnowvirt
	.exit
	.procend


	.export	$call_handlers,entry
	.proc
	.callinfo
$call_handlers
	.entry
	/*
	 * Find the user PCB (t1).
	 */
	ldw	ACT_PCB(t3),t1

	/*
	 * Stash the callee-save registers so they can be modified.
	 */
	stw	r3,SS_R3(t1)
	stw	r4,SS_R4(t1)
	stw	r5,SS_R5(t1)
	stw	r6,SS_R6(t1)
	stw	r7,SS_R7(t1)
	stw	r8,SS_R8(t1)
	stw	r9,SS_R9(t1)
	stw	r10,SS_R10(t1)
	stw	r11,SS_R11(t1)
	stw	r12,SS_R12(t1)
	stw	r13,SS_R13(t1)
	stw	r14,SS_R14(t1)
	stw	r15,SS_R15(t1)
	stw	r16,SS_R16(t1)
	stw	r17,SS_R17(t1)
	stw	r18,SS_R18(t1)

	/*
	 * Execute the returnhandlers.
	 */
	.import	act_execute_returnhandlers,code
	ldil	L%act_execute_returnhandlers,t1
	ldo	R%act_execute_returnhandlers(t1),t1
	.call
	blr     r0,rp
	bv	0(t1)
	nop

	ldil	L%cur_pcb,t4
	ldw	R%cur_pcb(t4),t2
	ldw	PCB_ACT(t2),t2
	ldw	ACT_LOWER(t2),t3
	ldw	ACT_PCB(t3),t1

	/*
	 * Restore the callee-save registers.
	 */
	ldw	SS_R3(t1),r3
	ldw	SS_R4(t1),r4
	ldw	SS_R5(t1),r5
	ldw	SS_R6(t1),r6
	ldw	SS_R7(t1),r7
	ldw	SS_R8(t1),r8
	ldw	SS_R9(t1),r9
	ldw	SS_R10(t1),r10
	ldw	SS_R11(t1),r11
	ldw	SS_R12(t1),r12
	ldw	SS_R13(t1),r13
	ldw	SS_R14(t1),r14
	ldw	SS_R15(t1),r15
	ldw	SS_R16(t1),r16
	ldw	SS_R17(t1),r17
	b	$trap_return
	ldw	SS_R18(t1),r18
	.exit
	.procend

	.end
