/* 
 * Copyright (c) 1991, 1992, 1994, The University of Utah and
 * the Computer Systems Laboratory at the University of Utah (CSL).
 * All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software is hereby
 * granted provided that (1) source code retains these copyright, permission,
 * and disclaimer notices, and (2) redistributions including binaries
 * reproduce the notices in supporting documentation, and (3) all advertising
 * materials mentioning features or use of this software display the following
 * acknowledgement: ``This product includes software developed by the
 * Computer Systems Laboratory at the University of Utah.''
 *
 * THE UNIVERSITY OF UTAH AND CSL ALLOW FREE USE OF THIS SOFTWARE IN ITS "AS
 * IS" CONDITION.  THE UNIVERSITY OF UTAH AND CSL DISCLAIM ANY LIABILITY OF
 * ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * CSL requests users of this software to return to csl-dist@cs.utah.edu any
 * improvements that they make and grant CSL redistribution rights.
 *
 * 	Utah $Hdr: norma_glue.c 1.7 94/12/14$
 */

#include <norma_ipc.h>
#if NORMA_IPC

#include <vm/vm_kern.h>
#include <machine/pmap.h>
#include <mach/vm_param.h>
#include <mach/vm_prot.h>
#include <kern/processor.h>
#include <kern/task.h>
#include <sys/varargs.h>

int debug_norma_glue = 0;

extern int pmdebug;
extern pmap_t kernel_pmap;

vm_page_t
norma_page_grab()
{
	vm_page_t   	page = vm_page_grab();
	vm_offset_t 	addr;

	if (page == VM_PAGE_NULL)
		return(VM_PAGE_NULL);

	addr = page->phys_addr;
	if (debug_norma_glue)
		pmdebug = 0x06;
	pmap_map(addr, addr, addr+PARISC_PGBYTES, VM_PROT_READ|VM_PROT_WRITE);
	pmdebug = 0;
	return(page);
}

vm_offset_t
phystokv(x)
	vm_offset_t x;
{
	return(x);
}

void
norma_page_free(page)
	vm_page_t	page;
{
	if (page == VM_PAGE_NULL)
		return;

	pmap_remove(kernel_pmap,
		    page->phys_addr, page->phys_addr+PARISC_PGBYTES);

	vm_page_free(page);
}

norma_map_copy_object(copy)
	vm_map_copy_t copy;
{
	int i;
	vm_offset_t addr;

	for (i = 0; i < copy->cpy_npages; i++) {
		addr = copy->cpy_page_list[i]->phys_addr;

		if (debug_norma_glue)
			pmdebug = 0x06;
		pmap_map(addr, addr,
			 addr+PARISC_PGBYTES, VM_PROT_READ|VM_PROT_WRITE);
		pmdebug = 0;
	}
}

norma_unmap_copy_object(copy)
	vm_map_copy_t copy;
{
	int i;
	vm_offset_t addr;

	for (i = 0; i < copy->cpy_npages; i++) {
		addr = copy->cpy_page_list[i]->phys_addr;
		
		if (debug_norma_glue)
			printf("norma_unmap: addr = 0x%x\n", addr);
		pmap_remove(kernel_pmap, addr, addr+PARISC_PGBYTES);
	}
}

vm_map_t
norma_find_map(space)
	pa_space_t space;
{
	task_t taskp = (task_t) default_pset.tasks.prev;
	task_t lastp = (task_t) taskp->pset_tasks.next;

	printf("norma_find_map: Looking for map for space %d ... ", space);
		
	while (taskp != lastp) { 
		if (taskp->map->pmap->space == space) {
			printf("Found it: 0x%x\n", (int) taskp->map);
			return(taskp->map);
		}
		taskp = (task_t) taskp->pset_tasks.prev;
	}
	return(0);
}

/*
 * NORMA code used kdb functions, but we don't support kdb.
 */
extern int	cnputc();

/*VARARGS1*/
db_printf(fmt, va_alist)
	char *	fmt;
	va_dcl
{
	va_list	listp;

	va_start(listp);
	_doprnt(fmt, &listp, cnputc, 16);
	va_end(listp);
}

/*VARARGS1*/
kdbprintf(fmt, va_alist)
	char *	fmt;
	va_dcl
{
	va_list	listp;
	va_start(listp);
	_doprnt(fmt, &listp, cnputc, 16);
	va_end(listp);
}

void
db_printsym(off, strategy)
	long 	off, strategy;
{
	db_printf("db_printsym: %x %x\n", off, strategy);
}

#endif NORMA_IPC
