/* 
 * Copyright (c) 1990,1993,1994 The University of Utah and
 * the Computer Systems Laboratory at the University of Utah (CSL).
 * All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software is hereby
 * granted provided that (1) source code retains these copyright, permission,
 * and disclaimer notices, and (2) redistributions including binaries
 * reproduce the notices in supporting documentation, and (3) all advertising
 * materials mentioning features or use of this software display the following
 * acknowledgement: ``This product includes software developed by the
 * Computer Systems Laboratory at the University of Utah.''
 *
 * THE UNIVERSITY OF UTAH AND CSL ALLOW FREE USE OF THIS SOFTWARE IN ITS "AS
 * IS" CONDITION.  THE UNIVERSITY OF UTAH AND CSL DISCLAIM ANY LIABILITY OF
 * ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * CSL requests users of this software to return to csl-dist@cs.utah.edu any
 * improvements that they make and grant CSL redistribution rights.
 *
 * 	Utah $Hdr: pmap.h 1.24 94/12/14$
 *	Author: Mike Hibler, Bob Wheeler, University of Utah CSL, 9/90
 */

/*
 *	Pmap header for hppa.
 */

#ifndef	_PARISC_PMAP_H_
#define	_PARISC_PMAP_H_

#ifndef ASSEMBLER

#define EQUIV_HACK	/* no multiple mapping of kernel equiv space allowed */

#include <mach/machine/vm_types.h>
#include <mach/vm_statistics.h>
#include <kern/queue.h>

#endif ASSEMBLER

#ifdef hp700
#define BTLB		/* Use block TLBs: PA 1.1 and above */
#define HPT		/* Hashed (Hardware) Page Table */
#define USEALIGNMENT	/* Take advantage of cache alignment for optimization */
#endif

/*
 * Virtual to physical mapping macros/structures.
 * IMPORTANT NOTE: there is one mapping per HW page, not per MACH page.
 */

#define PARISC_HASHSIZE		4096	/* size of hash table */
#define PARISC_HASHSIZE_LOG2	12
#define PARISC_MIN_MPP		2	/* min # of mappings per phys page */

#ifndef ASSEMBLER

/*
 * This hash function is the one used by the hardware TLB walker on the 7100.
 */
#define pmap_hash(space, offset) \
	(((unsigned) (space) << 5 ^ \
	  (unsigned) (offset) >> PARISC_PGSHIFT) & \
	 (PARISC_HASHSIZE-1))
/*
 * Do not change these structures unless you change the assembly code in
 * locore.s
 */
struct mapping {
	queue_head_t	hash_link;	/* hash table links */
	queue_head_t	phys_link;	/* for mappings of a given PA */
	pa_space_t	space;		/* virtual space */
        vm_offset_t	offset;		/* virtual page number */
	unsigned	tlbpage;	/* physical page (for TLB load) */
	unsigned	tlbprot;	/* prot/access rights (for TLB load) */
};

#define MAPPING_NULL	((struct mapping *) 0)

#endif ASSEMBLER

/* TLB access/protection values */
#define TLB_REF		0x80000000	/* software only */
#define	TLB_ALIGNED	0x40000000	/* software only */
#define TLB_TRAP	0x20000000
#define TLB_DIRTY	0x10000000
#define TLB_BREAK	0x08000000
#define TLB_AR_MASK	0x07f00000
#define		TLB_AR_NA	0x07300000
#define		TLB_AR_KR	0x00000000
#define		TLB_AR_KRW	0x01000000
#define		TLB_AR_KRX	0x02000000
#define		TLB_AR_KRWX	0x03000000
#define		TLB_AR_UR	0x00f00000
#define		TLB_AR_URW	0x01f00000
#define		TLB_AR_URX	0x02f00000
#define		TLB_AR_URWX	0x03f00000
#define TLB_ZERO	0x00080000	/* software only */
#define TLB_ICACHE	0x00040000	/* software only */
#define TLB_NOTUSED	0x00020000      /* software only */
#define TLB_DCACHE	0x00010000      /* software only */
#define TLB_PID_MASK	0x0000fffe
#define TLB_WIRED	0x00000001	/* software only */

#define TLB_REF_POS	0
#define TLB_ALIGNED_POS	1
#define TLB_TRAP_POS	2
#define TLB_DIRTY_POS	3
#define TLB_BREAK_POS	4
#define TLB_ITLB_POS    12
#define TLB_ICACHE_POS  13
#define TLB_DTLB_POS    14
#define TLB_DCACHE_POS  15
#define TLB_WIRED_POS	31

/* protection for a gateway page */
#define TLB_GATE_PROT	0x04c00000

/* protection for break page */
#define TLB_BREAK_PROT	0x02c00000

#ifndef ASSEMBLER

struct phys_entry {
	queue_head_t	phys_link;	/* head of mappings of a given PA */
	struct mapping	*writer;	/* mapping with R/W access */
	unsigned	tlbprot;	/* TLB format protection */
};

#define PHYS_NULL	((struct phys_entry *)0)

/* XXX could be in vm_param.h */

#define PARISC_QUADBYTES	0x40000000
#define	parisc_round_quad(x)	((((unsigned)(x)) + PARISC_QUADBYTES-1) & \
					~(PARISC_QUADBYTES-1))
#define parisc_trunc_quad(x)	(((unsigned)(x)) & ~(PARISC_QUADBYTES-1))

struct pmap {
	simple_lock_data_t	lock;	     /* lock on pmap */
	int			ref_count;   /* reference count */
	pa_space_t		space;	     /* space for this pmap */
	int			pid;	     /* protection id for pmap */
	struct pmap		*next;	     /* linked list of free pmaps */
	struct pmap_statistics	stats;	     /* statistics */
	queue_head_t		pmap_link;   /* hashed list of pmaps */
};

typedef struct pmap *pmap_t;

#define PMAP_NULL  ((pmap_t) 0)


struct vtop_entry {
	queue_head_t	hash_link;	/* head of vtop chain */
};
#define vtop_next	hash_link.next
#define vtop_prev	hash_link.prev

#ifdef 	HPT
/*
 * If HPT is defined, we cache the last miss for each bucket using a
 * structure defined for the 7100 hardware TLB walker. On non-7100s, this
 * acts as a software cache that cuts down on the number of times we have
 * to search the vtop chain. (thereby reducing the number of instructions
 * and cache misses incurred during the TLB miss).
 *
 * The vtop_entry pointer is the address of the associated vtop table entry.
 * This avoids having to reform the address into the new table on a cache
 * miss.
 */
struct hpt_entry {
	unsigned	valid:1,	/* Valid bit */
			vpn:15,		/* Virtual Page Number */
			space:16;	/* Space ID */
	unsigned	tlbprot;	/* prot/access rights (for TLB load) */
	unsigned	tlbpage;	/* physical page (for TLB load) */
	unsigned	vtop_entry;	/* Pointer to associated VTOP entry */
};
#endif

#endif ASSEMBLER

#define HPT_SHIFT	27		/* 16 byte entry (31-4) */
#define VTOP_SHIFT	28		/* 8  byte entry (31-3) */
#define HPT_LEN		PARISC_HASHSIZE_LOG2
#define VTOP_LEN	PARISC_HASHSIZE_LOG2

#define MAX_PID		0xfffa
#define	PARISC_SID_KERNEL  0
#define	PARISC_PID_KERNEL  2

#define KERNEL_ACCESS_ID 1

#define KERNEL_TEXT_PROT (TLB_AR_KRX | (KERNEL_ACCESS_ID << 1))
#define KERNEL_DATA_PROT (TLB_AR_KRW | (KERNEL_ACCESS_ID << 1))

/* Block TLB flags */
#define BLK_ICACHE	0
#define BLK_DCACHE	1
#define BLK_COMBINED	2
#define BLK_LCOMBINED	3

#define pmap_kernel()			(kernel_pmap)
#define	pmap_resident_count(pmap)	((pmap)->stats.resident_count)
#define pmap_remove_attributes(pmap,start,end)
#define pmap_copy(dpmap,spmap,da,len,sa)
#define	pmap_update()

#define pmap_phys_address(x)	((x) << PARISC_PGSHIFT)
#define pmap_phys_to_frame(x)	((x) >> PARISC_PGSHIFT)

#define cache_align(x)		(((x) + 64) & ~(64 - 1))

#ifndef ASSEMBLER
#include <vm/pmap.h>
#endif
#endif	/* _PARISC_PMAP_H_ */
