/*
 *  (c) Copyright 1992 HEWLETT-PACKARD COMPANY
 *
 *  To anyone who acknowledges that this file is provided "AS IS"
 *  without any express or implied warranty:
 *      permission to use, copy, modify, and distribute this file
 *  for any purpose is hereby granted without fee, provided that
 *  the above copyright notice and this notice appears in all
 *  copies, and that the name of Hewlett-Packard Company not be
 *  used in advertising or publicity pertaining to distribution
 *  of the software without specific, written prior permission.
 *  Hewlett-Packard Company makes no representations about the
 *  suitability of this software for any purpose.
 */

/*
 * IOMAP support
 */

#include <sys/param.h>
#include <sys/types.h>
#include <hpsgc/bsd_glue.h>
#include <hpsgc/iomap.h>

int
#ifndef _NO_PROTO
iomap_allocate(char *map, int base, int target, int npages)
#else
iomap_allocate(map, base, target, npages)
	char *map;
	int base, target, npages;
#endif
{
	int s, space;
	
	/* 
	 * look for a hole big enough for us to fit
	 * start at the target address so that 16 bit
	 * io doesn't eat space that can only be used
	 * by 8 bit io.
	 */

	space = 0;
	s = splbio();
	while (target-- > base && space < npages) {
		if (!map[target]) {
			if (++space >= npages)
				break;
		}
		else
			space = 0;
	}
	if (space >= npages) {
		base = target;
		while (npages--)
			map[base++] = 1;
		splx(s);
		return (target);
	}
	splx(s);
	return (IOMAP_NOSPACE);
}

void
#ifndef _NO_PROTO
iomap_free(char *map, int base, int npages)
#else
iomap_free(map, base, npages)
	char *map;
	int base, npages;
#endif
{
	int s;

	s = splbio();
	while (npages--)
		map[base++] = 0;
	splx(s);
}
