/*
 * This file modified by Patrick Dalton
 * on May 6, 1992 for use HPPA.
 */
/* 
 * Mach Operating System
 * Copyright (c) 1991,1990 Carnegie Mellon University
 * Copyright (c) 1994 The University of Utah and
 * the Computer Systems Laboratory (CSL).
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 * 
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Carnegie Mellon requests users of this software to return to
 * 
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 * 
 * any improvements or extensions that they make and grant Carnegie Mellon
 * the rights to redistribute these changes.
 */
/*
 * HISTORY
 * $Log: db_interface.c,v $
 * Revision 1.3  1994/12/14  19:39:54  law
 * Copyright stuff
 *
 * Revision 1.2  1992/07/08  12:20:43  dalton
 * Checkin before split to 1.0.4 release (by LBS).
 *
 * Revision 1.1  92/06/08  14:57:38  dalton
 * Initial revision
 * 
 * Revision 2.13  91/10/09  16:13:36  af
 * 	 Revision 2.12.1.1  91/10/05  13:12:12  jeffreyh
 * 	 	Added user space access and check routines.
 * 	 	Added UNIX tasak information print routines.
 * 	 	Changed db_trap to db_task_trap.
 * 	 	[91/09/05            tak]
 * 
 * Revision 2.12.1.1  91/10/05  13:12:12  jeffreyh
 * 	Added user space access and check routines.
 * 	Added UNIX tasak information print routines.
 * 	Changed db_trap to db_task_trap.
 * 	[91/09/05            tak]
 * 
 * Revision 2.12  91/08/24  12:22:37  af
 * 	Use generic console calls.
 * 	[91/08/02  03:15:32  af]
 * 
 * Revision 2.11  91/05/14  17:33:28  mrt
 * 	Correcting copyright
 * 
 * Revision 2.10  91/03/16  14:55:31  rpd
 * 	Added resume, continuation arguments to vm_fault.
 * 	[91/02/04            rpd]
 * 	Changed kdb_trap to use db_recover.
 * 	[91/01/13            rpd]
 * 
 * 	Removed db_nofault.
 * 	db_error might return now.
 * 	[91/01/13            rpd]
 * 
 * Revision 2.9  91/02/14  14:35:43  mrt
 * 	Added flag not to invoke vm_fault(), used in crashed kernels.
 * 	[91/02/12  12:15:09  af]
 * 
 * Revision 2.8  91/02/05  17:47:51  mrt
 * 	Added author notices
 * 	[91/02/04  11:21:54  mrt]
 * 
 * 	Changed to use new Mach copyright
 * 	[91/02/02  12:25:29  mrt]
 * 
 * Revision 2.7  91/01/08  15:49:22  rpd
 * 	Changed kdb_trap to return TRUE if it actually enters kdb.
 * 	[90/12/23            rpd]
 * 	Fixed to use kdbsplhigh/kdbsplx.
 * 	[90/11/26            rpd]
 * 
 * Revision 2.6  90/12/05  23:37:26  af
 * 
 * 
 * Revision 2.5  90/12/05  20:49:41  af
 * 	Adapts for new drivers, made clear we do not expect
 * 	any environment pointer in kdb_init().
 * 	[90/12/03            af]
 * 
 * Revision 2.4  90/10/25  14:46:35  rwd
 * 	Support for watchpoints.
 * 	[90/10/16            rpd]
 * 
 * Revision 2.3  90/09/09  14:33:17  rpd
 * 	Reversed return code in kdb_intr, so that the do key press does not
 * 	generate garbage chars (if kdb enabled).
 * 	[90/08/30  17:03:22  af]
 * 
 * Revision 2.2  90/08/27  22:07:24  dbg
 * 	Got rid of ddb_regs, exception status is kept on stack.
 * 	I still have to decide to switch over to dbg's simpler
 * 	way to access memory or not. [how do you trace another
 * 	thread's user stack in his way ?]
 * 	Created.
 * 	[90/08/14            af]
 * 
 */
/*
 *	File: db_interface.c
 * 	Author: Alessandro Forin, Carnegie Mellon University
 *	Date:	8/90
 *
 *	Interface to new debugger.
 *
 */
#include <sys/reboot.h>

#include <machine/pmap.h>
#include <machine/cpu.h>

#include <mach/vm_param.h>
#include <vm/vm_map.h>
#include <vm/vm_kern.h>
#include <kern/task.h>
#include <kern/thread.h>
#include <ddb/db_task_thread.h>

#include <machine/thread.h>
#include <machine/db_machdep.h>
#include <machine/setjmp.h>
#include <machine/psw.h>

#include <sys/types.h>

/*
 *  kdb_init - initialize kernel debugger
 */

#if 0
/* 
 * This stuff def'ed out for now.  Need to figure out how to 
 * set it up so that it can enable kdb at boot time.
 */
/* 
 * kdb_init not needed.  It is defined in db_aout.c
 */
kdb_init(argc,argv)
{
	extern char end[];
	extern db_max_line, db_max_width;

	read_hp800_symtab(end);
	db_max_line = DB_MAX_LINE;
	db_max_width = DB_MAX_WIDTH;
	kdb_enable();
}

/*
 *  kdb_enable - enable kdb usage
 */
kdb_enable()
{
	register struct restart_blk *rb = (struct restart_blk *)RESTART_ADDR;

	extern int kdb_breakpoint(), halt_all_cpus();

	if (boothowto & RB_KDB) {
		rb-> rb_bpaddr = kdb_breakpoint;
	} else
		rb-> rb_bpaddr = halt_all_cpus;
	init_restartblk();
}

#endif 0

/*
 * Received keyboard interrupt sequence.
 */
int	kdbactive = 0;

#if 0
kdb_kintr()
{
	if (kdbactive == 0 && (boothowto & RB_KDB)) {
		printf("\n\nkernel: keyboard interrupt\n");
		gimmeabreak();
		return 1;
	}
	return 0;
}
#endif
static char *kdb_trap_names[] = {
	"Non-existent trap",
	"High priority machine check",
	"Power failure trap",
	"Recover counter overflow trap",
	"External interrupt",
	"Low priority machine check",
	"Instruction page fault",
	"Instruction memory protection fault",
	"Illegal instruction trap",
	"Break instruction trap",
	"Privileged operation trap",
	"Privileged register trap",
	"Overflow",
	"Conditional trap",
	"Assist exception trap",
	"Data page fault",
	"Non-access instruction TLB miss",
	"Non-access data TLB miss",
	"Data memory protection fault",
	"Data memory break",
	"TLB dirty",
	"Page reference trap",
	"Assist Emulation trap",
	"Higher privilege transfer",
	"Lower privilege transfer",
	"Taken branch trap",
	"Kernel preemption trap",
	"Interrupt stack overflow",
	"Kernel stack overflow",
        "Interrupt stack lost"
};

#define	TRAP_TYPES  (sizeof(kdb_trap_names) / sizeof(kdb_trap_names[0]))

/*
 *  kdb_trap - field a BPT trap
 *
 *  Esp points to the saved register state
 *
 *  Flag is TRUE when we are called from trap() [because of a fatal error],
 *  it is FALSE when we are called from kdb_breakpoint().  Only in this
 *  latter case is the TLB info in *esp valid.
 */
extern jmp_buf_t *db_recover;

kdb_trap(esp, flag)
	register struct hp800_saved_state *esp;
{
	int    cause, type, s;
	struct hp800_saved_state *old_exc_frame;
	task_t task;
	unsigned kdbpeek();

	s = kdbsplhigh();	/* if not already */

	cause = esp->cr19;   /* would like to say IIR for cr19 */
	type = break13(cause);

	if ((flag == 1) && (db_recover != 0)) {
		/*
		 *	We can recover from this exception.
		 */
		db_printf("Caught ");
		if (type >= TRAP_TYPES)
		    db_printf("type %d", type);
		else
		    db_printf("%s", kdb_trap_names[type]);
		db_printf(" trap, pc = %x, va = %x",
			  esp->iioq_head, esp->cr19);
		db_error("");
		/*NOTREACHED*/
	}

	if (flag == 1)
	  kdbprinttrap(type, esp);

	/*
	 * Make things reentrant, so that we can debug the debugger
	 */
	old_exc_frame = db_cur_exc_frame;
	db_cur_exc_frame = esp;

	kdbactive++;
	db_task_trap(cause, flag, USERMODE(esp));
	kdbactive--;

	db_cur_exc_frame = old_exc_frame;

	task = (current_thread())? current_thread()->task: TASK_NULL;

	/* 
	 *  Must figure out how to reset the instruction queue
	 *  to prevent a loop
	 */
	if (!flag && isa_kbreak(kdbpeek(esp->iioq_head, task)))
	  {
	    esp->iioq_head = esp->iioq_tail;
	    esp->iioq_tail = esp->iioq_tail + 4;
	  }
 	kdbsplx(s);

	return(1);
}

/*
 * Print trap reason.
 */
kdbprinttrap(type, code)
	int	type, code;
{
	printf("kernel: ");
	if (type >= TRAP_TYPES)
	    printf("type %d", type);
	else
	    printf("%s", kdb_trap_names[type]);
	printf(" trap, frame @ %x\n", code);
}


void
thread_kdb_return()
{
        if (kdb_trap(&current_thread()->pcb->ss, 1)) {
                thread_exception_return();
                /*NOTREACHED*/
        }
}

/*
 * Peek/Poke routines
 */
extern unsigned kdbgetmem();/*forward*/

unsigned
kdbpeek(addr, task)
	unsigned int *addr;
	task_t   task;
{
	if (task == TASK_NULL && 
	    (task = db_current_task()) == TASK_NULL) {
		db_printf("\nbad address %x\n", addr);
		db_error(0);
		/* NOTREACHED */
	}
	(void)kdb_vtophys(task->map, addr, &addr, TRUE);
	return kdbgetmem(addr);
}

kdbpoke(addr, value, task)
	unsigned int *addr;
	unsigned int value;
	task_t   task;
{
  	if (task == TASK_NULL && 
	      (task = db_current_task()) == TASK_NULL) {
		db_printf("\nbad address %x\n", addr);
		db_error(0);
		/* NOTREACHED */
	}	
	(void)kdb_vtophys(task->map, addr, &addr, TRUE);
	return kdbputmem(addr, value);
}

/*
 *   kdb_vtophys - map addresses by hand, so that TLB is
 *   minimally affected.  But we do take faults, if necessary.
 */
boolean_t db_no_vm_fault = TRUE;

kdb_vtophys(map, addr, paddr, no_ret_error)
	vm_map_t map;
	unsigned int *addr;
	unsigned int **paddr;
	boolean_t no_ret_error;
{
	pmap_t		pmap;
	vm_offset_t	pa;	/* physical address */
	int		ret;

	if (map == VM_MAP_NULL) {
		if (no_ret_error)
			db_error("No VM map");
			/* NOTREACHED */
		return(-1);
	}
	pmap = map->pmap;
retry:
	pa = pmap_extract(pmap, addr);
	if (pa == 0) {
		if (!no_ret_error)
			return(-1);
		if (db_no_vm_fault)
			db_error("Paged out page");

		/* NOTE: we might get hung if map is locked!! */
		ret = vm_fault(map, trunc_page(addr), VM_PROT_READ, FALSE,
			       FALSE, (void (*)()) 0);
		if (ret != KERN_SUCCESS) {
			db_error("Invalid virtual address");
			/* NOTREACHED */
		}
		goto retry;
	}
	*paddr = (unsigned *)pa;
	return(0);
}


/*
 *   kdbgetmem - read long word from kernel address space
 */
unsigned
kdbgetmem(addr)
	unsigned char	*addr;
{
	unsigned 		temp = 0;
	int			i;
	extern unsigned		kdb_getiomem();

	if (addr == 0)
		return(-1);

	if (mem_size && addr > mem_size) {
		db_error("Beyond physical memory");
		return(-1);
	}

#if	BYTE_MSF
	for (i = 0; i <= 3; i++) {
#else	BYTE_MSF
	for (i = 3; i >= 0; i--) {
#endif	BYTE_MSF
		temp <<= 8;
		temp |= addr[i] & 0xFF;
	}
	return temp;
}


/*
 *  kdbputmem - write word to kernel address space
 */

kdbputmem(addr, value)
	unsigned char	*addr;
	unsigned	 value;
{
	int		 i;
	extern		 kdb_putiomem();

	if (addr == 0)
		return 0;

	if (mem_size && addr > mem_size)
		return 0;

#if	BYTE_MSF
	for (i = 3; i >= 0; i--) {
#else	BYTE_MSF
	for (i = 0; i < 4; i++) {
#endif	BYTE_MSF
		*addr++ = value & 0xFF;
		value >>= 8;
	}
	/* Might be text, flush I-cache */
	ficache(0, (unsigned)(addr - 4) & ~0x3, 2 * sizeof(int));
}

/*
 * Read bytes (??from kernel address space??) for debugger.
 */

void
db_read_bytes(addr, size, data, task)
	register unsigned *addr;
	register int	   size;
	register unsigned *data;
	task_t   task;
{
	while (size >= 4)
		*data++ = kdbpeek(addr++, task), size -= 4;

	if (size) {
		unsigned tmp;
		register char *dst = (char*)data;

		tmp = kdbpeek(addr, task);
		while (size--) {
			*dst++ = tmp & 0xFF;
			tmp >>= 8;
		}
	}
}

/*
 * Write bytes (?? to kernel address space??) for debugger.
 */
void
db_write_bytes(addr, size, data, task)
	register unsigned *addr;
	register int	   size;
	register unsigned *data;
	task_t   task;
{
	register char	*dst;

	while (size >= 4)
		kdbpoke(addr++, *data++, task), size -= 4;
	if (size) {
		unsigned tmp = kdbpeek(addr, task), tmp1 = 0;
		register char *src = (char*)data;

		tmp >>= (size << 3);
		tmp <<= (size << 3);
		while (size--) {
			tmp1 <<= 8;
			tmp1 |= *src++;
		}
		kdbpoke(addr, tmp|tmp1, task);
	}
}

boolean_t
db_check_access(addr, size, task)
	vm_offset_t	addr;
	register int	size;
	task_t		task;
{
	register	n;
	unsigned	*paddr;
	vm_map_t	map;
	
	if (DB_VALID_KERN_ADDR(addr)) {
	        map = kernel_map;
	} else {
		if (task == TASK_NULL && 
		    (task = db_current_task()) == TASK_NULL)
			return(FALSE);
		map = task->map;
	}

	while (size > 0) {
#if 0
		if (ISA_K0SEG(addr)) {
			if (mem_size && K0SEG_TO_PHYS(addr+size) > mem_size)
				return(FALSE);
			break;
		}
#endif
		if (kdb_vtophys(map, (unsigned *)addr, &paddr, FALSE) < 0)
			return(FALSE);
		n = trunc_page(addr + HP800_PGBYTES) - addr;
		if (n > size)
			n = size;
		size -= n;
		addr += n;
	}
	return(TRUE);
}

boolean_t
db_phys_eq(task1, addr1, task2, addr2)
	task_t		task1;
	vm_offset_t	addr1;
	task_t		task2;
	vm_offset_t	addr2;
{
	unsigned	*paddr1, *paddr2;

	if (addr1 >= VM_MAX_ADDRESS || addr2 >= VM_MAX_ADDRESS)
		return(FALSE);
	if (task1 == TASK_NULL || task1->map == 0 ||
		 task2 == TASK_NULL || task2->map == 0)
		return(FALSE);
	if (kdb_vtophys(task1->map, (unsigned *)addr1, &paddr1, FALSE) < 0 ||
	    kdb_vtophys(task2->map, (unsigned *)addr2, &paddr2, FALSE) < 0)
		return(FALSE);
	return(paddr1 == paddr2);
}

#define DB_USER_STACK_ADDR		(VM_MAX_ADDRESS-HP800_PGBYTES)
#define DB_NAME_SEARCH_LIMIT		(DB_USER_STACK_ADDR-(HP800_PGBYTES*3))

static int
db_search_null(task, svaddr, evaddr, skaddr, flag)
	task_t	 task;
	unsigned *svaddr;
	unsigned evaddr;
	unsigned **skaddr;
	int	 flag;
{
	register unsigned vaddr;
	register unsigned *kaddr;

	kaddr = *skaddr;
	for (vaddr = *svaddr; vaddr > evaddr; vaddr -= sizeof(unsigned)) {
		if (vaddr % HP800_PGBYTES == 0) {
			vaddr -= sizeof(unsigned);
			if (kdb_vtophys(task->map, vaddr, skaddr, FALSE) < 0)
				return(-1);
			kaddr = *skaddr;
		} else {
			vaddr -= sizeof(unsigned);
			kaddr--;
		}
		if ((*kaddr == 0) ^ (flag  == 0)) {
			*svaddr = vaddr;
			*skaddr = kaddr;
			return(0);
		}
	}
	return(-1);
}

void
db_task_name(task)
	task_t		task;
{
  register int n = DB_TASK_NAME_LEN - 15;
  db_printf("TASK 0x%8x", task);

  while (n-- >= 0)	/* compare with >= 0 for one more space */
    db_printf(" ");
}


/*
 * db_disasm: interface into the instruction disassembler.
 *
 */
db_addr_t db_disasm(db_addr_t iaddr, 
		    boolean_t Use_special_format, 
		    task_t task)
{
  db_addr_t ret;

  ret = print_instruction(kdbpeek(&iaddr, task), iaddr);

  return (iaddr+ret);
}


