/*
 *  (c) Copyright 1987 HEWLETT-PACKARD COMPANY
 *
 *  To anyone who acknowledges that this file is provided "AS IS"
 *  without any express or implied warranty:
 *      permission to use, copy, modify, and distribute this file
 *  for any purpose is hereby granted without fee, provided that
 *  the above copyright notice and this notice appears in all
 *  copies, and that the name of Hewlett-Packard Company not be
 *  used in advertising or publicity pertaining to distribution
 *  of the software without specific, written prior permission.
 *  Hewlett-Packard Company makes no representations about the
 *  suitability of this software for any purpose.
 */

/* UNISRC_ID	@(#)37.4   87/03/06  */
/* HPUX_ID: @(#)filehdr.h	37.4     87/02/19 */

/*****************************************************************************
*                               					     *
*  HP9000s800 Object Modules (Program files) consist of:                     *
*                                                                            *
*     Header                                                                 *
*     Auxiliary header(s)                                                    *
*     Space dictionary                                                       *
*     Subspace dictionary                                                    *
*     Loader fixup dictionary                                                *
*     Space/Subspace string area for space/subspace dictionary               *
*     Initialization pointers                                                *
*     Data pages                                                             *
*     Compilation unit dictionary                                            *
*     Symbol dictionary                                                      *
*     Fixup information                                                      *
*     Compilation unit/symbol string area for comp. unit/symbol dictionary   *
*     Data for unloadable spaces                  			     *
*                                                                            *
*  A word of every block which is associated with a name is                  *
*  the offset within the corresponding string area of                        *
*  the beginning of the name (word aligned).                                 *
*  The name is length-preceeded , \0 terminated and padded                   *
*  to a word boundary.                                                       *
*  When the blocks are in memory the offset will be                          *
*  replaced with a pointer to the string.                                    *
*                                                                            *
*  A file location entry of zero always means that the entry in question     *
*  is absent.                                                                *
*                                                                            *
*  An index number of -1 indicates no entries                                *
*                                                                            *
******************************************************************************/

/*              union definition for pointer to names                        */

union name_pt {
                char        *n_name;
                unsigned int n_strx;
              };

/*              struct definition for the system clock                       */

struct sys_clock { 
                   unsigned int secs; 
                   unsigned int nanosecs; 
                 };

/* macro definitions to give simple names to union structure components */

#define NAME_PT name.n_name
#define STR_INDEX name.n_strx


/************************************************************************/
/*                              					*/
/*      HP9000s800 magic number, system id and SOM version        	*/
/*                              					*/
/* 	The magic numbers defined for HP9000s800 are:     	 	*/

/* These definitions correspond with magic.h */
#define RELOC_MAGIC	0x106		/* relocatable only */
#define EXEC_MAGIC	0x107		/* normal executable */
#define SHARE_MAGIC	0x108		/* shared executable */
#define	DEMAND_MAGIC	0x10B		/* demand-loadable executable */

#define HP9000S800_ID	0x020B 
#define VERSION_ID	85082112



/***************************************************************************
*               the header consists of a single record                     *
****************************************************************************/

struct header{
        short int system_id;                 /* magic number - system        */
        short int a_magic;                   /* magic number - file type     */
        unsigned int version_id;             /* version id; format= YYMMDDHH */
        struct sys_clock file_time;          /* system clock- zero if unused */
        unsigned int entry_space;            /* index of space containing 
                                                entry point                  */
        unsigned int entry_subspace;         /* index of subspace for 
                                                entry point                  */
        unsigned int entry_offset;           /* offset of entry point        */
        unsigned int aux_header_location;    /* auxiliary header location    */
        unsigned int aux_header_size;        /* auxiliary header size        */
        unsigned int som_length;             /* length in bytes of entire som*/
        unsigned int presumed_dp;            /* DP value assumed during 
						compilation 		     */
        unsigned int space_location;         /* location in file of space 
                                                dictionary                   */
        unsigned int space_total;            /* number of space entries      */
        unsigned int subspace_location;      /* location of subspace entries */
        unsigned int subspace_total;         /* number of subspace entries   */
        unsigned int loader_fixup_location;  /* space reference array        */
        unsigned int loader_fixup_total;     /* total number of space 
                                                reference records            */
        unsigned int space_strings_location; /* file location of string area
                                                for space and subspace names */
        unsigned int space_strings_size;     /* size of string area for space 
                                                 and subspace names          */
        unsigned int init_array_location;    /* location in file of 
                                                initialization pointers      */
        unsigned int init_array_total;       /* number of init. pointers     */
        unsigned int compiler_location;      /* location in file of module 
                                                dictionary                   */
        unsigned int compiler_total;         /* number of modules            */
        unsigned int symbol_location;        /* location in file of symbol 
                                                dictionary                   */
        unsigned int symbol_total;           /* number of symbols            */
        unsigned int fixup_request_location; /* location in file of fix_up 
                                                requests                     */
        unsigned int fixup_request_total;    /* number of fixup requests     */
        unsigned int symbol_strings_location;/* file location of string area 
                                                for module and symbol names  */
        unsigned int symbol_strings_size;    /* size of string area for 
                                                module and symbol names      */
        unsigned int unloadable_sp_location; /* byte offset of first byte of 
                                                data for unloadable spaces   */
        unsigned int unloadable_sp_size;     /* byte length of data for 
                                                unloadable spaces            */
        unsigned int checksum;
};

#define FILHDR	struct header
#define FILHSZ	sizeof(FILHDR)
