/*
 * This file copied by Patrick Dalton from the mips machine dependent code.
 * It is being modified for the HPPA
 */
/* 
 * Mach Operating System
 * Copyright (c) 1991,1990,1989,1988,1987 Carnegie Mellon University
 * Copyright (c) 1992,1994 The University of Utah and
 * the Computer Systems Laboratory (CSL).
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 * 
 * CARNEGIE MELLON, THE UNIVERSITY OF UTAH AND CSL ALLOW FREE USE OF
 * THIS SOFTWARE IN ITS "AS IS" CONDITION, AND DISCLAIM ANY LIABILITY
 * OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF
 * THIS SOFTWARE.
 * 
 * Carnegie Mellon requests users of this software to return to
 * 
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 * 
 * any improvements or extensions that they make and grant Carnegie Mellon
 * the rights to redistribute these changes.
 *
 * CSL requests users of this software to return to csl-dist@cs.utah.edu any
 * improvements that they make and grant CSL redistribution rights.
 *
 * 	Utah $Hdr: db_trace.c 1.1 94/07/21$
 */
/*
 * HISTORY
 * $Log: db_trace.c,v $
 * Revision 1.2  1994/12/14  19:39:56  law
 * Copyright stuff
 *
 * Revision 1.1  1994/07/21  18:29:03  mike
 * original version
 *
 * Revision 2.10  91/10/09  16:14:11  af
 * 	Fixed db_trace_cproc.
 * 	[91/09/19            rpd]
 * 
 * Revision 2.9.3.1  91/10/05  13:12:51  jeffreyh
 * 	Supported stack trace and register access of non current thread.
 * 	Supported user register access.
 * 	Fixed incorrect register reference.
 * 	Fixed guess_procedure to limit search and skip leading NOPs.
 * 	[91/09/05            tak]
 * 
 * Revision 2.9  91/05/14  17:33:59  mrt
 * 	Correcting copyright
 * 
 * Revision 2.8  91/03/16  14:55:54  rpd
 * 	Fixed db_trace_cproc to handle wired cthreads.
 * 	Fixed db_cproc_state.
 * 	[91/03/10            rpd]
 * 	Fixed tracing of swapped threads.
 * 	Removed unnecessary uses of db_get_value.
 * 	[91/03/01  17:49:16  rpd]
 * 
 * Revision 2.7  91/02/14  14:37:26  mrt
 * 	Split trace function from ddb interface.  Added (optional)
 * 	code to walk through a user-mode, optimized CThread program
 * 	like the U*x server.  Should make it MI someday.
 * 	[91/02/12  12:18:59  af]
 * 
 * Revision 2.6  91/02/05  17:48:14  mrt
 * 	Added author notices
 * 	[91/02/04  11:22:18  mrt]
 * 
 * 	Changed to use new Mach copyright
 * 	[91/02/02  12:25:54  mrt]
 * 
 * Revision 2.5  91/01/09  19:56:45  rpd
 * 	Fixed stack tracing for threads without kernel stacks.
 * 	[91/01/09            rpd]
 * 
 * Revision 2.4  91/01/08  15:49:38  rpd
 * 	Quit tracing after hitting a zero pc.
 * 	[90/12/31            rpd]
 * 	Changed for new pcb organization.
 * 	[90/11/12            rpd]
 * 
 * Revision 2.3  90/09/09  14:33:21  rpd
 * 	Fixed trace when hitting a breakpoint in user space.
 * 	Fixed to trace crossing over to user space for any
 * 	thread (e.g. "t/tu <thread>" command).
 * 	Documented magic adjustments of pc and ra in various places.
 * 	Locals work, provided you either do not strip or use
 * 	my fixed version of xstrip.
 * 	[90/08/30  16:56:46  af]
 * 
 * Revision 2.2  90/08/27  22:07:51  dbg
 * 	Fix some sloppy type (non)-declarations.
 * 	[90/08/27            dbg]
 * 
 * 	Got rid of ddb_regs: all lives on the stack now.
 * 	BEWARE: to make this possible the "valuep" field in
 * 	the db_regs array is _not_ a pointer, but a displacement
 * 	off the base of the exception frame.  Use _exclusively_
 * 	the (MI) db_read/write_variable functions to access registers.
 * 	Updated db_getreg_val accordingly.
 * 	Cleaned up a lot the trace functions, I hope they still work
 * 	without a valid symtab...
 * 	Made apparent that we cannot get a stack trace given _only_
 * 	a stack pointer, as the stack is not self-describing.
 * 	Not sure why local variables do not come out, xstrip problem ?
 * 	[90/08/20  10:09:54  af]
 * 
 * 	Turn debug off
 * 	[90/08/18  00:17:56  af]
 * 
 * 	Created, from my old KDB code.  History summary:
 * 		From jsb: better stack traces.
 * 		[90/04/23            rpd]
 * 		Changed tracing code to reflect 11 arguments for all traps.
 * 		[90/05/23            rwd]
 * 		Switch over to the MI KDB, fixes and cleanups.
 * 		[90/01/20            af]
 * 		Created.
 * 		[89/08/08            af]
 * 
 */
/*
 *	File: db_trace.c
 * 	Author: Alessandro Forin, Carnegie Mellon University
 *	Date:	8/89
 *
 *	Stack traceback routines for MIPS
 */

#include <mach/boolean.h>
#include <machine/db_machdep.h>
#include <kern/thread.h>
#include <kern/task.h>

#include <ddb/db_variables.h>
#include <ddb/db_task_thread.h>

#include <machine/db_trace.h>
#include <machine/cpu.h>
#include <machine/psw.h>
#include <machine/db_symtab.h>
#include <mach/vm_param.h>

extern int start();	/* lowest kernel code address */
extern unsigned db_maxoff;

#define IS_DOUBLE_KERNEL_DATA(offset) \
  ((offset) >= ((int)&(((struct hp800_kernel_state *)0)->fr12)))

#define IS_DOUBLE_USER_DATA(offset) \
  (((offset) >= ((int)&(((struct hp800_saved_state  *)0)->fr0 ))) && \
   ((offset) <= ((int)&(((struct hp800_saved_state  *)0)->fr31))))

#define IS_DOUBLE_SIZED_DATA(offset, use_kreg) \
  ((use_kreg) ? IS_DOUBLE_KERNEL_DATA(offset) : IS_DOUBLE_USER_DATA(offset))

extern	db_sym_t	localsym();

/*
 * Machine register set.
 */
struct hp800_saved_state *db_cur_exc_frame = 0;

static
db_setf_regs(vp, valuep, op, ap)
	struct db_variable	*vp;
	db_expr_t		*valuep;
	int			op;
	db_var_aux_param_t	ap;
{
  register		*regp = 0;
  register                reg_offset;
  int                     kreg_offset;
  boolean_t               valid_kreg = FALSE;
  int			null_reg = 0;
  register thread_t	thread = ap->thread;
	
  if (db_option(ap->modif, 'u')) 
    {
      if (thread == THREAD_NULL) 
	{ 
	  if ((thread = current_thread()) == THREAD_NULL)
	    db_error("no user registers\n");
	  /* NOTREACHED */
	}
      if (thread == current_thread() && (USERMODE(db_cur_exc_frame)))
	{
	  regp       = (int *)db_cur_exc_frame;
	  reg_offset = (int) (vp->valuep);
	}
    } 
  else 
    {
      if (thread == THREAD_NULL || thread == current_thread()) 
	{
	  regp       = (int *) db_cur_exc_frame;
	  reg_offset =  (int) (vp->valuep);
	} 
      else if ((thread->state & TH_SWAPPED) == 0 &&
	       thread->kernel_stack) 
	{
	  kreg_offset = kern_save_reg_idx(vp, &valid_kreg);
	  if (valid_kreg)
	    {
	      regp       = (int *) KERNEL_STATE(thread->kernel_stack);
	      reg_offset = kreg_offset;
	    }
	  else
	    {
	      regp       = &null_reg;
	      reg_offset = 0;
	    }
	} 
      else if ((thread->state & TH_SWAPPED) &&
	       thread->swap_func != (void (*)())thread_exception_return) 
	{
	  /* only PC is valid */
	  if (vp->valuep == 
	      (int *)&(((struct hp800_saved_state *)0)->iioq_head)) 
	    {
	      regp       = (int *) (&thread->swap_func);
	      reg_offset = 0;
	    } 
	  else 
	    {
	      regp       = &null_reg;
	      reg_offset = 0;
	    }
	}
    }
  if (regp == 0)
    {
      if (thread->pcb == 0)
	db_error("no pcb\n");
      regp       = (int *)(&thread->pcb->ss);
      reg_offset = (int)(vp->valuep);
    }

  regp = (int *)((char *)regp + reg_offset);
  if (IS_DOUBLE_SIZED_DATA(reg_offset, valid_kreg))
    {
      if (op == DB_VAR_GET)
	*(double *)valuep = *(double *)regp;
      else if (op == DB_VAR_SET)
	*(double *)regp = *(double *)valuep;
    }
  else
    {
      if (op == DB_VAR_GET)
	*valuep = *regp;
      else if (op == DB_VAR_SET)
	*regp = *valuep;
    }
  return(0);
}


struct db_variable db_regs[] = {
  { "flags", (int*)&(((struct hp800_saved_state *)0)->flags), db_setf_regs },
  { "r1", (int*)&(((struct hp800_saved_state *)0)->r1), db_setf_regs },
  { "r2", (int*)&(((struct hp800_saved_state *)0)->r2), db_setf_regs },
  { "r3", (int*)&(((struct hp800_saved_state *)0)->r3), db_setf_regs },
  { "r4", (int*)&(((struct hp800_saved_state *)0)->r4), db_setf_regs },
  { "r5", (int*)&(((struct hp800_saved_state *)0)->r5), db_setf_regs },
  { "r6", (int*)&(((struct hp800_saved_state *)0)->r6), db_setf_regs },
  { "r7", (int*)&(((struct hp800_saved_state *)0)->r7), db_setf_regs },
  { "r8", (int*)&(((struct hp800_saved_state *)0)->r8), db_setf_regs },
  { "r9", (int*)&(((struct hp800_saved_state *)0)->r9), db_setf_regs },
  { "r10", (int*)&(((struct hp800_saved_state *)0)->r10), db_setf_regs },
  { "r11", (int*)&(((struct hp800_saved_state *)0)->r11), db_setf_regs },
  { "r12", (int*)&(((struct hp800_saved_state *)0)->r12), db_setf_regs },
  { "r13", (int*)&(((struct hp800_saved_state *)0)->r13), db_setf_regs },
  { "r14", (int*)&(((struct hp800_saved_state *)0)->r14), db_setf_regs },
  { "r15", (int*)&(((struct hp800_saved_state *)0)->r15), db_setf_regs },
  { "r16", (int*)&(((struct hp800_saved_state *)0)->r16), db_setf_regs },
  { "r17", (int*)&(((struct hp800_saved_state *)0)->r17), db_setf_regs },
  { "r18", (int*)&(((struct hp800_saved_state *)0)->r18), db_setf_regs },
  { "r19", (int*)&(((struct hp800_saved_state *)0)->r19), db_setf_regs },
  { "r20", (int*)&(((struct hp800_saved_state *)0)->r20), db_setf_regs },
  { "r21", (int*)&(((struct hp800_saved_state *)0)->r21), db_setf_regs },
  { "r22", (int*)&(((struct hp800_saved_state *)0)->r22), db_setf_regs },
  { "r23", (int*)&(((struct hp800_saved_state *)0)->r23), db_setf_regs },
  { "r24", (int*)&(((struct hp800_saved_state *)0)->r24), db_setf_regs },
  { "r25", (int*)&(((struct hp800_saved_state *)0)->r25), db_setf_regs },
  { "r26", (int*)&(((struct hp800_saved_state *)0)->r26), db_setf_regs },
  { "r27", (int*)&(((struct hp800_saved_state *)0)->r27), db_setf_regs },
  { "r28", (int*)&(((struct hp800_saved_state *)0)->r28), db_setf_regs },
  { "r29", (int*)&(((struct hp800_saved_state *)0)->r29), db_setf_regs },
  { "r30", (int*)&(((struct hp800_saved_state *)0)->r30), db_setf_regs },
  { "r31", (int*)&(((struct hp800_saved_state *)0)->r31), db_setf_regs },
  { "cr11", (int*)&(((struct hp800_saved_state *)0)->cr11), db_setf_regs },
  { "iioq_head", (int*)&(((struct hp800_saved_state *)0)->iioq_head), db_setf_regs },
  { "iisq_head", (int*)&(((struct hp800_saved_state *)0)->iisq_head), db_setf_regs },
  { "iioq_tail", (int*)&(((struct hp800_saved_state *)0)->iioq_tail), db_setf_regs },
  { "iisq_tail", (int*)&(((struct hp800_saved_state *)0)->iisq_tail), db_setf_regs },
  { "cr15", (int*)&(((struct hp800_saved_state *)0)->cr15), db_setf_regs },
  { "cr19", (int*)&(((struct hp800_saved_state *)0)->cr19), db_setf_regs },
  { "cr20", (int*)&(((struct hp800_saved_state *)0)->cr20), db_setf_regs },
  { "cr21", (int*)&(((struct hp800_saved_state *)0)->cr21), db_setf_regs },
  { "cr22", (int*)&(((struct hp800_saved_state *)0)->cr22), db_setf_regs },
  { "bsd_goto", (int*)&(((struct hp800_saved_state *)0)->bsd_goto), db_setf_regs },
  { "sr4", (int*)&(((struct hp800_saved_state *)0)->sr4), db_setf_regs },
  { "sr0", (int*)&(((struct hp800_saved_state *)0)->sr0), db_setf_regs },
  { "sr1", (int*)&(((struct hp800_saved_state *)0)->sr1), db_setf_regs },
  { "sr2", (int*)&(((struct hp800_saved_state *)0)->sr2), db_setf_regs },
  { "sr3", (int*)&(((struct hp800_saved_state *)0)->sr3), db_setf_regs },
  { "sr5", (int*)&(((struct hp800_saved_state *)0)->sr5), db_setf_regs },
  { "sr6", (int*)&(((struct hp800_saved_state *)0)->sr6), db_setf_regs },
  { "sr7", (int*)&(((struct hp800_saved_state *)0)->sr7), db_setf_regs },
  { "cr0", (int*)&(((struct hp800_saved_state *)0)->cr0), db_setf_regs },
  { "cr8", (int*)&(((struct hp800_saved_state *)0)->cr8), db_setf_regs },
  { "cr9", (int*)&(((struct hp800_saved_state *)0)->cr9), db_setf_regs },
  { "cr10", (int*)&(((struct hp800_saved_state *)0)->cr10), db_setf_regs },
  { "cr12", (int*)&(((struct hp800_saved_state *)0)->cr12), db_setf_regs },
  { "cr13", (int*)&(((struct hp800_saved_state *)0)->cr13), db_setf_regs },
  { "cr24", (int*)&(((struct hp800_saved_state *)0)->cr24), db_setf_regs },
  { "cr25", (int*)&(((struct hp800_saved_state *)0)->cr25), db_setf_regs },
  { "cr26", (int*)&(((struct hp800_saved_state *)0)->cr26), db_setf_regs },
  { "mpsfu_high", (int*)&(((struct hp800_saved_state *)0)->mpsfu_high), db_setf_regs },
  { "mpsfu_low", (int*)&(((struct hp800_saved_state *)0)->mpsfu_low), db_setf_regs },
  { "mpsfu_ovflo", (int*)&(((struct hp800_saved_state *)0)->mpsfu_ovflo), db_setf_regs },
  { "pad", (int*)&(((struct hp800_saved_state *)0)->pad), db_setf_regs },
  { "fr0", (int*)&(((struct hp800_saved_state *)0)->fr0), db_setf_regs },
  { "fr1", (int*)&(((struct hp800_saved_state *)0)->fr1), db_setf_regs },
  { "fr2", (int*)&(((struct hp800_saved_state *)0)->fr2), db_setf_regs },
  { "fr3", (int*)&(((struct hp800_saved_state *)0)->fr3), db_setf_regs },
  { "fr4", (int*)&(((struct hp800_saved_state *)0)->fr4), db_setf_regs },
  { "fr5", (int*)&(((struct hp800_saved_state *)0)->fr5), db_setf_regs },
  { "fr6", (int*)&(((struct hp800_saved_state *)0)->fr6), db_setf_regs },
  { "fr7", (int*)&(((struct hp800_saved_state *)0)->fr7), db_setf_regs },
  { "fr8", (int*)&(((struct hp800_saved_state *)0)->fr8), db_setf_regs },
  { "fr9", (int*)&(((struct hp800_saved_state *)0)->fr9), db_setf_regs },
  { "fr10", (int*)&(((struct hp800_saved_state *)0)->fr10), db_setf_regs },
  { "fr11", (int*)&(((struct hp800_saved_state *)0)->fr11), db_setf_regs },
  { "fr12", (int*)&(((struct hp800_saved_state *)0)->fr12), db_setf_regs },
  { "fr13", (int*)&(((struct hp800_saved_state *)0)->fr13), db_setf_regs },
  { "fr14", (int*)&(((struct hp800_saved_state *)0)->fr14), db_setf_regs },
  { "fr15", (int*)&(((struct hp800_saved_state *)0)->fr15), db_setf_regs },
  { "fr16", (int*)&(((struct hp800_saved_state *)0)->fr16), db_setf_regs },
  { "fr17", (int*)&(((struct hp800_saved_state *)0)->fr17), db_setf_regs },
  { "fr18", (int*)&(((struct hp800_saved_state *)0)->fr18), db_setf_regs },
  { "fr19", (int*)&(((struct hp800_saved_state *)0)->fr19), db_setf_regs },
  { "fr20", (int*)&(((struct hp800_saved_state *)0)->fr20), db_setf_regs },
  { "fr21", (int*)&(((struct hp800_saved_state *)0)->fr21), db_setf_regs },
  { "fr22", (int*)&(((struct hp800_saved_state *)0)->fr22), db_setf_regs },
  { "fr23", (int*)&(((struct hp800_saved_state *)0)->fr23), db_setf_regs },
  { "fr24", (int*)&(((struct hp800_saved_state *)0)->fr24), db_setf_regs },
  { "fr25", (int*)&(((struct hp800_saved_state *)0)->fr25), db_setf_regs },
  { "fr26", (int*)&(((struct hp800_saved_state *)0)->fr26), db_setf_regs },
  { "fr27", (int*)&(((struct hp800_saved_state *)0)->fr27), db_setf_regs },
  { "fr28", (int*)&(((struct hp800_saved_state *)0)->fr28), db_setf_regs },
  { "fr29", (int*)&(((struct hp800_saved_state *)0)->fr29), db_setf_regs },
  { "fr30", (int*)&(((struct hp800_saved_state *)0)->fr30), db_setf_regs },
  { "fr31", (int*)&(((struct hp800_saved_state *)0)->fr31), db_setf_regs },
  { "prevssp", (int*)&(((struct hp800_saved_state *)0)->prevssp), db_setf_regs }
};
struct db_variable *db_eregs = db_regs + sizeof(db_regs)/sizeof(db_regs[0]);

struct db_variable db_kregs[] = {
  { "pcb", (int*)&(((struct hp800_kernel_state *)0)->pcb), db_setf_regs },
  { "r2", (int*)&(((struct hp800_kernel_state *)0)->r2), db_setf_regs },
  { "r3", (int*)&(((struct hp800_kernel_state *)0)->r3), db_setf_regs },
  { "r4", (int*)&(((struct hp800_kernel_state *)0)->r4), db_setf_regs },
  { "r5", (int*)&(((struct hp800_kernel_state *)0)->r5), db_setf_regs },
  { "r6", (int*)&(((struct hp800_kernel_state *)0)->r6), db_setf_regs },
  { "r7", (int*)&(((struct hp800_kernel_state *)0)->r7), db_setf_regs },
  { "r8", (int*)&(((struct hp800_kernel_state *)0)->r8), db_setf_regs },
  { "r9", (int*)&(((struct hp800_kernel_state *)0)->r9), db_setf_regs },
  { "r10", (int*)&(((struct hp800_kernel_state *)0)->r10), db_setf_regs },
  { "r11", (int*)&(((struct hp800_kernel_state *)0)->r11), db_setf_regs },
  { "r12", (int*)&(((struct hp800_kernel_state *)0)->r12), db_setf_regs },
  { "r13", (int*)&(((struct hp800_kernel_state *)0)->r13), db_setf_regs },
  { "r14", (int*)&(((struct hp800_kernel_state *)0)->r14), db_setf_regs },
  { "r15", (int*)&(((struct hp800_kernel_state *)0)->r15), db_setf_regs },
  { "r16", (int*)&(((struct hp800_kernel_state *)0)->r16), db_setf_regs },
  { "r17", (int*)&(((struct hp800_kernel_state *)0)->r17), db_setf_regs },
  { "r18", (int*)&(((struct hp800_kernel_state *)0)->r18), db_setf_regs },
  { "r30", (int*)&(((struct hp800_kernel_state *)0)->r30), db_setf_regs },
  { "cr0", (int*)&(((struct hp800_kernel_state *)0)->cr0), db_setf_regs },
  { "cr8", (int*)&(((struct hp800_kernel_state *)0)->cr8), db_setf_regs },
  { "fr12", (int*)&(((struct hp800_kernel_state *)0)->fr12), db_setf_regs },
  { "fr13", (int*)&(((struct hp800_kernel_state *)0)->fr13), db_setf_regs },
  { "fr14", (int*)&(((struct hp800_kernel_state *)0)->fr14), db_setf_regs },
  { "fr15", (int*)&(((struct hp800_kernel_state *)0)->fr15), db_setf_regs }
};
struct db_variable *db_ekregs = db_kregs + sizeof(db_kregs)/sizeof(db_kregs[0]);
  
/*
 * Find kernel register for the varible given.
 */
static int *
kern_save_reg_idx(vp, is_valid)
     struct db_variable *vp;
     boolean_t *is_valid;
{
  register i;
  struct db_variable *kregp = db_kregs;
  
  *is_valid = TRUE;
  for (i=0; kregp != db_ekregs; kregp++, i++)
    if (strcmp(vp->name, kregp->name))
      return kregp->valuep;

  *is_valid = FALSE;
  return 0;
}

static struct hp800_saved_state hp800_reg_env;

struct hp800_saved_state *
db_get_reg_env(thread, user)
	thread_t	thread;
	boolean_t	user;
{
	register struct db_variable *dp;
	struct db_var_aux_param aux_param;

	aux_param.modif = (user)? "u": "";
	aux_param.level = 0;
	aux_param.thread = thread;
	for (dp = db_regs; dp < db_eregs; dp++) {
		db_read_write_variable(dp, 
				(int *)((int)&hp800_reg_env + (int)dp->valuep),
				DB_VAR_GET, &aux_param);
	}
	return(&hp800_reg_env);
}

/*
 * Stack trace.
 */

void db_hp800_stack_trace();

void
db_stack_trace_cmd(addr, have_addr, count, modif)
	db_expr_t	addr;
	int		have_addr;
	db_expr_t	count;
	char		*modif;
{
#define	F_KERNEL_ONLY	1
#define	F_TRACE_THREAD	2
#define	F_DO_LOCALS	4
#define	F_PRINT_FRAME	8
	unsigned	flags = F_KERNEL_ONLY;
	long int	the_pc, the_rp;
	register char 	c;
	thread_t	th;
	extern unsigned current_kstack;
	vm_offset_t kstackp, stackp, framep;

	while ((c = *modif++) != 0) {
	    if (c == 't')
		flags |= F_TRACE_THREAD;
	    if (c == 'u')
		flags &= ~F_KERNEL_ONLY;
	    if (c == 'l')
		flags |= F_DO_LOCALS;
	    if (c == 'e')
		flags |= F_PRINT_FRAME;
	}

	if (flags & F_TRACE_THREAD) {
	    if (have_addr) {
		th = (thread_t) addr;
		if (!db_check_thread_address_valid(th))
		    return;
	    } else {
		th = db_default_thread;
		if (th == THREAD_NULL)
		    th = current_thread();
		if (th == THREAD_NULL) {
		    db_printf("no active thread\n");
		    return;
		}
	    }
	    if (th->state & TH_SWAPPED) {
		register pcb_t pcb = th->pcb;
		db_sym_t sym;
		char *name;
		int proc_off;

		sym = db_search_task_symbol(th->swap_func, DB_STGY_PROC, 
						&proc_off, th->task);
		if (sym == DB_SYM_NULL) {
			/* no procedures ? wont be able to do much */
			db_find_xtrn_task_sym_and_offset(th->swap_func,
						    &name, &proc_off, th->task);
		} else {
			db_symbol_values(sym, &name, 0);
		}

		if (name == 0 || proc_off != 0) {
			db_printf("Continuation %x\n", th->swap_func);
		} else {
			db_printf("Continuation %s\n", name);
		}

		kstackp = 0;
		framep = pcb->ss.r4;
		stackp = pcb->ss.sp;
		the_pc = pcb->ss.iioq_head;
		the_rp = pcb->ss.rp;
	    } else {
		if (th == current_thread())
		    goto no_addr;

		kstackp = th->kernel_stack;
		framep = KERNEL_STATE(kstackp)->r4;
		stackp = KERNEL_STATE(kstackp)->sp;
		the_pc = KERNEL_STATE(kstackp)->rp; /* ??? */
		the_rp = KERNEL_STATE(kstackp)->rp;
	    }
	} else if (!have_addr) {
	no_addr:
	    th = current_thread();
	    kstackp = current_stack();
	    framep = db_cur_exc_frame->r4;
	    stackp = db_cur_exc_frame->sp;
	    the_pc = db_cur_exc_frame->iioq_head;
	    the_rp = db_cur_exc_frame->rp;
	} else {
	    db_error("You can only do this by fixing $sp, $pc and $ra");
	}
	db_hp800_stack_trace(count,framep,stackp,the_pc,the_rp,flags,kstackp,th);
}


void
db_hp800_stack_trace(count, framep, stackp, the_pc, the_rp, 
		     flags, kstackp, thread)
	unsigned int framep, stackp, kstackp;
	long int     the_pc, the_rp;
	thread_t     thread;
{
  static char	*ubanner = "*** User stack:\n";

  char		*name;
  db_sym_t	sym;
  unsigned int	argp;
  task_t        task = (thread == THREAD_NULL)? TASK_NULL: thread->task;
  unsigned	proc_pc, new_framep;

  int		narg, nloc;
  unsigned	minpc;
  boolean_t	top_frame = TRUE;
  struct frame_info f;
  int		proc_off;
  int 		Nest;
  struct hp800_saved_state *reg_env;

#if 0
  if (the_rp)
    the_rp -= 8;		/* backup jal */
#endif
  minpc = (unsigned)start;

  if (USERMODE_PC(the_pc)) {
    if (flags & F_KERNEL_ONLY)
      goto out;
    else {
      db_printf("%s", ubanner);
      minpc = VM_MIN_ADDRESS;
    }
    reg_env = db_get_reg_env(thread, TRUE);
  } else
    reg_env = db_get_reg_env(thread, FALSE);

  if (count == -1)
    count = 65535;

  while (count--) {
    int dummy;

    if (the_pc == 0)
      break;

    bzero(&f, sizeof f);

    /* 1: find out where we are */
    sym = db_search_task_symbol(the_pc,DB_STGY_PROC,&proc_off,task);
    if (sym == DB_SYM_NULL) {
      /* no procedures ? wont be able to do much */
      db_find_xtrn_task_sym_and_offset(the_pc, &name,
				       &proc_off, task);
      proc_pc = the_pc - proc_off;
    } else
      db_symbol_values(sym, &name, &proc_pc);

    if (name == 0 || proc_off > db_maxoff) {
      proc_off = -1;
      sym = 0;
      proc_pc = guess_procedure(the_pc, task);
      db_printf("%x(", proc_pc);
    } else {
      db_printf("%s(", name);
    }

    /* 2: find out more about it */
    get_frame_info(sym, framep, the_pc, &f, task, reg_env);
    nloc = f.nloc;
    narg = f.narg;
    
    argp = FRAME_ARGS(framep);

    the_rp = db_get_task_value(RETURN_POINTER(framep), 4, FALSE, task);
    if (the_rp < minpc)
      the_rp = 0;

    /* 2.5: Print arguments */
#if 0
    /* 
     *  Should check to see if the args have been saved to  the stack
     *  but for now we will assume they always are.
     */
    if (top_frame) {
      register int i;
      for (i = 0; narg && i < 4; i++, argp += 4) {
	if (sym = get_next_arg_sym(sym, &Nest, &sym)) {
	  db_symbol_values(sym,&name,0);
	  db_printf("%s=", name);
	}
	if (saves_arg(the_pc, REG_ARG(i), task)) {
	  db_printf("%x", db_get_task_value(argp,
					    4, FALSE, task));
	} else
	  db_printf("%x", 
		    getreg_val(REG_ARG(i), reg_env));
	if (--narg) {
	  db_printf(",");
	  sym++;
	}
      }
    }

#endif 0
    for (; narg--; argp += 4) {
      if (get_next_arg_sym(sym, &Nest, &sym)) {
	db_symbol_values(sym,&name, 0);
	db_printf("%s=", name);
      }
      db_printf("%x", db_get_task_value(argp,4,FALSE,task));
      if (narg)
	db_printf(",");
    }
    db_printf(")");

    /* 3: give more detail about where we are */
    if (proc_off != -1)
      db_printf("+%x", proc_off);
    if (f.filename)
      db_printf(" [%s:%d]", f.filename, f.linenum);
    db_printf("\n");

    /* 4: possibly print local vars */
    if (flags & F_DO_LOCALS) {
      unsigned w;
      boolean_t isReg;

      while (sym = get_next_local_sym(sym, &isReg, &Nest)) {
	if (Nest != 2)		/* 2 = top-level locals */
	  continue;

	db_symbol_values(sym, &name, &w);
	if (isReg)
	  w = getreg_val(w, reg_env);
	else
	  w = db_get_task_value(w + stackp, 4,
				FALSE, task);
	db_printf("\t%s=%x\n", name, w);
      }
    }

    /* 5: find out where we go after this */
#if 0
    db_restore_regs(reg_env, stackp, proc_pc, the_pc, task);
#endif

    if(STACK_BOTTOM(framep, task))
      goto out;

    new_framep = db_get_task_value(PREVIOUS_FRAME(framep), 4, FALSE, task);
    if (new_framep == 0)
      {
	struct hp800_saved_state *ess;

	db_printf("Exception frame found:\n");
	ess = (struct hp800_saved_state *)((unsigned char *)framep - FRAME_SIZE);
	dump_ssp(ess);

	the_pc = ess->rp;
	framep = ess->r4;
      }
    else
      {
	the_pc = the_rp;
	framep = new_framep;
      }
  }
 out:
  return;
}

/* 
 * Try to find a procedure beginning even though we don't
 * have nay information about it.
 */
guess_procedure(pc, task)
unsigned pc;
task_t	 task;
{
    unsigned	    w;
    int             bw_pc, fw_pc;

    if (pc < (unsigned)start)
    	fw_pc = VM_MIN_ADDRESS;
    else
        fw_pc = (unsigned)start;
    if (fw_pc + db_maxoff < pc)
	fw_pc = pc - db_maxoff;

    for (bw_pc = pc; bw_pc >= fw_pc; bw_pc -= 4) {
	w = db_get_task_value(bw_pc, 4, FALSE, task);
	if (inst_return(w))
	    break;
    }
    if (bw_pc < fw_pc)
	return(fw_pc);

    /*
     * skip all values that are zero
     */
    for (bw_pc += 8; bw_pc < pc; bw_pc += 4) {
	if (db_get_task_value(bw_pc, 4, FALSE, task))
	  break;
    }
    return(bw_pc);
}

int curproc;

get_frame_info(sym, fp, pc, f_info, task, reg_env)
	db_sym_t sym;
        vm_offset_t  fp, pc;
	frame_info_t f_info;
	task_t task;
	struct hp800_saved_state *reg_env;
{
  /* if we don't have any symbol information then we'll
   * just give some default information
   */
  if (sym == DB_SYM_NULL)
    {
      f_info->narg  = 4;
      f_info->nloc  = (reg_env->sp - fp)/4;
      f_info->filename = "";
      f_info->linenum  = 0;
    }
  else
    {
      f_info->narg = get_num_func_args(sym);
      f_info->nloc = 
	get_num_func_locals(sym, f_info->narg,(reg_env->sp - fp)/4);
      X_db_line_at_pc(db_last_symtab, sym, 
		      &(f_info->filename), &(f_info->linenum));
    }
  
  f_info->saved_pc_off = -20;
  
}
