/* 
 * Copyright (c) 1991,1994 The University of Utah and
 * the Computer Systems Laboratory at the University of Utah (CSL).
 * All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software is hereby
 * granted provided that (1) source code retains these copyright, permission,
 * and disclaimer notices, and (2) redistributions including binaries
 * reproduce the notices in supporting documentation, and (3) all advertising
 * materials mentioning features or use of this software display the following
 * acknowledgement: ``This product includes software developed by the
 * Computer Systems Laboratory at the University of Utah.''
 *
 * THE UNIVERSITY OF UTAH AND CSL ALLOW FREE USE OF THIS SOFTWARE IN ITS "AS
 * IS" CONDITION.  THE UNIVERSITY OF UTAH AND CSL DISCLAIM ANY LIABILITY OF
 * ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * CSL requests users of this software to return to csl-dist@cs.utah.edu any
 * improvements that they make and grant CSL redistribution rights.
 *
 * 	Utah $Hdr: pdc.s 1.5 94/12/14$
 *	Author: Bob Wheeler, University of Utah CSL
 */

/*
 *	Routines to do PDC and IODC calls when virtual memory is enabled
 */

/*
 *	pdc_iodc_call(int (*routine)(), unsigned pdc_flag, unsigned arg0, 
 *		      unsigned arg1, ... unsigned arg7);
 *
 *	routine is the entry point to the pdc or iodc routine.
 *	pdc_flag is 1 if this is a pdc call
 *	arg0, ... ,arg7 are passed to the pdc or iodc routine.
 *
 *	Assumptions:
 *	All return blocks must be in equivalently mapped memory and the call
 *	must be made in kernel mode. The PSW-Q bit must be on.
 */

#include <machine/asm.h>
#include <machine/pdc.h>
#include <machine/break.h>
#include <machine/psw.h>

	.code

	.export	pdc_call,entry
	.proc
	.callinfo
pdc_iodc_call
	.entry
	break	BREAK_KERNEL, BREAK_PDC_IODC_CALL
	nop
	nop
	bv,n	0(rp)
	.exit


/*
 *	pdc_break
 *
 *	In order to get the machine into a known state we issue a break 
 *	instruction and then jump here from the break trap handler.
 *	
 *	The requirements for PDC and IODC calls are:
 *
 *		PSW-Q bit is on, all others are off unless this is a 
 *		PDC_PIM_HPMC request in which case the PSW-M bit is on.
 *
 *	PDC and IODC routines adhere to the calling convention and thus
 *	we don't have to save or restore state for them.
 *
 *	The PDC needs 7K of stack space to operate.
 *
 *
 *	Assumptions:
 *	We assume that the routine pdc_iodc_call was invoked and that this
 *	routine did the break instruction to get us here. The Arguments to 
 *	pdc_iodc_call are:
 *
 *		pdc_iodc_call((int (*)()) routine, unsigned pdc_flag, 
 *			      unsigned arg0, unsigned arg1, 
 *			      unsigned arg2, ..., unsigned arg7);
 *
 *	routine is the address of the routine to call (either pdc or iodc 
 *	entry point). pdc_flag is 1 if this is a PDC call (used to check if the
 *	PSW-M bit should be set). ar0 through arg7 are the arguments to call
 *	the PDC or IODC routine with. 
 *
 *	Restrictions: 
 *	Can't take any interrupts or traps that can't be resolved in locore.
 *	(i.e. this routine is not reentrant) We also can't take an ITLB miss
 *	while the PSW-Q bit is off so this routine must all fit on one page.
 */

	.align	2048
	.export pdc_iodc_break,entry
pdc_iodc_break
	; When we arrive here we assume that all the registers are in their
	; inital state on entry into the kernel. We can use any caller saves
 	; registers since we know that we were invoked by a call to the routine
	; pdc_iodc_call().

	; Get us on the interrupt stack and see if we have 7K of stack space
	; for the call.

	ldil	L%tmp_saved_state,t1
	ldil	L%isp,t2
	ldw	R%tmp_saved_state(t1),t1
	stw	rp,TSS_RP(t1)
	stw	r0,TSS_FLAGS(t1)
	ldo	R%isp(t2),t3
	stw	sp,TSS_SP
	copy	sp,ret0
	comib,=,n 0,t3,pdc_iodc_break_on_istack

	; We're on a kernel stack, put us on the interrupt stack

	ldi	SS_SWITCH,t4
	copy	t3,sp
	stw	t4,TSS_FLAGS(t1)
	stw	r0,R%isp(t2)

pdc_iodc_break_on_istack

	; now see if there is 7K left on the interrupt stack for this call.
	; Add 64 bytes for the call frame.

	ldo	7232(sp),t2
	ldil	L%interrup_stack,t3
	ldil	L%INTSTACK_SIZE,t4
	ldw	R%interrup_stack(t3),t3
	ldo	R%INTSTACK_SIZE(t4),t4
	add	t3,t4,t3
	comb,<,n t2,t3,pdc_iodc_break_stack_ok

	; Don't have enough stack space, panic...

	HANDLER(ISTACK_OVERFLOW)

pdc_iodc_break_stack_ok

	; start saving off the state to turn the Q bit on

	mfctl	iir,t3
	mfctl	isr,t2
	stw	t3,TSS_IIR(t1)
	stw	t2,TSS_ISR(t1)
	mfctl	ior,t3
	ldi	PSW_Q,t2
	comib,<>,n 1,arg1,pdc_iodc_break_no_m
	comib,<>,n PDC_PIM,arg2,pdc_iodc_break_no_m
	comib,<>,n PDC_PIM_HPMC,arg3,pdc_iodc_break_no_m
	depi	1,PSW_M_P,1,t2
pdc_iodc_break_no_m
	stw	t3,TSS_IOR(t1)
	mfctl	ipsw,t3
	mtctl	t2,ipsw
	stw	t3,TSS_IPSW(t1)
	ldil	L%pdc_iodc_break_qbit_on,t3
	mfctl	pcsq,t2
	ldo	R%pdc_iodc_break_qbit_on(t3),t3
	mtctl	r0,pcsq
	stw	t2,TSS_PCSQH(t1)
	mfctl	pcsq,t2
	mtctl	r0,pcsq
	stw	t2,TSS_PCSQT(t1)
	mfctl	pcoq,t2
	mtctl	t3,pcoq
	stw	t2,TSS_PCOQH(t1)
	ldo	4(t3),t3
	mfctl	pcoq,t2
	mtctl	t3,pcoq
#ifdef	PA89
	rfir
#else
	rfi
#endif
	nop

pdc_iodc_break_qbit_on

	; The Q bit is now on, build the call frame for the PDC or IODC call
	; ret0 points to the original call frame

	copy	arg0,t1
	ldo	PA_FRAME_SIZE(sp),sp
	stw	r0,FRAME_PSP(sp)
	copy	arg2,arg0
	copy 	arg3,arg1
	copy	arg4,arg2
	copy	arg5,arg3
	ldw	FRAME_ARG6(ret0),t2
	ldw	FRAME_ARG7(ret0),t3
	stw	t2,FRAME_ARG4(sp)
	stw	t3,FRAME_ARG5(sp)
	ldw	FRAME_ARG8(ret0),t2
	ldw	FRAME_ARG9(ret0),t3
	stw	t2,FRAME_ARG6(sp)
	stw	t3,FRAME_ARG7(sp)
	.call
	blr     r0,rp
	bv,n    (t1)
	nop

	; returned from PDC or IODC call, cut back the stack and restore the
	; interrupt registers to rfi back to the caller.

	ldil	L%tmp_saved_state,t1
	ldo	-PA_FRAME_SIZE(sp),sp
	ldw	R%tmp_saved_state(t1),t1

	ldw	TSS_FLAGS(t1),t2
	bb,=>,n	t2,SS_SWITCH,pdc_iodc_break_no_switch
	
	; we came in on the kernel stack, restore the interrupt stack pointer

	ldil	L%isp,t3
	stw	sp,R%isp(t3)
	ldw	TSS_SP(t1),sp

pdc_iodc_break_no_switch

	ldw	TSS_RP(t1),rp

	; turn off the Q bit and reload the registers for the rfi

	rsm	PSW_Q,r0
	ldw	TSS_PCSQH(t1),t2
	ldw	TSS_PCSQT(t1),t3
	mtctl	t1,pcsq
	mtctl	t2,pcsq
	ldw	TSS_PCOQH(t1),t2
	ldw	TSS_PCOQT(t1),t3
	mtctl	t1,pcoq
	ldw	TSS_IPSW(t1),t1
	mtctl	t2,pcoq
	mtctl	t1,ipsw

	rfi
	nop
