/*
 *  (c) Copyright 1985 HEWLETT-PACKARD COMPANY
 *
 *  To anyone who acknowledges that this file is provided "AS IS"
 *  without any express or implied warranty:
 *      permission to use, copy, modify, and distribute this file
 *  for any purpose is hereby granted without fee, provided that
 *  the above copyright notice and this notice appears in all
 *  copies, and that the name of Hewlett-Packard Company not be
 *  used in advertising or publicity pertaining to distribution
 *  of the software without specific, written prior permission.
 *  Hewlett-Packard Company makes no representations about the
 *  suitability of this software for any purpose.
 */

#include "prefix.h"


/*
 * Unsigned integer multiply algorithm with trap on overflow,
 * by Jerry Huck and Viggy Mokkarala August, 1985.
 *
 * Based on the integer multiply algorithm, but because of the lack
 * of a shift and add and trap on unsigned overflow, the case block
 * has been increased to 4 instructions.
 *
 * For cases 8 through 15, if sh3add does not overflow, we can be sure that
 * sh1add and sh2add do not.  Therefore, we use the same code as for the
 * cases 0 through 7 without checking for the overflow.
 *
 *
 * Editor: Michael J. Eager
 * 
 * 
 * ROUTINE:	$$muloU
 * 
 * 
 * DESCRIPTION:	
 * 
 * 	$$muloU multiplies two single word unsigned integers, giving a single 
 * 	word result.  A trap is generated if the value will not fit in a 
 * 	single word.
 * 
 * 
 * INPUT REGISTERS:
 * 
 * 	arg0 = Operand 1
 * 	arg1 = Operand 2
 * 	r31  == return pc
 * 	sr0  == return space when called externally 
 * 
 * 
 * OUTPUT REGISTERS:
 * 
 * 	arg0 = undefined
 * 	arg1 = undefined
 * 	ret1 = result 
 * 
 * OTHER REGISTERS AFFECTED:
 * 
 * 	r1   = undefined
 * 
 * SIDE EFFECTS:
 * 
 * 	Causes a trap under the following conditions:  OVERFLOW
 * 	Changes memory at the following places:  NONE
 * 
 * PERMISSIBLE CONTEXT:
 * 
 * 	Unwindable
 * 	Does not create a stack frame
 * 	Is usable for internal or external microcode
 * 
 * DISCUSSION:
 * 
 * 	Calls other millicode routines via r31:  NONE
 * 	Calls other millicode routines:  NONE
 */

DEFINE(tmp,r1)
DEFINE(retreg,ret1)
 
	.subspa	$MILLICODE$
	.proc
	.callinfo millicode

	.entry
$$muloU
	.export $$muloU,MILLICODE

	combt,<<=	arg1,arg0,$loop
	add		0,0,retreg
	add		0,arg1,tmp
	add		0,arg0,arg1
	add		0,tmp,arg0
$loop	zdep		arg1,30,4,tmp		;get LS nibble and double it.
	blr		tmp,0		
	extru		arg1,27,28,arg1	
$zero	comb,=,n	arg1,0,$return	
	b  		$shmcand		
	nop
	nop
$one	movb,tr		arg0,tmp,$addtmp
	comb,=  	arg1,0,$return
	nop
	nop
$two	sh1add,nuv	arg0,0,tmp		;trap on overflow
	b,n		$trap
	b		$addtmp
	comb,=		arg1,0,$return
$three	sh1add,nuv	arg0,arg0,tmp
	b,n		$trap
	b		$addtmp
	comb,=		arg1,0,$return
$four	sh2add,nuv	arg0,0,tmp
	b,n		$trap
	b		$addtmp
	comb,=		arg1,0,$return
$five	sh2add,nuv	arg0,arg0,tmp	
	b,n		$trap
	b		$addtmp
	comb,=		arg1,0,$return
$six	sh1add,nuv	arg0,0,tmp
	b,n		$trap
	b,n		$fiftnx
	nop
$seven	sh2add,nuv	arg0,retreg,retreg	
	b,n		$trap
	movb,tr,n	arg0,tmp,$fiftnx
	nop
$eight	sh3add,nuv	arg0,0,tmp
	b,n		$trap
	b		$addtmp
	comb,=  	arg1,0,$return
$nine	sh3add,nuv	arg0,retreg,retreg
	b,n		$trap
	movb,tr		arg0,tmp,$addtmp		;same as for $one
	comb,=		arg1,0,$return
$ten	sh3add,nuv	arg0,retreg,retreg	
	b,n		$trap
	b		$endloop			;same as for $two
	sh1add		arg0,0,tmp
$elvn	sh3add,nuv	arg0,retreg,retreg	
	b,n		$trap
	b		$endloop
	sh1add		arg0,arg0,tmp
$twlv	sh3add,nuv	arg0,retreg,retreg
	b,n		$trap
	b		$endloop
	sh2add		arg0,0,tmp
$thtn	sh3add,nuv	arg0,retreg,retreg	
	b,n		$trap
	b		$endloop
	sh2add		arg0,arg0,tmp
$frtn	sh3add,nuv	arg0,retreg,retreg	
	b,n		$trap
	b		$fiftnx
	sh1add		arg0,0,tmp
$fftn	sh2add,nuv	arg0,arg0,tmp	
	b,n		$trap
$fiftnx	sh1add,nuv	tmp,tmp,tmp	
	b,n		$addtmp
$endloop comb,=		arg1,0,$return	
$addtmp	addb,uv,n	tmp,retreg,$trap
$shmcand extru,<>	arg0,3,4,0
	addb,tr		arg0,arg0,$loop
	sh3add		arg0,0,arg0
$trap	
	zdepi		1,1,2,arg0
	addo		arg0,arg0,arg0
	.exit
$return	MILLIRET
	nop
	.procend
#ifdef milliext
	.origin 0x00000200
#endif
	.end
