/*
 * Copyright (c) 1994 The University of Utah and
 * the Computer Systems Laboratory (CSL).  All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * THE UNIVERSITY OF UTAH AND CSL ALLOW FREE USE OF THIS SOFTWARE IN ITS "AS
 * IS" CONDITION.  THE UNIVERSITY OF UTAH AND CSL DISCLAIM ANY LIABILITY OF
 * ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * CSL requests users of this software to return to csl-dist@cs.utah.edu any
 * improvements that they make and grant CSL redistribution rights.
 *
 *      Author: Bryan Ford, University of Utah CSL
 */
/*
 *	File:	mutex.h
 *
 *	This defines POSIX-like blocking mutex locks for kernel space.
 *	(Maybe they will eventually be exactly POSIX,
 *	to make it easy to move stuff in and out of the kernel.)
 */

#ifndef	_KERN_MUTEX_H_
#define _KERN_MUTEX_H_

#include "lock.h" /*XXX*/
#include "debug.h"

struct mutex
{
	decl_simple_lock_data(,interlock)

	unsigned	locked : 1,
			waiters : 31;
};
typedef struct mutex Mutex;
typedef struct mutex *mutex_t;


#define mutex_init(m) \
	({ simple_lock_init(&(m)->interlock); (m)->locked = (m)->waiters = 0; })

#define mutex_lock(m) \
	({ int rc = 0; simple_lock(&(m)->interlock); \
	   if ((m)->locked) rc = mutex_lock_solid(m); \
	   else { (m)->locked = 1; simple_unlock(&(m)->interlock); rc; } })

#define mutex_unlock(m) \
	({ simple_lock(&(m)->interlock); \
	   assert((m)->locked); (m)->locked = 0; \
	   if ((m)->waiters) mutex_unlock_notify(m); \
	   simple_unlock(&(m)->interlock); })

#define mutex_clear(m) /* do nothing */


/* These are only for use by the macros above.  */
int mutex_lock_solid(struct mutex *mutex);
void mutex_unlock_notify(struct mutex *mutex);

#endif _KERN_MUTEX_H_
